package OpenXPKI::Test::CertHelper::Database;
use Moose;
use utf8;

=head1 NAME

OpenXPKI::Test::CertHelper::Database - Test helper that contains raw
certificate data to be inserted into the database.

=head1 DESCRIPTION

    # instance of OpenXPKI::Test::CertHelper::Database::PEM
    my $cert = $db->cert("alpha_alice_1");

    print $cert->id, "\n";     # certificate identifier
    print $cert->data, "\n";   # PEM encoded certificate

The following predefined test certificates are available:

=over

=item * B<alpha> PKI realm

=over 1

=item * set 1 (expired certificates)

    $db->cert("alpha_root_1")      # self signed Root Cert
    $db->cert("alpha_signer_1")    # Signing CA   (signed by Root Cert)
    $db->cert("alpha_alice_1")     # client Alice (signed by Signing CA)
    $db->cert("alpha_bob_1")       # client Bob   (signed by Signing CA)
    $db->cert("alpha_scep_1")      # SCEP service (signed by Root Cert)
    $db->cert("alpha_datavault_1")     # self signed DataVault Cert

=item * set 2 (valid till year 2100)

    $db->cert("alpha_root_2")      # self signed Root Cert
    ...

=item * set 3 (valid from 2100 till 2105)

    $db->cert("alpha_root_3")      # self signed Root Cert
    ...

=item * set 4 (valid till year 2100)

    ...
    $db->cert("alpha_alice_4")     # client Alice: a revoked certificate
    $db->cert("alpha_bob_4")       # client Bob    a revoked certificate
    $db->crl("alpha-4")            # PEM encoded Certificate Revocation List
    ...

=back

=item * B<beta> PKI realm (valid till year 2105)

    $db->cert("beta_root_1")      # self signed Root Cert
    $db->cert("beta_signer_1")    # Signing CA   (signed by Root Cert)
    $db->cert("beta_alice_1")     # client Alice (signed by Signing CA)
    $db->cert("beta_bob_1")       # client Bob   (signed by Signing CA)
    $db->cert("beta_scep_1")      # SCEP service (signed by Root Cert)
    $db->cert("beta_datavault_1")     # self signed DataVault Cert

=item * B<gamma> PKI realm (valid till year 2105)

    $db->cert("gamma_bob_1")       # "Orphan" client cert without known Signing or Root CA

=back

=cut

# Project modules
use OpenXPKI::Test::CertHelper::Database::PEM;

################################################################################
# Other attributes
has _certs => (
    is => 'rw',
    isa => 'HashRef[OpenXPKI::Test::CertHelper::Database::PEM]',
    lazy => 1,
    builder => '_build_certs',
);

has _crls => (
    is => 'rw',
    isa => 'HashRef[Str]',
    lazy => 1,
    builder => '_build_crl',
);

=head1 METHODS

=cut

=head2 private_keys

Returns a HashRef: C<certificate alias => private key data (PEM)>.

=cut
has private_keys => (
    is => 'rw',
    isa => 'HashRef[Str]',
    lazy => 1,
    builder => '_build_private_keys',
);

=head2 pkcs7

Returns a HashRef: C<certificate alias => PKCS7 container>.

The container consists of the certificate chain for each client: Root, Signing
CA and Client certificate.

=cut
has pkcs7 => (
    is => 'rw',
    isa => 'HashRef[Str]',
    lazy => 1,
    builder => '_build_pkcs7',
);

################################################################################
# Methods

=head2 cert

Returns an instance of L<OpenXPKI::Test::CertHelper::Database::PEM> with the
requested test certificate data.

    print $db->cert("beta_alice_1")->id, "\n";

=cut
sub cert {
    my ($self, $certname) =@_;
    die "A test certificate named '$certname' does not exist." unless $self->_certs->{$certname};
    return $self->_certs->{$certname};
}

=head2 crl

Returns the PEM encoded CRL for the given realm / generation.

    print $db->crl("alpha-4"), "\n";

=cut
sub crl {
    my ($self, $realm_gen) =@_;
    die "A CRL named '$realm_gen' does not exist." unless $self->_crls->{$realm_gen};
    return $self->_crls->{$realm_gen};
}

=head2 all_cert_subject_key_ids

Returns an ArrayRef with the "subject_key_identifier" of all test
certificates handled by this class.

=cut
sub all_cert_subject_key_ids {
    my $self = shift;
    return [ map { $_->subject_key_id } values %{$self->_certs} ];
}

=head2 cert_names_where

Returns a list with the internal short names of all test certificates
where the given attribute has the given value.

=cut
sub cert_names_where {
    my ($self, $attribute, $value) = @_;
    my @result = grep { $self->cert($_)->db->{$attribute} eq $value } keys %{$self->_certs};
    die "No test certificates found where $attribute == '$value'" unless scalar(@result);
    return @result;
}

=head2 cert_names_by_realm_gen

Returns a list with the internal short names of all test certificates
in the given PKI realm and generation.

=cut
sub cert_names_by_realm_gen {
    my ($self, $realm, $gen) = @_;
    my @result = grep {
        $self->cert($_)->db->{pki_realm} eq $realm
        and $self->cert($_)->db_alias->{alias} =~ / - $gen $ /msx
    } keys %{$self->_certs};
    die "No test certificates found for realm '$realm', generation $gen" unless scalar(@result);
    return @result;
}

=head2 all_cert_names

Returns an ArrayRef with the internal short names of all test certificates
handled by this class.

=cut
sub all_cert_names {
    my $self = shift;
    return [ keys %{$self->_certs} ];
}

#
# Methods below were copied from autogenerated code by tools/testdata/_pem-to-certhelper.pl
#
sub _build_pkcs7 {
    return {
        'alpha-alice-1' => "-----BEGIN PKCS7-----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\nMQA=\n-----END PKCS7-----",
        'alpha-alice-2' => "-----BEGIN PKCS7-----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\nMQA=\n-----END PKCS7-----",
        'alpha-alice-3' => "-----BEGIN PKCS7-----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\nMQA=\n-----END PKCS7-----",
        'beta-alice-1' => "-----BEGIN PKCS7-----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\ngZiuwpS/pMbPbyFLXCYZ6edQwFWBz9FEnPHxgplXZruV33k7S9e4EaEAMQA=\n-----END PKCS7-----",
    };
}
sub _build_crl {
    return {
        'alpha-2' => "-----BEGIN X509 CRL-----\nMIICFjCB/wIBATANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQBGRYDT1JH\nMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGzAZBgNV\nBAMMEkFMUEhBIFNpZ25pbmcgQ0EgMhcNMTgwMzIwMTQyODQzWhgPMjA2ODAzMDcx\nNDI4NDNaMF4wOgIBDRcNMTgwMzIwMTQyODQzWjAmMAoGA1UdFQQDCgEBMBgGA1Ud\nGAQRGA8yMDEwMDMwNDA3MDgzMFowIAIBDhcNMTgwMzIwMTQyODQzWjAMMAoGA1Ud\nFQQDCgEFoA4wDDAKBgNVHRQEAwIBADANBgkqhkiG9w0BAQsFAAOCAQEAPy0GetWI\ntQ71nrM9CROU9YstElIuQ3j+gnJdLBhg3PIVEBG6UZ4KnJa92YYQCX8qLUAhePlq\nYZfmCtW1ja8ZHNQsMbD1Jh4dGVMJ53XBlc94Uczfh8Y2ndHrxlFiulECc8Ut23bL\nkFT25vAhnpAEjKCkU8CZiVlI0oCo/7ASX6wf0WcTMgUDPrRnO9uO8K018u24XNUQ\n0mqvBTKjAKCCcnFH81+gyCeULnUQDIZomD+T8m8wm6/2lctJbfaWNm7md4ziDjri\nc4ruOix74BME0rQb96qDSc9jKkegPQ9c9AaxcEzA71uvmAPXsvg8xBHNchgkVq6z\nlbY1vhZ5Ih/ahw==\n-----END X509 CRL-----",
    };
}
sub _build_private_keys {
    return {
        'alpha-alice-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIkef539cUMjICAggA\nMBQGCCqGSIb3DQMHBAhcZ9O8m/uwqwSCBMi8YaCHUvoXFIGzIgyHI4ECpJ+oyMh/\nrKmaLRyuHY2d8sBJODb8x0V+i8YyOeA8pOQqgu8NJ9Z+bPcfYKXrmJOVz0xYD0Td\nN2l9QrPnu3T0iXmZFlV6tdh6GTsbq0Ik5Nl4oDz9ue2RG0EkaSqeM5f7fsJgqZ7V\nn984e9bBzZk6Vz6Vm6iZdVKINdKKDmpElxaJCQG0W9ipZjmf0Z6CNf4VWuTSgZW8\nGxtSiuRDE7eNMei1Sx3kHXDRpVIchBN7fY/tSYZ9tml0aHEzUHiEy6BzZZhxNUhA\n928H5vwbdgpOozkjjWJjwmLo9r6kZssSGNpUT2R9awfnmeelNseEJz2QeE/TIYZ3\n6666VgARR35Cs00sG3qD/ppOGXkWPcdVBOMzHW+AnNZEf2QQB1oy/hAduWrPZ4P7\nnlBdYzmCfJBpKjWJD8SEq90LcMrn+Jopbrw8j48Qk2aFm0qJIhxkQg2NCaM+ZJaH\nOtbhAPZJSjifeVOmUtiAd6KILxB6kvYNaOprCbQ29YUeU6JoxhuVwKgbelhqQ9Bf\nw02lmRVWkyDxTlON98OTIgvZ0pN8bSPkmJ1Rh56MM3CkqJ+lniJo3GNyVAMXkOc5\nVz3C2h9k0aXtGXgaqtFpWFEGuCqbUVATdVriISV0uJ+qf6u470grnoriYinCieYZ\nuByxqcGTKLqu63e3gfsyeabYHYKuHYz0CppSgsAtHjRCZVUTtte0YksOXAr5RrES\nNXxrnu06+YYmPsGg2RpAZBynaIWFdwP5zYQOyb8oCfUKWart/vyvLKnSZn+QUF5s\nkV28j8WjSs5gTJ53rW2JEfF/ueDOL1MleDFVNCmXberubJitK3qsFPsLkzLwO+2T\npmgcZ4di/WiV2G4/P5lOTN89txmtQtKejjUUC/3Vzq3QxWviSaVj5/rfKm1k2VvI\nmUhW4EJH/NGi+ngBoBHogCXm9jekuNLqk5YHSAZ1UljB/eWMckgfVyVTgsVZ82wV\n03BjAHf0xZepyNZukI7omuOHmoVjahWJf+yTpRtPB8u4vKTCbw6FoNE3kLgXpBBN\n2GgxeAsmcbg0ShM8mNxFdrvUB+R/CX+zXNYtXq0xyRTpfPUd8jBPUHDtI8FNh/vq\nnoA5ZlrlVgJ5V33Dcs7N+EDCJRo0NruU2Ds52EyDj/EFPFEXzJCbvHlWGGz0ktYj\nkweaRiiQwWmpV1/KgfKa6YpQ5o4vQiF3Fv4U0rikBgOsjxppTCIKkt1awfXPpKpE\nXyJPeiR0yhAqupd1S7BmPv4U0LhYlh9rsnASc+iYklzSBJ66ehp2Vzur2UJxqtIR\nD2nLpguLvbXDfonckDkkji9Ee4C30q4VrVvFmuDIW9hFOKOPN3VG7tzRFWqVMWDt\noaKVM04lT2RaqtHmXweOVJyzB6nJk653q4HduRAqflcM2fEpb8uYBm2dFxFX7idc\nu+iZQ0YNlPybPKtP5mEzOIpGFBrECIYoE7OXsJ5D6gixe9gTUtnPfiGw0h6WDQ1s\nYvMZYJT8cq1TFpwam9EsAyN7Tl6rWyK9cqCbUETionrhzxFfBSYWume6PvkeYIjt\nS0b/138JCgVQV9HGAnw170jNjWRhLv5d9uzxJ1qy6o9J2JgX79Qfy233L9R3cg2g\n0QA=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-alice-2' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIbS7mxb0z8NACAggA\nMBQGCCqGSIb3DQMHBAjULcZINVBwjASCBMg3BWpXQu6cK1JLqXaKsiGi1vO55ZNa\nbGWLf9g399NtTjy0MgFo5isabZps1YYmf1Cufqkts0VXKaO6M7wu51D2YaE4XtDj\n1zJHvNsdSBOP1d5aZOsPRol2fBECfTIq12G0ulhGqQ2N+6KvpHsGnYDUT5EFr3Yd\nwOBSYJCslSSA/xYPC3O7vhg12zAb4KOdGz/6HZqYtPHNJ/cTMC1PejHBgNUW+MWT\ndHLbnINJOnkxn2Y/VERFCFNvZr0loA6f6zoB7R1VvUWAc+kBrt4v024Tm3lhuOUD\nczn2D0K7c9sY5sl+uXDSG3l4MF2rs4dvEfvPy2FTaq61rXmRXl239LIdj/w53kyS\nHHTTylUZSab6rmKCA9eVE+KMlx6tp7ZHTKhRz6kVHO95NIsAdss0oLIfjXbbLzNC\np0A3rKd7M2aChrLAwuprDHlAu5zDlg6389WlPr1aDl5GGEwWoBxubV824iEm+/Y6\n4Pwmxs7q7bbM/hmERxysTYntT6JjAWAPFfA/4Ry56ShgYyRQ+JYikOF2wsetCLzr\nCnzvfiZbSVZflPWzZUyk0PgZ4c46VPSqysfDFLmPfvsAVMCBQLWLA8RXLdbKaWok\noxjXC7su28yMANBqz9D4Spx6QDbngL7WmkqMnXu7j6RQkeMLnZKXLRKGkszyv3fg\nRgbGXIvjhx7d//VXM7Yr1pf5aPZWJxPP/Wl1M3P4LCfTKwmHv6DbOH9IRJbS2ZIk\nsvcUL4oK/MhrEg0EN5tvzoAIIwC65X5kwEA/CdrDOkg5YuPC2P+EbNA++zQBcCEt\nW+kiAeXt9Dlg36L2RZOVKFqdDBUleE2eszt5+DoPEUCB8UqYlSeTJAHjNjHKGFBV\nDmmLrkF9zSaGfAwZd1rHb5B56VaIR6Lbnjrmc2OTUIh1wj81ONL77iCzXNFtYTeG\n61Arkeh1eVx+tMg/fzLxW/0Ah/yjPNGvCvrraxp8mgnxY6tisvd9os3TYmoTYyvJ\nwZ7fj+ZFaYicLsRuxEbYIOFt4A5q22EIqc2TbLcpCRdBXeAHcqolGp2srJwjHZZ/\nphDwo7IrIkM+R3b421fnMxii/f9YQV24W/8x5CBGEHV+Dv0DgZrotVbVUVIjpUnD\nvDWtAso0m1ijvp4mFJgDFcsg/szDcQXDw/8PYNDohU+S5OLOafDshqnE5BRfhL0G\n/ppxS1o+Lz1jTntyB48cDf6QI1hLXO4m4ipAjycbqa1asTO6pQ7n8ZP9rzugz3wh\nfldjOJ/2JPSQzwJabcIRmYVrSCWf3dWf2COnBrfQgWJRpFuTOVW989m8NtAJx9Bz\n66ax6jted8CreTg4YWH9RxfXi/AKaj+GpT0Mq10zWlCTrkNQv1it2rF4pApzVWMr\nhxQKGHXZoLzrlP/E0vknUjnB3Dk2sGJFhZ0iqAb15FJaRMUKIhns1pxEiIV2R87g\nHYAUhLrvmaTfu8BGD+gpypCtQdo7la+hOKygq/SCLv1BNQ42y5Hng0cZ2S8V4oY2\nd7A/6Hrx7IoWLZAeFhKCxBvT6lwyGon3unzBjQGgfvgQt6UcatvDCEHMkaqPlj4o\nO58pwwjGwSqLV86jxNN/DBr3HArC6bYU6uuxTJzAXblomG7MsDFR0lxZqwxL0ZtR\nQXM=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-alice-3' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIommT179loaICAggA\nMBQGCCqGSIb3DQMHBAiF0WIdrO8sZgSCBMilHTSoJrz6uPyR9YqmpzpnEzBvr+Lq\n1l2ILm52DwYJIaKbcllYoBpgsuY1Bu0GqWzHhfZkD0hWOiPui0SxrOEowkSbp6X4\na+6lmf7UB3+YYlyNi34VianjalT5zOnpFKvFecHYIgwTTokxNMsrp3HFOT7VzSUf\niaoTvui94BrGCSSWCHIXVTXyAsgd8dYfBiZ8TrJwuPhSbiXon6B6lnvcTZ0ZUBc1\ntYtqcQzP2eH+xlauiOzAFQp2SeEs/HYgRa//SO2fyxAg3EkBl0A5/rVUod1+NOqC\nmzQem61aSSQlip1uqxKQ9vfZGrGYr6ECVR8hC4Be8n5cXZZ1UAW80Ut/DCwUku4/\nRS1J/5LRHOGbKH3OepLZ8jn3n8D8WRbrpSbxTrBtGBdzClDqrGr/o5x+iki+1VzE\n30rYeceOF2crGwj4pCIRVGoUQV3xnfJvCaw1ucOMcwqlXQOU/METYVoMUffBIplY\nNBqmJjBUrCyysRufoRNMU3t8T4+IPog6nPgXIxPRjVji8G7RMn8yNY/rnjTbckUY\nRExI73Tn1D5yNqFbH4+6CLV9cz2UAC/JGWIrfiHvRrwz4gKDRawTRoFWaB2KSOAv\nLkE+44F4zJgBKE62c3bszSCfSDHinC38lEzXUOAQRMYU08NjPkB/ERiJYkIlzQW7\nOiKb0ksuhVEUxrdXV3qraVJiN8aJugWuOS/+nOjFi3xczgNhiohFLa4inQnuI1rb\npWoPghbTDwKXxeJBMRCZZ7RcM5LfKGpUwrRDUt9kQUZlsxGNvdgtasJldZYEGr+w\nOPbEqFMLN5L+Wg7DhB12DnOaVsNvt/1JxAOYw38yzqOsdCjD0jVheA2fKuU6Iw2/\nLFrT2RZh+yVzFUHnCbWJEYls7Wsb6C6ncFf6p6CBtK3QSmNNwPLJwjYxraS5neLL\no563YCf6gWBg0X1yWijMqXcjyJEzU/OGeMI9s6CYD1EM96Zo+SxfBpJ6QlIrTjeD\nCYmkLC703Z/5PT7hD3IRhReomQ0DebIBvhz5BE53QyWhRtiT05u45zhLEGGckXqB\ny0qx+NpBGM0iJ7e7Cy3z9Uku9D/3G/iOV5TfiRzwJfcIAeM8ohMJI58EwQdEOqLf\n9AAcvv/nQUnP5HL1t+0q/2Ag8GGr58Qpm8hM2kSysmovSGxpHbw9Fcyf5Zw84bXq\n6LSEwVT1hbxN0JNVnG37CMI4VkRta8EtuctpCvpyLElCnT540yS0GWThm9QQfPxf\nh+HrKfuZPsayY07byPMUakdxUd2SteQ7pgwgIEd5bGyT+23+PjD/sz9wQaqOaDNg\nD1NsADv7Xo2rFEfiOOFXqr3OMVZFkKh9sPqFiI0r1/cNciD4yErcEx3bZ0gNWNiP\nhGI/InZ8r3mmNOgvvz877gDcgPvKc2HUbVBfC5conb+j30PA08vGk895kLPQszUY\nCKMXof99U5foJ5evYfLZKyLVDS1+QwsnNVzJ0tiO/L9bgErSQFkTWA3cBH7gwF4T\nVmcQ7xEdAjfgdg+rW7x3eTE+wOkIrK+KnP5kzMQAwcbi/M62bH+eRgQQiyP8QceT\nePCCrwKZZfNAmMUqkRTmlefyQencwgVw8JEG0fchGTe+/1S5JkCyiQOmMMLNP9c5\nTqk=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-alice-4' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIt5XFFj0B5K4CAggA\nMBQGCCqGSIb3DQMHBAhei0LAFh+fjQSCBMh35Nz9FoB1DpOvwpHIwJ3KtxGoAK8A\nulRyOW0xyNyFfD/BrrqkNf7njoiRSp0P/6erwNzjWB1OuQPVVUyG4n+KfGzdi0mV\nplvK/erocK1tSbo15pomtpCBGBk5RQGrX6holIXxo/GKB40kZTTLpfy4Yvsgpkf9\nLaBmjqSoSfSEGTWwwbLdT5Kll5EggJoH/pa0MbRBWMgS8/6l4CyA1Wu3FwUBS4+Z\nTIEKg/SJuonN1tsRztlrYnTQbAc+T3lRLl4DK9BzUVP/jEquxe5hLEko54devjq/\ntUTIg8mqLuZ/LGqP8+0oakWnIcCJSdMeKwtL6u5qtVcxTyLJ6kt9V1luYyzzM5fN\nXT7+hPUZNWSyE7AbaKuosGNo/EUkODbka3MzJiN7VPCVWCRfNasojjByxgZFxN9n\nsOUcqeV8kbyNlQkPQlKFp9ULEcCC1Mr/ouhWMtlPRN+cATYx4NqO2j9SLPj4+tQP\ndUiQ6wL3TLLRcoFX5vjezKYa7V87Nk8KBM5dryU5lWXrkmm6mT+3yyJJolDuUEFw\ndfjyqt5G3UQE5vKawLT2qR7xsURyiVc0y/leGpElkoOUP56smEQjnUUGT2Ezc5RG\nbk0gMr7k6iefeR8kxU0Mn7ukLtTG93ROCpkGgovcEPRWRlTc2fS6jxpr8/1kq5Tn\nx71Oyt/jOPJSScel15bxLswRuj+ZlPP+5ieVMEwt4lZgQdiGyDa+FEecBqM7btbO\ncbUu6wwXYMd0HLgZdnWtsfJ4+ZFg1MyLvBescZDtj/Eo1Y0oRijY0yqXn5lGK4pM\nFH/3XHFhsgmh/tNfx+ilFr7DYnLZVDTz5ado6XApr+9qL7g0T5LCJxgjTV/ZIpNf\nodfpBh8rhuIBA4TIZ0E28pzo3x+LKZYz1G7v3A9EvGXCHdXm2CZwv4ErWjo6aIga\nCeXNDSnRjU+ccQuiO+HUXbg3qo/2Tb8o572LzZ21eeO4MozKZX4nxAL2dRafcjE8\nudBXGdQtlz8KEKX1M43A6p2lA2bIqipUHost3F2+y/GxuuFCVR8cNE8wv0qUyvdi\n1v8RYD8SvyTM0eOLRYAwjtNKxmikcStK8S0kwzHbO2k7rKUjJ78UBEBbRgkfwAFO\nt08SvDdTKrH5VkBrL76J4zjhezVAldf0Oe577mHMpix3cotrAl0uJZZg2151Whns\nWtU9PiFl3NA+vjSI3IzLbmLNpsUX2CdkCLBomiLBF21ckazsIBPfnbYud6AcgIDY\ni2GGT9YNFqn6XRINT3D2fLpbmVJj7QLU7dGXUm27d/g3cutntjp1vxpDV4tgdTxi\nA8MN3+uDH/iztb25Dunnbyks1f0GPvztYMd46fWjnB13k84U83N9EfXFWeeQF/Gl\nsN0F3DKcSWwSJ+/51GZMSkhCvp722Im8VrVSZ4udoXy1vHqfl6SmBX3n6dR2bZaG\nkpiSLxTnkTXWlx5gmbf8JEB5g6z+iw2ets7vRP9X9kh7Bi98S8jvmrCLE95EHoQ3\nSOsxwUIwyLxc6j6cgRivXA3t9F0XB/AgSgyzjChsajpTG3QyyhvM3pFt2gi2QChl\nk+k7iHDMjEmsWBE0lo5r0qe2kFQfxFpXRaQEIxrdBf7IJZPFQikWYG1AX7PMaxkU\nDwY=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-bob-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIGLqESjTYLwgCAggA\nMBQGCCqGSIb3DQMHBAj/ySR1dkn/NgSCBMgDjPrEsJU4bb5iuYUrXFjSzJ7sjBxB\nT74reLNHkIp0CLjUZzjqAXnZeM+FLkh1LxILUMJUWWCiwHpRx6ZQQopmC+qMWXar\nNQ3d9CR42pULUxryHiDM0wbtozz2xnwNHhj0vRyS0Kb3IZcWbSbbfxyxw1QeRiAM\neeAbu9iUJWQXh1TGntSZof3389gY0zRekM0OhSg6mfhraYYs5ZFbqoD+MMpTqcIn\nIrYs94KOaWQiMPEcN0JupdzFztTsze/kh3rBpWWblzKx30Mgt4lBDRkIvM8gHQiF\nAlU5wecttXAm7q0HTvHw+aVO8zJMX5Y9JwZ0P0j9Xpr+8JQ9pA/BaRuNry1i9/HL\n+u9kjnzJoiy9vW37PzC2wHspnf44bONqKOaFQuTGT9FHK1Ed9cgaAHalig9a7epa\nrJStIU9iae9JQccqmIt1X8J47u6Ni+FUHBRpfZVouR6Z5zOTuK4apUaEZaFnLSxn\nL4o2NRVz3I+I8AU/fobodBn6lvMgAfv3BcxXosBDA5yMPyXs3vnsDBemDRuw8oCz\nULDafUnN35bgDcpU0lx2MrDNd525QNJrisxnlamZhscaoiFHgNlNHO2IuVKPsXOK\n62PskvzUHokrkqGFez1tuaywrbFHlIfJmx82NSO26dLiQ4jeoYto+CSInp+4cN0l\n1DYh8e2Rmk/XKIg0ughhb3XfgFNuaRJVMAErI7+I57h8z+KsG92d/QrUzVrh7/j4\n7zJcwzt2EUNQVnb6tCRF273VOLOxFi+88w3y6dS8DghCslxXShtERzXgQtGXKr8q\ngqkB/5PMhbla+KMFZm+n8GrDuJXNBWlLNXwuZNGyXM8OZbCt1CDuJRTN3UJ0x/kn\n89WSg5mYWP8OnrCFtnHtjlHDqKNPKspiPVFF7vwY4xY6JPzxL20rtIKzzlRcXpHh\n2Ip1TudOEcjKsbuSqJTuMByEp2OX9rd3l2b3vzEO529RKyWGPBjwqKzkwPAylKWI\nUgeNU8E5se6Bcalsg+LwJpy0dnsFU1hZUnz+Z9ZM7WZeazriUSy2hvcA1Puhw7qK\n0e8a8vbPI8TOLS5k7hKXyfh1CzUl2FuD81Yq5K0uckfVGzc9Llg7lsWseIDB3HA6\n/6Tl5Ucs9vYFHnATYrUsT/EYR+vhqBnerDnUSS7JtvDV725Tb241w7puB/O/J00t\nNsZaUNOTSB7MIMwT/JJK/QThN/nDGnR5vAXsipnJuerQuIhf8xwO0ah86AvX9Lgw\nQO9Ibk2kNyydR+GygDtjgC6uBqPyZlhxSDCbL0IqRmrwM5XnFKejfMEweWMxqzsb\nO1A5I2dIR/4S0g6uB4ka3EJWReEfXM1mGc3OthjeZY4GrKPYl6wyGk0YmKN23cNz\nQCBm+eoAjv7KjVcSLCkZBzKvaH1YTYUsw3vi9lYIjQF06VBcJICOFKUHFHKbh14J\nnw7FCUEz0LZfVo2IDJzE6imWfbiWI1JEzcW1zN3JMINOfG74yB7nmKyUg+aJe9dh\ncheSDC5iFWxE0Tg0l2HQ5lyoA/MscPd7WklOLongpXaNeF7nQBpqzBkWVw0eaKaw\nhICOlbm2MAe6dxgk3Zn9deslRtiW/nXZAcX0CSd7K2F7Wul4LvD+pYrfjKRpbqjz\nIdo=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-bob-2' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIHaWYzUoTp9YCAggA\nMBQGCCqGSIb3DQMHBAgIrZVw0eoB3gSCBMiUe9/Y0VUXBO4y+98QxoH+LKMCATKs\n7I7QjFZP5ufrt44OWD+dnBF6jZ0IbaV559fxGxxmWyNwhS+PFfqCIAC9wcJb+IVv\nwjR6gM0Jxfg8R2tfzg7N7dIvtxtqgro5BQkj3cnvaSGqnSDGJfj/Q9Vxtm6xyCE7\n0rvf1X4VE4KvFegTE/gfTJ94C8LyEjAFZyIjFsNJTpBsan1mx61+v/jSDB/Qb8Qt\nE2tMWbRTv/iiuWw3Z+sapRdhL6Lc0OulJinQBgfDgdfqmf40TWpYc88FJ+J7XIUN\n0Q7OD9Dv3/ORw7I5AoTDklKrGFGYPJqryZXUCTB92CQHGO/Fz9MYRm9qIuimfzm4\nP4ztsP8mgP8V1w6MVpPbca1lvTx6vBnyX81nWfvOKqWuf3IAYlSszqhUTfBDqSKJ\nUXL/hvwaDj4kAeVlY9hVto592lN3fCZiyhe7yL6NllPuElFjbzUcTkvA9xdPUwpI\nBXIA2cAdTttTUtiPWff4/wddLwZD9rIDksE0Ngbo2a2agcQRHtjfARzQ76xa/aNu\nu85yS0rlOi2c5ZHwzXof54+732gryclCJA5ytSzmDLrs8mdoXrEEDnr4AoZQojXR\n/LuMegN29tyaWSemiol7k0jPhOQbm9x2T4S4KdbQlwHAy7pmOFqfJ87NgZbqdRkb\n9qVP7aoy7+5fblfVFnGJjOXKrj7wqErXVdx/SU2NmCjktSyoaBHaSttDoAGRCkJi\n85i8f1PSwNLjQVjxok86zL/d+UBq186bpg6n2N0PSAlBdbF/sgbKeLWm9Ix9A+h4\nqkGzsJavVHJGWbRa1rslZCx2FkeJAZ2Ovu9SYY07jTcec/iiwXcw/uXkXainNxUw\nNnUpZZQPopGaQIj2JUsierxlr5VRI3Eoj1XMyFahg3bbssmwjdm5/NNDVpXoQGWP\nfDwOvQ9YyivoTC3sDVlzVFomltzNJPvuvS5BzRK+V+I3PnMmy9/1RWVlvAzHEDsC\n3uLSyhYGOKZTo7qKsQfRcLWDZoFxiijlLJaTt7O6FXqFXpp3VP/1BiUI+DxME+5V\n7VWvPLeUjMfOPXM64CvZ1Z2BjfTN0z3nNh77GXYi4OerA32gKDnxNGsdnuP8gmCH\nuOaqd+7FaWbM/xuj9HAzG+HKRPLAB7z+Ssb7xQ1Ud1VEbNT+ELs7oJA+B2WaalkE\n9k5vZZt0XvJLOMBsj6jj/APrMUCJIpuIgALXpWSQZwsV/HDIs18JCy+0wfTgUXuj\n8UX24Haxe0lsuW+qHptDnNBLwBvup+b0Ot5bAWRsmWd0s7QrACeWNU2kojAojo/m\nz9q2BNJrldrtDSQgjBhy855Sgq/l3aO2b64i0YbeUdT6+piFLQWyjcLP1zvHpYjb\nNOcWvHncOE561om5dB7F0S5OtLVWHO+mE9na7Ctm6IRx/aWL5/ul+I+IaziFnQbK\n9wsCxdW1Md58UbA6u0/MOqj1qr5O6sGwQ3Ue9aLrBrKSbwafWUa9xjgfmJBsdcxD\nNZ0e8APgEYNFHCYnTvLvsp+CZHsmZRGWtq4Afid9oE4QjIHcAR8pktrFqm1jjXV6\nST4xjTHzCekP46IlFt5TKssvhBy5zp7NJzin2E70Oebjhn2GZpn7Mb0rqrDh2v6+\nESs=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-bob-3' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIeP2T/J5jXCYCAggA\nMBQGCCqGSIb3DQMHBAg21LQ/72YP9ASCBMiJtu5aDsm3N/QDmEdriGhhVBBqvAbg\nVUZHhyLGNHlycKSIPF/MzI3kNr07D/bvD2fwQzExt3PrXeATGCbdIbXGuHlTS/K4\nYjo9MzV6ZDgRaq/F6w5qv65q/XNNoYA+S2nCmvTGW11XhoUNv7uA6tOIcBQMhucr\nktZQpNBifxWL6X2wvTRyG9As/kTFkU4hvPKkjzjqs2jIRS2E9bNEZM3rC+vxYuQN\nCB2ogA1QefA55xX/h+k5LFjtPeCSZEDxZmjWJA10wV4AezhIvqWD/FrCLC145uTJ\nXj8ISYswq46wEDj1xegtBPLK9nhORn8Qp4DmObt+AMBs+OUPt0YUOE9bErBCLPA3\nkSjMfDYMUA53eR4NSqDsTOu/tMhXSekdmpgUU2mGSgToZrxtCxZw5I3VaDZCASWi\nAR4cfAvAOnGoiY7wg/wGM9IeRpF2bTjv3yAJUBp5D3Zlxyv1oDIbKrZ52bajY77I\nyx1rSgoHc0xUmphf9i4NFL4rSAenZtrMb0Mt0GMAUgeS2lAT9EqkjrWQ8d9S4/r5\n3zZH40qnODwBCK7fxAZgBeO7rTosygW0jgUh+0j9Mt0JjNPCMS4eifheG6vnRcxq\nldVUsS6oLWxzYA0064BOoRtArPVk4WdXzMyBnDT2MkMnlVbH+rQJ1ikU+OSda5FR\naL5tBmkTUmCqp70KOrzq5Y5YAHDt8p0ljWqPpnYjtCe4n2phITSlgllfktqcQPKw\nNlZWjwoi6VwEBD66GRSIZz3vkeJXyOtndQdIVgqGVkoqCDyyCnGF35Qukf0jjqyk\n6fqqGxXarxRNrW3DHolfIc6l9A2ZFn8STnzRvfZZWkSFbcjRZnVh2+A6FaTAewxS\nii5WBN/vE4PFe61sxrHGB6ITLDjZT7wylE8TlROloHgI06samU8edwt3IgZ0LMAu\nac60JSkdgTWCbpqIAvDITDI9QrIzxuSSyLhiCOD5XxIjjmHhYE36jbPPsZqlOZWD\naw2sOpi1NmAtfE6Eg/3bJs3Tt/Y5eMd7k3f6nBVEHyPtVB9Upob387Y/g8sUANLR\nsurb++cyIC2lh6yYn2Z9cNUaZ+5xLn1WMQ3m6O9+/DeRkS4o785XKqfCMVnyeyya\naFpkAyj68/WI+T2hYpAAVTN5Cz8qqdEu//jLOcOMiAxyZliUWyfA5Xakp5FP2ftM\nda0bzK0RQO9HshzMeJSQKu1/0G4xVu+LhE/X7+dkKXRDpwxtV31i8p41gbldRZrA\nWANO835zV6ouTS1U0EsL7fMq5POhf4Hnooe5JT8fkCAJi//SGYTZps5mIYaTgy7n\nN1CfmagTGNS/SsCpW4fzZ1hFRqs+lvXV/YpitpOTvj290kbdwvp40cDC/ErW+QMd\nG5AthLrygDxH4nAWKsSUip2eRZChxeZBYZxcmmGxeD+xgZ/ULbptADhZPvLzy1vd\nHynssjKKdOsHPiRh/oltnb8yV7NA0UqqoauH2FyzQ3gC6/3Kq7ueX2T30gKvMb7B\nbVP/V2KBJ9/Uwq34tTsXRxpj13wtW4/a1Isl5fuPBBNVoJOdKVv6M3CIKytZDfYp\nuNCJI+i/GA6NwGlyPowpfa+nFm/KYebym/qbBNvJB4mMHoxHZnc/ywbpaUKvV7Xe\n4hA=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-bob-4' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIhcxG0+YYrQICAggA\nMBQGCCqGSIb3DQMHBAgLnyd1gDbI+wSCBMgnfgB9dZ8tUoknJx8ZHh5Y9ZcFWBqV\nSu6A1Qkh66zRSN8CBdx2ZCAcpGk2DuB90tq3X38g7qmwcjzd3mFg3b1I+B7Pg8Ro\n6tCzagF2b20cdCPI8r17seGtv2gJgbnoXUs7uLEJ/JlPD2G4oxfuUvsfDoS6Wn4h\n5ci3TUK+2aMp0FX7mNezc+b//PKDi69jJGgMkO0/cxtR28a38w9ipVuEY+S6GpFV\ne4/TrLgFSrJsqMCAMEvB8xaO15UpddsOFdrozsVOsePlwVQVKTHyl06615UjMuYN\npqiAZwhFb9N6rX/nzyNWNKdHDevHdfHrivyYyGkBoXc+11kPczWlmOUBawEwJdtv\nXuBMbWqTGCyhBvNkIWCr+Xtouo5b9Vigetsa0iciqMGRlFL73zWnMnAo/zz9/rim\nDn99eIsdBBIbulQHTARyKruwqU7TrkfXp+INrg05/IvtslOzTX9p+g0Z995zn9zt\n9DFhN3Zj2wjvco1IxaXct6n4SlTngMVXks/87BBvtwIiS9XAeZt1yNY4reXMYEyX\nTlnb94q9gwGWKEYfVqRiMafDS7YIqCw0yj5DLckSEDX3gwGyuoPiHoZ8xesN7w8x\n/X7QHkbMJH5trla2IuYf1IJb8eSK4245WSDg78Nt8OsEOYHNe/WEw/GFDALY9foA\nEpv3FboWqpctsTNaSCPTd3pYtdKiQBc3gAMvY0tR6MG8TsPkGXVhaPpoGVOIw5LA\n4kQvyvwQ7tpL8Wf9NxvjDrHMvajpz2lo2gnnMPoPsuodAB+zQzfvqCP9YZ+Pw6uV\nR907YpFaZed0PVr/+Rc4yA6ZkqBCNLhd0b5aSTxbT35SOEFqTJ28LEyqqrASNR2Z\nznkv9RyJyccSHZNRHqJpz6WijH/5b1muEGRKFBx3B3tbrrFod3zCSn4M9MytanBu\nSExNtpqcJEjgAOGSrLYFrUzMVhIVBKBW5QzuiR/fpvW3fWVmdlC772JSIpFPOGbE\n3Q62y49m3OzDAit98nOpVVTdRI0iDgO4MEtK8o8YDsmXHvDvP7SvMRVuNfDciigu\ndrTPydJVKCalWdNtc3+CVbfVPg4LpIygqfDM5w5MVmLTKJ2Qddi85Ow65KwYjdBs\nxIMp45LsdMDiavBgUZrbkZwEZ5WxopnVuI69qC49GQu6ZzqiIwpLkcT6Oc2rfYwO\nyEe+P8+pnCG9XkQXh+r64C8lE+TddXtuyJGeVcRvYsCVB2WoTADcLcEAoRzpHLjO\ndZheRttRGky/SP50HVvNopZOdORA+yWzFi2Ii+cyUDh2IQ/UBRQbMkSSaROUOQL0\nzeABnTik9HwU4x5wrTiAa01Wndp0sJ3nIP5zjOrWIH9bApSuummFu5gpzn40fUAJ\nYdnLf+ekFf0NZWrWApPnXzEuRlGtXSYVWWTXzCZQa33UJJWGVnnPP048ScVAR1lq\n+wZH3kZZULcdKMkMN/hqU3cZ596Oz5aEQRH+SfJHNzrVzsbbDqvK46kEdk4BMFrB\n7jpEZEZikaIN7v/s/cuN6l3eAsbxRdNqudlHPfKA7Wa2h2TxsZUuMy3nglNMnHul\nvcdfbzYZFiIcCReSM4ZLQe9x4Uar5ylzQrK12JwkqqceP18ngaJqg/3oF4EWRiKG\n5lA=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-christine-2' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIpE/yMz58k2YCAggA\nMBQGCCqGSIb3DQMHBAgYMyckOTgPIASCBMi+2Nh8subiNKr1MY2uHqtZYbMZDvdc\niSWqdZ3DLkrbKowtnwKazEvTGNOtlPFqJVyJODPDko/TQh33fIXlruChR7A62GrR\nClg5xAJb68gmMhKAS0UGtdb2SzXnxzXjyMBZwONYA+fuTPRhJkNfSEwV4hNa67CU\nGv2fNA9zWxOzt1OK352dCpQRDeFACch/USlICvbDKUSXIy2Skcd8OJOgP/tSrsU4\nB1F7ruqKhn4Ctt0tR0Eae0hU2XUwxYGfu6kkDAxR9baDFkWrdH25iu7KqsATjEax\nEb92o0cD5UQB3P57pfCE44jzWQQtQVj2sb8/nbk3Sbn9CYFoPrYoPnPinXQiCSyD\niKFYhuEHTNgwgxdqGjVi/Vf1rbGQB1ae2eY4RrNuGe513H1iWTbbeL+Jbm9QNitH\nkd2iYDfna1p3oYtt6kRjYKYvO8/qnA0EWe6hZzj0l1kOZ3dprwAtJYJoiU0QUiXN\nU7WFBTTwJuH4ThdnwL2M2IXY7fgybuN10wXvywjbEmQMAZ0wOr7VwhKW32N1lj7S\nDhSKjxSCx7dYVRwV7N938gN6PWiRYrKwrCwVod0mHcIEQXSUK3uRZWlAF5YWK61z\nVus/xHsVvuMAtDxKrxEGU4c44affgZG5tcFLIDaQMaa1i9KvCMMHRy6J55vb6U5D\nkGvKiR4Gx+B/R7R6lbfxGN1I8PPPpLcsUFX0/bRhw15P77a/lLpM4pguj69JK0KR\nowTIutwh+IdI7WuoT2l6wyYsDQyJZKXHgaRaKaw+deqSGpOAmyTMbv+Xm+9L0sFa\n6BsuGESywLMQW0VfNBG5jM2JOOY3pJOlVqiaOxtVUcs/t+Laldra29LawMmcCVQR\nGv2W6s9GDoZddG9XccPU1eF80HAVt31sKwIAj3CI2DFT53Pw2W9cW+vyiYFG+h1h\nhGIrGHA62Tw7iBD0RXD3gZ29DL1COozZzKUHTU6+qvTM8esrhOotorNGvyLohlPi\nuh/urXH7ApGruwCT1HVU9oIwDFqR4uGorkwxWt5bfJWqggHPWTSZZRwldcCnqpRr\n+FQP2FqsBGAIW7Cyttw2HAe5Fi72NRIXSv4c0/CCIJlG1AHBdwF8OIRd7pUAdTFt\nwlUe4m8ifdwXxQpfTPmF3JXK4L3EZx5nZiINyp2T6GZrRAtCHX/1HmZa9DCX5eov\nOiMIiBUtXXeKa7xZ5yc8HZ//kd2KnQRH17Za4rtaq3lqUSMqLDyHGx2xVyWnvQxO\n4n1CbAol95qrKSYI9TUuGCjRZ/GCRk9O3JqJrNursmNx4blYu9EU9JDO3KVG88YN\nVzMvrCA+XPYYQuquK8Q5ermTNuwtlsO2A5wiRNkOcf/vVmtDJTqnOxGsRA9bD2zH\nMwN/r67Z/1sqj1eoeN6o5KaBGstrDwvZ2/stKRlBX70vgfSLHrU85YqfPtmFhxLG\n9JCIrPCKg6NVWOmtqYQWhrBcQm749hCC+tMORHeJk0SAQfp4ewOITBX/dxztDkP6\nCg5JSkWocdMz0K0WnIc3Rhpqs+XSxLXOStc664Ewc85Qo+jEqDNUIBYog/ztRP4p\ngtQs96MjVJVVnEFuHNNKlcEwnHhaXIUjwCViDgynA2Pwx4oU6ptgjQAHNrxK/QuJ\nLwg=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-datavault-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIRADbcX3yyIwCAggA\nMBQGCCqGSIb3DQMHBAiSe+j4+QAIkgSCBMgxQwEvJEDWuc5gGDjMw1uYbXQCh64q\nLvGEaPDh67DRgUzkkWs9RQY7+qNpCMA72Yjd2qC4q3GfMVNroJXtqjyTcgEuAdDu\nqpeb/LD6cpnRLK8t722tCtS3IBflHia7j6a+sqVa5KXUXCR16km70Zqvv8SeMCxn\nxrS941Q9Sf3cmzSat1u4FAccnAPuroEraq4mNl6V84tw2Wo83FPlvSppqlFHGzFc\noNB5nHLAot7I7gHDv/Red5d8YxelbpwkJHVg0ufcoeRVnGSklKevzhq/rj3CdeuG\nI+YVVm+Onr61BQuawfLO45f88gpDnOXZwVJqorUTvMTOHRrEB7AfS2WuTPpEGMBy\nNPO7sFufvuhs1cDMH9TPLpC9FY01GqaJVGTkTNRpxgWWfDjSwXixnTOZWljnsJFJ\nCBBEirDqxAdd1Ed7YWj9PdcTHTiErF2G0Y8wgIcfWprcWPzSoI4dUOwR5Mp+KPSJ\nN+FdkPAo16l/emudWq+QY3P18i/rmypxkKnaPWguQZoHDoJnLiby4h2wr3IwMLhF\ncb189gMVPFkZYe00d6fkFzGHSUau7CpufMfW6UMWNiiqPbNVcpDGUw81axzLbOgG\nxdk1pCOtbi4PyppBCCurNvt+jk0kJN/EQBdTJUod/egC8JzAE0aFh38wJA7B+ALJ\nvFTCMi/iHj8nHa1ygesRiRkDdRA0Dogm1idY618WYBeSU7u7B/udBOW0qRgBSVmA\n4Lnq6LiGB4XXzD2+8mNM3F1lAED2kdfJcKC+U2U/ZMnN/mAlIluMErp4T3UDBjrB\nkpO4bOqa8lysZQqBwx7g6tnAa/fxBejP7hXL3ayU3C41R3DTqXahXyiz9uvhQG5J\nWVy/qsw6TtdKzCNcIQH7sbnp7309f4A1cActCzTaoFcw+lrs1HOgk3Ul8z2O3NI1\nNdsB7K9xszE+e7qrFz7SCY5xH9jR1oR2nplzNWYlY8ts8xYZ2jeSp8n1gPKTJ7vr\nYuwgfu0wGIhjWvipvEKYroBdDfA8YejiLY0N94KgBEdUi23uDq2eEZD2X37FEP1w\nrPW0gXmE8bKkrxIyirkuLihcpjMNpchi37yr9pfo3KAKr8cejnAIv6EKwp/DCB2d\n5/uA11QNl775wzxD1tDV3VCR5Fwno/BO2+PyJBlmkgtrWf9lwut/nZML0lgSX0Ag\nFgjLExNFifFxOfFdQL8rhnIAKNuMzQ/iViPKVPT+VgjD1FV0SNsoatwSHK/klof+\nC/hL6T4BgnsojTw2mFRCH8UNEfM9+BvTvwhFYeQP0lcXbq62GMlbFcZz7YrXq4AS\nbq1AIuctPEysZpFCqzdKV8xDKYBom79Phtn8d0Z9dVAuAaWqhZ/5tRSGunhdK9AJ\nEP02ngseQs+QfLGVAseAMUGuh7a40unL2g4Rk1kJIItMlDfizI+62j2tvyuajDrb\nXMrB/PwVWGdzguGrMk01j0Q8nsagv1aAreNP++q7Ex/JRSywYGZMirQuDC3uQ1eQ\nQVOQtd+ozhzjtI+8RUN9vEnPt5ng3dLi7M03to1lt7D7SEfpzEPHwIMjIJtEYxoB\njYv1nPea+27EvgXuoRiW8GRqnvxPgUp5CLWBJrgTuue6ocZWhm5ij8U610FRrEr/\nd9U=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-datavault-2' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIIUOMhnLS7SYCAggA\nMBQGCCqGSIb3DQMHBAgS/YhSuB2ENwSCBMhrVA0EwpdhlxZ1xSvMdjvj09uKDMej\n00ZFAiOgAg2SX9Tq979RJFUcxkInSeunWDfNbuwS3Ioa/smmhLHPnHPozP3OzYlQ\nYy96y4VBWXSHlBvON2pcHLHg+FIvn3Xn9GKpNXmuIM/PWw3xikvTl51GeN0bt39F\n9Jyx6DIjVx+oP6kPTuS/MyoaxplVt2WSQw6QdF+FYsBw3JYdDTJKFagbUiudYGyQ\nLOR8xOGFVdlshQsa6BNi0Kc/M2LhwGp3KvfECDzPKo8qJXeH096NxwKJy3NcIcoE\ng6hUHpCsJenggsC8cp9roe9zj5pKcVFGgUSWeups+U2U2tZyqPKFaA4QC8rdwwCV\nZs3kSuYd68cRArMzX5tr9jXurH4rpDK4xYQIF+MWDLtbttRY0loVLKTHMEn9gHxC\nNI0P1izK+ZKaI5N0XDVG8aZFNefY6XwJlbbhnJx7Z/IfgXh10PqBTVvmtCRgh+m9\nMbrOq07JxrOjmgBU8rSyVcqoz2N32xBcgSoHkZeE5d3wQwOpL4YRtxZEijq6QJw1\ncEXUqYHRKlPE6s0eoKRRCxqpUNxvICsZjQ0PbOOGxpimuZ7SPiTOFSfShTA1G+J0\nEeLHGGv/ec5AMauD7siepiCORVUggvhwbdyNXD2FU5s1SPfleN4JoE0taajCSy6k\n7fNI0qyWUoyguSvJMTMUyb+meMMZy0l3O10I55KA/cE9q/y7xCQz9aein1gNDFRG\nr/MNgmcLQZWYEZF/bl3Qe16jFi6TLJZPtvzwIEoTXVG0G8eHkj2kp0cC/q5zoWVA\noFqDN1ivYG1a83PBFRgdUWdwMzPCuDvq8us7gwcG91rW+TPCWdpCpRFGoV6cjOwY\npLoNfGpmCqPVtn8qT6ramzGxToJlM7DhAn/hoXk+83fVZjTbCral6u5avlpi2moQ\nxtxV15uq07Ayix0d0R/Wug7UWpslqJZsEYkzogJW1UhGm7djAn4uhQ+Li5FxCilo\nxxaTKZzlXqdNZLlwofFnIG83ycovFfoWcLjcI+cz1qvg0h1Oqns2uGounm4mi8XH\niSi/QFhHZl0k4wYeQrbs0AJPGrS2lE7O7GR9Uo/0Dixx/CpqPDmucVKVoBuV1ZiI\n2kGWeJIqLNIHfpiAI6BOqqiZZWV6cRvSmKvtiO+FG90x7YDa/zzfWTH2NnaZzEB2\n9VYX94kA/ZfY3TYN21Ary53r7fNuq9jJbRAN0Ly/6hYWKSA00IfZE3Rt1za8G5dE\nwL24z6/IKEJMGlb1ZatVNMWrzEjrKs1vXBZJdg9sypblSaNg2aSUQJtVF/sEqfky\nWWam2Abn0TAiCk0C/0P1qqXol5wYtyPqnqzqHnwbnMDCadHQ2o/ulJC7szvjH7kR\nahwaaDj2D8N5f4kdB9aBuzGppH2SVZ/g2xyXlwhjS66F5RFVZAVdf95EH1Brb5l+\ntyShYtYifUIgMC9lSp8hycNnMzmW23vnoCG4s+7PgzRh/4+HNhBWiQXnWQ5HDuEP\n52ygeq3LSkqw3vdZfba2/Hjzc6CjkjOOd5+B1g0XZ5qbXziFklah3GFqnSPnLEVd\n1Ors9/v3WB55AsRKyRjsti5OfeyW92p1pCQ7IPnGp4hcLXr8P6RbbzQbixIjewkI\nU4w=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-datavault-3' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQINs7ySxKd4WcCAggA\nMBQGCCqGSIb3DQMHBAgghrPj87IjqgSCBMge96GYyJ4XSKeNkmgiZJgqSinRkD2N\nfy4ojy0tZa2rmdtNd4sC0JKNWBvMedOAhUPauti64y6+6hI0TcJq5XNr7NOEN6mK\njq4e48Kjj6b7Raqs5xixVa7tQBDZ9R9jVGy37QgF5Gy3WAz7Py05mT8UBQhvM3a7\nirrt2NVAOX3A6KcbFxcGrXNu5ePEDJd5lA5hK/7DOw3Mq/uxOml0/Xs+nDoThwIN\ntF4Y0gg7FRe9tHrIGm7vOF5Jv0Gs4bRpnf8v8Navcp4w6SFVf7HiQL2oM+pQfyKm\nBmRsUXR3AIYFRM/utr1J+lNWC12pmFFLrt8DwYqmjGF3a3K/nvFnj3b84lgoK8vh\nQVyFvEiILlQippQXRycnrECdHYWncbWcDPd4Zhyiz4mAuYwtJTHIzpIfstjv+EuS\nfOPhlS3P9kSfnMJx/WzOLAufb+wVFQFY7hGYb2mY02y1obtzauQsLtGERL9PfNUC\n6Ls4kWUhYfmfw7RyAuMKWkklIXoAnlW1afO59qEN2LJzy3/03vXYsmFcuL+NgX4e\n/AzkTN7di263Ai8ywFuXdE5j1xBXSgPE04r5k9DyL9lZN3hB0gHzyl89Zf9LcFRb\nb7+cu9zx7t0JN2MRuKDx6XuaciDktvvPnpB3igyWxBYX9fC4NUDRk4M1lY+J4ueG\n2qs1jO5DgIFiFCRIltvKsQEHkeTIMPiBVgcDeD8/hRUf8igLZ0QTbhhonqT7fRHI\nF9pdDbhXGCbiyt/xpY+6MPpyZqb21tNVSIU3KCt9pta2PB2qLko6IlMdBaZlsGJR\nzT23HtzSSeL9oh7bx+w+i0x9mAO/Olwe2ndSb8b/y9QIB93H9g9xgzwnGnk85w3N\nRFwvrE4+gH4WlGs70SX33SfIFm1mhKLd/LocFzVTLmAzBDK0ymAuQ2UqYKd0g2PA\nx6etj7XbTiORaukj85hpwTfsZAC7doUBDplpq3semk+lvadrCpREbmz5ymAQuDAH\n9ALkCj0aASMnRZDZpKN6fitUSFZtDDEpLQ4YZnB6ot8d0OBMej8UsHpJCk8euERu\nAmS1PxgfOyoIs3d6DV6fdDlonF3cYwYtKk/OMxnRZ+RDqFp9+C/uJmVQ1oJTvsHL\nMOzzVbkgqVh46Tn3yk9cF8+BOQvpu5D3GI5QOQt7cQglr0FpxA3bXjLP37VSVlrb\nOZGChcRnriyElDQ+L8PXY7/2vwaBH1ShprIXvQwj2Zc9ejz/Qrx/272WsiIUfyUg\nfa9wmREwF1dV/AmTqRDyhEaX+3LTCb2M4/sBYiGr75VdWFqpH1UtnKis4t3kiut9\n1/0N2AYD8Bs1o1pxUNl3qxYRPQ/xWPQlADFijY2nvrHVhMpwZeBa9eOoipE1B+ZV\nKGmYXMuocqxgw48DVQzRyO/ptaeClkIAALi0d8PWCXUewfh2K19PmUVjEPaH0ERI\nmvXKlGN89GBXlimJg4+3lUvKzYAVwpH/QL2OgBNBzkpSFvXQetfkB5LjhdJf/BBn\niaqEvC2auVyUFLqPmIUVN9iin3OIDB2z88D01JzdWi5GobXkV3T21biwkRhiYCPa\nLjjmc/u6zkoVaAccSRbrAIJYcCOADjSkM5RqVh1swzhzkJ1UHOObCgmc79uU4vde\noms=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-datavault-4' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIZnQE3ZHFwsoCAggA\nMBQGCCqGSIb3DQMHBAjUTE8rSTnx8QSCBMgeIAtLypS5zZ0zaC9RbRmWQvN2jsuc\nw4lsWfklrlUO113JK/9Rn/BcQzE+h/u18lRmxP7luSlcyIWwNRF9wdQrv/Vs4OXK\n0dWz14bYj1npHhYiDxYZ6sfFq8oSeiuNwmwo/IKiT0xHT/UKStZwwVvFWPFjor2K\nxdMVGSi/eEBtatz8tbjAjb1xHfT8EL4bymgEGTsepduuBNsmw+JynMOCFh31GiPL\nnw1m+97rGDUY+ZqVVF38ZpFelWWGIPJr6an4X5V09IWmj0THO+15lnjXd6fUXlhh\nWKXjzCeXFNEoVa/6CL35T6J1fTml+AlTRcYMDh/TWDUdzGVA4NODOZDeQq3ymfuV\nP9IqXiSqFuTqY/uVn7qNuswOIaCTKWCcafiBzYCcZfSWceUKRFV7hkXV5/yfDl7/\ncGzJI5phQbudJaD+FpVUz/qlZs9wu4SLdvjPGwqiMr6eyhFFUnw1AHkg6tFRChJy\nXRde73jl/HCwtTGBUhmhKiF7pgRHpVOGNVgzZ6QWp3OiVlQU2eIHEHWsCUY0/qIG\niHCPVVdCJCWkSLRW8cn4Ei6XL3h2hTWsA7CJXdjUQLTcKt6azHQVqN6DBMCZzv0T\n1AzsZdI/JfIMyoLa34kAe01N6KSI1EBYQeM0fTIoCVI8IQASl9qI4CroDgPB6n3e\nZOMf81gFWs2yoYkwmEOHb6MqgITl+O908WxDd44mIzf1WjjmIjqqoVTSf8KWAk3y\nCl7h5Y21v4IyBvf7Plp1kWFUxr+voLnX5cnGWDcK55FydLjiOTTSKXCMLr206Lrm\nc6re2xtVSUx0mm9AHPxrjtWMtaG6PHY3OFYoxd3MlaRjWF/hQLY0TktrrJm5oztc\niDSSoHV7OF2M4wHdhTPDQbFL/PgFst3/YukLsJK4ZHdGbJb9ff9XVJoRcVhm4BcJ\n2em/lAgBrUAdfHSCCPMr7xuFZqrO4xoCBNNw7OdWGx2f1A8timbT/in4wwwJnONq\nciJkLNdvarIY5J5NwvIvLzlGgBYWqs9W9RVJP8uUHtZ8WIYYFiOa0HFFcoCkT2+W\nOyPremf9pSJxzpLN4fcFnODxYj4ytbMRsaJG3cC2PGtTX9B9tJy60YHJ/9tlJuir\nq04Xnw7/8QMeeykmqrV9oAmbIHEzKRQK4y3yANi0FFjAy6AfnbkzVjymxLq0+EBx\nszIVEW+kX4HOb9vGM54RzgOqmAttxJu7EEK1nZh28CM7vTsm1PXiVoezcYIhnxYQ\n9qnweEUkeOO9IJPEvXKsW83hho7obJDVxulZ1c7ZpxbMY2KEF59osRCse3XomY3a\nxWcB40PEZS70fRb/V9RN2ijs9tEjNXaebwgOKikD2zFhvPaLddjyrsQrCLit8fux\nxcHO3sJfWJgKSTDscnyiy/tvGYMOYFYnYGNVQ73oLugQvCTUXJuKdpgpjByrImXl\nbyH7SMbWUVTvblQg/vsjDu6SV9BhM6doejWwQ5mGNmGIfQgNLMBAzhCiukF3mCo1\nekiOZ0DYR01NbJLcOjJtjN8CUI/HE5t+KtyZkj09/wpb9eEIrWOzbnBxseJ3ipuO\nh+NoVIgMIu7beW9YbNlwS7BzS5Jnv4qc2Xkk683vLqm0X6XyEY/SjmycTgHUnOgw\nOdM=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-don-2' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIdJyifEO6oN4CAggA\nMBQGCCqGSIb3DQMHBAijR2pLlg67NASCBMiPl5KObcz8/68Lrpe6eTwCHID2VgDU\nSIp3xbvwU2CYJb9S5Ck89MvW0YqorRhDJVnobmtm29RfyNqZTVQD70vctqys9gtn\nHZ5pC7SNa/KMcPRqfqe05g9y0uUc4gxwKc04ArVdkt2YS6q9yyXuNzsXN8zzwPje\nskklqqjajhdL3WgkdD9IMliruHKG89R67XmhccToLXLMyPtb+oBkCWAh2QYy/vVr\n4LRolz6CY1+25aAGJjvCs9MfY5aksca+Ji9/rE6cYcbqVaw5AWDvMmE6Wpmhkjf3\nryqJW6ZMEb2FvT4/h1djpsyoely5bRIuDxlOHsJUNyIr9QcGJEuDnQNC4zwKDonP\nUKeOwwf3+bRwYeQ8b5KGutGchkITLmyHd3uatNeGSNMF4gZBYRmMfXOJEypUYYvc\nwL6ORwMDeasCJQ3ljNKatqcmvfRjJeHLyLns23xxIn7dYY9TQ+jjeuBJCIYTyJQh\nJ8V3BmKicVwyNjzvwoOjZg1IT/RM0Wd+jwAmQ49UE57lfqw/wP/Y6Z8GoXskRoTe\nsTS/q4H1QPYpIOYigzrMiGCpNjj1q2KYQ2Yg/U18Q1X+A/OEtmyqRct/esuaRAXJ\nl42CtFDlW8A78TLxkmismvwrBzpBV8KKhzdCLogGeqg7S6/vjHPIbdcSCeOvXUrs\nOjrnltAXsAOaZdNoOqZNp3zFumD+OUSqdqzwdydQ1MbJYBZVRbE6N2w9qgcYGMWQ\nRNxvZnwIIruf9fybylBR2/uuSnS1BK5YJ9yeZ2kfNbuG12X5MoidEKr6gJdBVHdi\nX99rv+5/e8fwouHGi8tnylFTycZgrHdFAtIoLdiusZ3/ItPgn9SkdcQAyCnw3uzD\nMfQ5QpRoVqTc7KDlHa2Ms5lOA7u9QwB22mXZglRle7e5qs2anD1hXis82+NvF1mi\nDeUq6zmbMJlbYjZQ9tlCJnrZBVcdS7Cqp4R2B3t72IrlUB6Axhdb/Ut0KNsNDv4B\nfbyve7SRNDoEuMSGlgirqRJAQ3dcawVBrrZlPX2YpcG5lg1JV4AsxjeUt+5D2l7w\nIxCq6Ro2EObsRrqXPirp9xu0H+64obmoH5p89Om2kBBp3jSDRqXacJ++azAgtBFB\ngC+67xlAYMK5LQ6bUJRd3spCWmWfT/8cVKqaNe04/EnYo2ZsGsrX7HM3MOvN68sD\nG3u5wX99z9nfMZ8OcJZPNM01LPBhBCRbKJog100bGsOn3TrcTTvIHHofog2OIeHt\n+Wc0c8UGgPb2TKFgdLyJcLbUSJMhaDRJLQmBeX7hzfKfbZA7baYL96jUR1/yuRQj\nwrpRzS1bKzeiQ5g0hVYbQcodrd1amK52fOe1LFA8QOH2K/sxQw0n1rS0xeDM3B+5\nhEQW5smjDbIODmHQEougwjDWKNbrXs5gRPFOy4Wa8uF4spBHvB4kNIcU0nbW+X9X\n8Qo+MYiYJtq9aVANgidM5NXitTfMmtkAIBOD6LHdPE96btBzVcbnlohJ0hW4CaSH\nQcD36WEwBEvPBO8wOJkHepfFC0QCQgCLHbMwEkHcKh9CT8Ibteus9vKSmdrxyKxK\n+9BufBaRmxYpCQUyjv87+DB/h2VIkcJiPKogjnCmiEpA7weFBlHpnyAkMWkG07j/\nQds=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-root-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIvooKywaWn8cCAggA\nMBQGCCqGSIb3DQMHBAhlRcWGrJk/5ASCBMjIQ4yp2ceLYJPJY1e6h/5TFki3aLPt\n4MdGULTnEhoX+HEEbfDBELmRQawZ8TjDBjsU+FlxEKnV3av+9Z4lcMrXNfehcSN3\nI8zFSfj/SKYUeqC8nN0q/BdUHHaBDsyqdwSKe27r6KgzwRPn6DgFUQOvAnmAgTXx\nORbvdyUn3QRaFx7imeDelSd+t3APmmcsneePscYCdGcKpG+itAvH8aAyAaA17TX/\npIDek91yRHzqzoC6cmB5MWHDqQ2kys3xDaeBzFL/kYG123grUybeL71mzfPB82cH\nHnpke2xpShdvmOZu+/bkVo/HVUVi4XpEflhmtjkceKo/qDuP1gHzLRUAmE49MMob\nteS3PAmeYNSvS3zZ3xopkYeD20WhOhRl2XPKzd8ZvpX3PpuA89D6lvSVdApdp1Zc\nfGmu3MsCNGKvEDI8GZB+/DDwPB84f0xZZ74MmZaC23Buik6ymdfaq1wftI8PDzMc\nKjn5GLb10KLOo4WriAO4Bk7jKIVakmyfEQZLbZZH+0rBuf+6P/YrMEDeBiyDPnP7\naIWWUoXIN8OEslBGe5oErqoPkwjh6Zrz78DljlLLxnAyugikaH/VneG3aRvuIiQd\nPa2SNE6nC0FRecuoc82NoXeptdiseLMD7OijzChWzQNDSw5YMi+MApYqcgJTJjpJ\nVCvij4WZjZcUYc6qTVEhjoBbXKA1YZ4xq/RIv0ZeIFL+PN0o6Nx2UENIcatGUxla\nQ592kW624YqucBOBZDXAtfgpPmBS61KcO5H0hIFb0kAqt9R5DeeSmVmeqRM+NBFh\nPzItGQ9O2OsGz0shGm6BfP12WfZWmZgYPeCpZSZQvPkgOCBoaG9q4YcZnnNl3gnb\nv3EbC8B6lEHM/eg1GJhlKU1+WK7NvxoivlIPVsvUBeL7UDkBvAwcw9C2Cz4SyEQc\n7PjrZEhpPDQPICbteNtqFEprZ4aFh1g/oLZCIH98Iq9aNxdgSOXwYUaalPTf9+89\nZvtbHHz1HpEQ/aVfvGdjTJrqxO1RtvunurZO20MYPyqVlknZCq1mYrxxLhG76WbW\nS4vF43dM9+swT6vP522Noi9LhfYnO42yFD1Vt+OS6XATKGMTIaOlv/V3kFsvURSx\nDkQ/3WM2T0TpEzfg+NuzyUzZg6IbiI4OC3iGOpQwquFY3RH7CcIntZLWTbKxZz1J\nywlgZV2n4xG1RCypyEVSsL5RLcDGWD2UzEo2A+sH07e275R0UFxfe+PociKQMfxD\n9o+d16mfsDr4l74157iGt1D3smQFj78FUmqkSO2pZocV97E8/stZJ5gM7oIda8Re\ng55ZMG/LjBMDW2dZ7pCxC3LtjMz5aF3LYq777Ov1uwW+/p6KF+ekBEZvl48kUhbI\nT4QaKCCCoZx1tdQEHR5LAlqeiuNWCZ7MpV686q9x9gq/zza28NbNvf+rRUDCtqRF\n4rnYieF76JAkjphkwkyR4lOvGatFZPClGl/vC/ey7Z2MwfWmujsdWDj9veNGQlfq\n9jEdtUiE6l2Rc4zwuoWGqPhRC1D1wT2HJf9P2nsWVCcCOWtdq4snJgkeFWYQTEZi\nDobOlAjWL5dbNU2Q+VrrcHYQwKRfAEexQ8Da5mEkcQ01Nc85doXPsH9fftE1b0DN\ngpI=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-root-2' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQILDs3UGPDpvECAggA\nMBQGCCqGSIb3DQMHBAjlrFaJPdCWjwSCBMjRRRlAvrgb0d51DMTarcLbplcueStH\nvfVMzIUngCMpv9xBXHf2a8QrM2BFTT2o4FXBrUSuyxbgjL0fNQKpxh1eGqQ1CckM\n1VxD0ruiCTV1i/I0p7GkWiObbydBB2HHgBiR/9OCiYvq+EuClxQexZ6E8l8QbEPf\ni68eY0LIzZ+sUfdkxW6Q/At4LQ0zSJCNv7TLKnHlh+j78fGSKstG/QdopG2zeSnK\neaPRbxeyqc+7n1NuSWoqd4EJFNAu2LyWrRY7kuQS5ZrNxDqmoEhC4dCE2PsrJl5J\nYIcUBFktcTfpEr8UxAL1ad2/0rYms3tNBqgkm00+yetFMwSRJkRDb9ZHIoTEL9IF\nl1A8FN5gIHRXhbuc+OTkV5Lhgj5Zb08QYg2Rryp6VRh16B7YVvfNBG3uQKqX8nOr\nb+Mi9MGJFAY2CIzgjjD82AJO7pXrck3G47LbiTD0bRI+ApQN7/6c1piEh0PMMwpJ\nf8i/OkW2Sh69qy3iIm3vWTydlVlL63OH5bH7v1RFw43Cytavto6nTmz7gVi4IR+T\ncr+dWK5lB4FLYIuUqs2V79qaNo7F3QB+ERpDaoQO9ip8TPpn9y8I6N9UBMH/ghZ9\nQKOugfh4gEpaZEWigqDLPGZo+/stQx4Ppb0acUAJeYHWkkECMoTUD/puecnkHhSr\ntRFVgjjF1eWpqk4+BBFqTRbbidZMk3fp94TPC8uGjEWpD5mi4tu8J3AHe92FDuNU\nwfEpChd9uKLIGmfSitvrKt+Jh6BqVdmPhCWrxooxsq20cSxMU7i7SgjzBXO4G/Tz\nXrxhLAdU5SjcqwHcI7PvfGCSpOycwXf0D9vQgLPEh86GhWp3/oQxfF0vCubE3Koi\nDmBRZIZfKK3aXviqH3ahw+WHu17RBW5dvySoeA5t+3EH8AJKEG6XrpVohNtR0ZcV\nfSJF9mdEfEbaM2gNWGpiFcX6N8Gyn4ncr5uKqWVKjYneMQjel6IEp53j2aCQLuPg\nGoKXqaRC5lU2V8YFcO+KQ/sC5rTmxg+cUYd8QOZS8MWalnZ6TSAL85R7eA7hk5TQ\nm2BzlxdrNLPk30nOyvzS9UKnYBBqQRPbWrgBBQGcsA5QsELg6jnU72TNMb04Nkr3\nIO55dSWVWx6PZA2wNWyCDkX3dC0gSYrA201c11OFZrTTDAoQX9MBG5L+3SdjndK1\n2Nr+5cMTN8qFppnuHg4vlOPguG8Z95qQHOMT2LLAhRrOC/+s/RmugY5RsMECThQi\naOQL6KtZMaglrNHFbNoDiPnh0HSTB9X/loMkrmfO/nEN04sbqo0c/fs5LlUjcyRX\nTMON+W9cOyRyYsIUM4uOiNSqOu0qMBzfPqyirwsExXtSBuqNs0FDE8rhKZAKh8qQ\nwwJxZsnYjzh3kTHFTtdOWy2TGaSwRUZ1Z89tymYHh2bb4ypPlu3sor3BFiTm0Htk\ninDX7XOjAGP+HScfDEwJHjBe6vlZd4n1fAy8fwVA0KanTOhpnxoO/He7y9LjOQPY\nnNOwq/mL/nflKWZsOJij7ekRql3OdNtNzF5OuKYA1/xzSrzDStF7+lJBgGCRPbCb\nMLNv7Flp3Y5nllMH/rIDsy0zqXzvsniy7TR1NJ+fylsn1qqWAvrGs9dql5f1nSbR\nsWU=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-root-3' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIgR3dgHkfcvQCAggA\nMBQGCCqGSIb3DQMHBAiw78nd6XvwKQSCBMjTj5lwbo22pfbZJfy0SozWofzWdPni\nhJdU0LCdUf9nGC6ix05FxDpRPFCzzTXi/Ay1Da5lvx0CBWDb/lylPTpUZk/IT2wR\nxMLdQ3euZoUHFZ50T1SyzQlwZb3Nflb7pOFZ1+a5N+ZsSwwnKgYc+GkzZkl+7Gzu\nUMY4Fx+AQWw2kQkZJL3p2cy9ld8CnkR7BT6uYrXSCb4xo+JiaBljrboTW1Ea1A5q\nRrDSpioAbzCQKa/sezGS/z/WZAqMAXCy50MI3ikmrdRvet7ASc5oZQ4zUncMBv7z\nMMnTodu/S1opmm/AeQumvngbQWEhiQD4SMYxVYjW1oqWwhfuwMXwle7/JycS7QSt\nQqGFzwgIpQZF55d/3nbftJoUenYwi+VF1xHCk2yP8s/Bni7EtSwHtdcdJcx5DgRb\nF4ssmruRwd9HSnFFVgwxc/XFYUMYBMn3vpxJz0xXmgP9thsb8cLJ44m1HLNsHN7Y\n5vllmakYfcZ/AZ8t1w9XZHy5jlNjyHSJZEy8wepo1RMikXe8CVfDA7mOcZxF/oGU\nRUR4PgVUA6ixwOrSxQuPW19PTFXVOHq0PH+K2c0MZngEMF2I0lcsY/iBBC/IQWnM\nrrX5AtnXUs/dBV6EQ7VLji9BKWNHF2+WcRDyXkZK5UeZoQ5hQAnWXSM/1xPH/vQI\nFVGO/9X+68y1kg2QAVdYfxUgArRCA0hDftZBtZhKE/OYkZEVjOPsM20w3ghw3X4E\nZf9Wexn+e40bXO2XQokqMJAs1KGiOShi8myyITx96qbkYbMaplkhsbCTdYIdwePp\ntGLNAUAM3BT6mH0ohOGRDALfkDfdw5stQelr3pwSKkR+lxuaSzI3bMeaIZ6PcvuK\nQejtCIwZgCqQN6Wkb5DGey6Pewztt2DLClIrHXMiwIAcT9PjbhJRc+5laBYxaSFu\nSIsnsw5bipi8eO2IRxTEzu5ZiMdLV6/aUDFmXDuKX/6Vv2erhIQimJ/ydODkjtTN\nnR58LgNMEzWT0FXoK4QVy2wCUIWJSiQkSwaRGuC6wacbx2fJMbotdLhdzfKJ774J\no5UvfEO1QBC0dFP8ynr1vnU6wM6vFAwYViZhArvwt0dyDC/DYCa9n3U8iFl1k4dY\naj2jic2zQBlxmcTq6kyQjL84U/E0tWjYNfz7Doi1f1LDbuBoH1FHMnoR7Pew9LcG\n+8KSprfZ6s44igauq/ZODf2Fe5mbkDqH4Fp2Q43sCxWKu+NEzldH9Q6xiaoxD+Yp\npwt/dyQ3MonMA8CqWPGaSp1F1rXO+vuJf13Y2n3vc6ReVrE4VOwU1gWE22xX4X8Y\njNaXHkecaCUkW7GKPQu+p8IeBZ4uocDUXbdodQ8ltb+0lptTIJelp3lAth2HJK0M\nuHnYwlJpI8MWHSDUepxzh7fSUeZ281Wxm92ZE0CMefubxs4cl0vSXgJ/EHIwaXk8\nUrwmCCscw457U+fk/1KCytbB+XczbRpvjSwUJAWcZ6I86nIWvRixhAnw8ECg1z3E\nQ2tG4KxGWKVQTDyOA/Ziq9E7DpYXsLXx6ze2epJ7OMMwzXNZedv03pHdQjuvBzqQ\nivGmijWDFbBtndXupk5G0zAzZKpMFuYr1qfzK/KLKi5EQxnQSDjNsgtVGFZKPOwi\n3Cc=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-root-4' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQI9hGHKPi9oAwCAggA\nMBQGCCqGSIb3DQMHBAgBeZZOI7TIsASCBMg2jGFarB+mljyUJG4N3blb5AJWnVxW\n90awiLTdQ0iW9qJxvG80pRybowHufmiIi0SRqYwjir+yNSL+LOeWa3teShKQQL5G\nGp6wI2WKFBcIuxfRQOx026FtWp1qxn7KCU+iFRHOYkI18wX83J112J5Iz4UbBNrD\npQ8+ugT3WznvYPz5NuAh8ARyK5PlmB4jTPY0cv2cXYD6g5x3E8D9LZvaHts1nKbq\nr4YaF+UKOUmrPSXb1GXBsafdRbf+maQKCWz6i6j5T02wX+MwIB6+qAgojewXKAuy\nTfiGhOjBevr3ARv5ilJzVAb+/px+gYc1peEzuV/T6s/D29NxhnWOlCZmH7t9QtOm\ntbnvjJWGdAsur7g5rUS9jpw0mjwho8Dq1sm02Ed2FdHpsTGLjAVOcYtorfzrBE9z\nf88yBO4YRkHMnJ727uqu9jsbkYog/nlX3xkKiHqw9uDe5oGqU8ZZdspFI3sKm9UY\n7seu0+YZzcr0IX38jccLSbBMdQ0p9LyDgMsZOqUF3GxkAQGpq/zjXhrnS+xCEW31\njfR7wn4/tmGWQ/ooQY+pNttkczPtSSxEeYG8F6TsOulExRBE+tqisqlX+eUP3TVb\neYBlEo/mkdLVLEu+6tWdL1diJK8Igy2AAHjfEKXucsrBFqw4POKoiJIkjoClsDTI\n6QVTGvX4upGsv1IsnoshA7nk7bp9mf1b1ifEgnYKv96LcEG1zq91mIk33nPolIY9\nTT56Lv+LC1TE8BMjjfuQK2aFfz/uxxLVJf6KLXpiP4PdBvAofjEavF6Ki1jMbFvz\nmuolna9/VRPyoGwONt3Sh9ENE9agDCJlD9Gt+ea0TWkcZjiZ+RNCRKBDAdh6u8qQ\nwXPQx8Xmyxxl6QDOo/3Fy4/5emPqUzHH7FwJUqhNLeosHWMaVgIp6QTe2CTt1pll\nzZ/c77j18OqVHWbVcjPKs/9EuVw72D/w/EAvV0u1MOWfIufzUy5cRMZ5SETOoZtb\nh9aUSAbGwWxh1Z3tIqvqSwwfEJx1DyaYZpFL6Ja4WdY/du4YqygUNamLgtfronCt\n4t3l1vGwkUWSdpYBaW+kNIrJJ5Aqo/5612IQmKVu+BXxbwSOrv7NfQxwhpgUurvO\ntEeNAs/A57jhuU2ZgYBURiwyMtfnLwporyf4yskmfhHeeGF0o6xSRzPxfG/8eRIT\n0CWpEbtsrL50LXBsjdRfSFz9krtsCd4ZjBmDaELEPxjrCEyMC6rwuX22bagNHt0K\n5FQodCqoMrl8HXqG83JUQVgBaSYpGS66qoWk/+WtQhHCDj/4s0o0axWhkn1KEGaP\nc9ScBRyJVe7ri3AIdBKxs5sDDR7DelOQVKj5G6uTTpgySyJbl2g7MgOPi40Xa57M\nceIDOL+e+oW7hyisoG+AimYS2NGSoq9OXl14G4c6JRjrg9oySWi/0CpU2g6KdJpb\nRCLsDdDZsNqU/wTLCCwrK4lIRaCI+o7c3531bx8pu9LzwaVxe2IXJQyay/OD6Hkb\n+IG8PWdLI90sDgGcDoLSce7nVK8hujQKUlLkRndRsUJLbzdr3mFEQd6CYbJ9daN6\nIwJcIUY6A3ee9+yiiyBDu+gHCrYSEyNAE+rPSYLit7r2XUJVtkaibp0qfBxpfmsK\nqPk=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-scep-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFBjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIOg7Ta8MdmLsCAggA\nMBQGCCqGSIb3DQMHBAgHzdAXN3tE+gSCBMCrfpNvrELZdBVUwjUoq0rdYStKO4l4\nd3LQWoV7d1+dvaR6B5Myk1+boFUMmcLftgAu7U9++GkXXp+t/HGwJ66f/FuL3bXl\nbl7nloweVDBmk0ioYnmYioDieHdO13VYzUxDk9yDvUhzGZ9EloQCTDv/PD1lckRm\n/HjbrtfFXOTqSYaCKPKow59407oSwMawn+nn4Nd9zoAdA4EomYvDJDF+rm/AQrs+\nKENlRu7z0QYNx8lHblbEnG3svsjanpcAxdWeHLhHoMDkh0rN5P9O6UbL/CosYsd+\no2hvi6zYa6lvUnjiNKBroZbYboAIZQywsPzubd9wDyncipw5C+ITTO/Vx/PRL/3Z\nVtmCHScmOSkGpAzsA1CTO+a+iuF+0CV4xGhf3UDebRoZfPPURbx2eMrD8UyhxyYu\nGDAnJr/Azpi4vzqEAbbsoQqzjEKwxWGSl6Dgt34ecksKYKx9YKfyrm1yeBw3bAUL\nMZtKqJqIrptSz+JDdnCndxId5cLlLyRuIFCuEcuA/PgoaIm1Ljnbjlr+rBX3ezGL\nIkVWUeXtxRCbIw7pV5XWXBk/9DfgqOyWJRJAZaA5e7Tk33XaZKuXCIr7tjO8dh/U\nBcywhp1rzrwyqSJq+hkXBXzIL2+8C8g/nhpGFNc2T/DeG2BBGPieAVTTmX2zolpp\npA9bah/Ocv+t7mcGwCKuAUMihrqL4dCbWAE+Q9Ib/dX1hCKhfDV27lW9Vn8MoaxX\nx6ClFbZkJqOjS9UjN+3Gmzy7qT7uBhFKgG9VuAj1FwIolBtT3lHwuMSipLFrm7iV\nJEUPQenI3knXlFLAv1igkXT3eMxGmrQ5GxSPyB8SouL89mF0iHuvfWQICXrAIeLw\noTgFMg6W780nO9VhcsfnXKIy2hHgOa7rGQAb4pTriec5MPEiKsey99elG3ewwznm\ncaQKi1W6dKWS3iLrNxIIRsdT+plMqxHBptfcNEZf7n9CO0gLdBCSZDy+OinEfVxC\n8PvyRrIE9xeceiLDtph46QFBoNavb70BcmN7FuTgo9U8bMSszrOPBwst2HGTz/NE\nTws2BFtj8oFKSVu8ElhXBJ+Tcwg2um5QZGVZEZBu+9CLtxjvv0m9jUfwK5sISYCe\nJUghrbjwF3zP5geApNriQ7QhOR64qux9EpYI2lA5MKQ7imv73DOHtPeal4Z9xxOY\nN6DWvSQWL0D4S/ufm43fD3aGLnBfHTQHwvQikOVoEKh4J7F6/lP7H+H1jlzFSKGn\n2ClypTCDpSFMs8nN2mbeavJb+8tQrV+YD1+ABKHOl87/9fij+DXAZM2AfDXTyafq\nworNgJ5mKgMxCEtX4Qct+UbsNISifNjHElUeH0yc97BZVrF4bnYjasKDHq1jC/R3\n9KBPA1E6Ef3OKMutjHpsKgSUivLl4mN8Kja6m4vAwc2WbjP/YPRAY+ejrpLTNCbl\nbKnhx3QiIfUq/ISvgYYyjAZSKuOYWirFuZjVg8na6tSA9KFJ7vbATBedIPpCyWwq\nFxxck9p4F7Ue24CbIbdxLzrodHSdQO0wdR4ZEHNenKs4Wvngw+LPtHe8mU3lw1w+\n90Dr2h5AzcGCRx8wyDG/voMfilZ6UqBYvri8Ax+eTY1QyCR6+VGxCPKr\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-scep-2' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIev+RaSw+asQCAggA\nMBQGCCqGSIb3DQMHBAi4bikOBOXyYwSCBMhZlw9b4uuLY8sFAJn1TfQFZ0xRJMOK\nsBWz+0SXfPirncHzVUe/vBCFXGEqYlfAyOG29+7ckuACXKEBV9JPwqFhzh1AxwlC\nMzzCabD8DCslLeMYMs8VpsyuU2cuk9KmrT5PkHbVrbmFl2GznfZ9GV41CN/XQt6m\nbqo3hx+Ui1L0pSnFdY4qC/KT6mibdh0Jx0KSfWUo/U4pQRcJfPTBET7YPr76R5Ky\nIXFIjrtMEgLxvXqn7S8Nwi++DM0geZODH28sGw9GdiwGNY2u58NVyr9rp+xAD/bG\njZUTA8MaQViMoxkyI0BwQAdcjIJiZWYYFfH9t/3XvC7ipcFYV18ASO3eGql8M9ru\n2Fw9WT2EJef6uRhux6HdmJHE0qw6d/A+3cbnt7IyleIAY7nz3v6MagNHvkC2CKRv\n1MXcU4B/zTIMOYK5M83E1zHuq7C3wlJ4x41sO6ckUMpSxo6NoV2GaiGOAJ1UJQun\nmiY5HFI7YzTiRnSd0xRpUEY2Lat8ZH2ZDePKDPhFFTbx63mPyv3P5fSgFZ1LssDv\nze/YSADtiBnFOXD1gGRSrNoY85aZjN4fDgTWY7jGa2cQz/hr61dAmS4J9jJ1lOAm\n5M1Bc3xDcysczkCMTB38sD3iHvnDxqcM2fsdsJhuERln9i4dIz162a9Vw593FJry\nra2g66gJUxJfuYa9yxuUOM0QDBkNXHIBi5yp5lIo1OqJLnzsE/5HXa0atlvpu7Bf\nG4Cm3Z3ulP2rr0KcWbfFj2n+nuDu5Oe1wy+0q0ATA4df1rOAfm6jEI3ZvhpJoWvK\nDkK4ddakeMPD49qzh/22LlCHyBBOuN6xiAYeK0tjn5qGWU42Iud9g69LBir5QKTb\nJek6XhI0eHA2V77ODq+W11zErRm4vpmGT66JUUDuPobboTl4TwhUr7Q1UHD521Q2\nzJ60KV9L4HbkZO7+OGJ6FQDXbF6IEQfTbgDD3RuDAXGrurRm+qux8usy91pYKQHf\nxah601pK6jbtXNb4aMNYW0d3uFd6lf6I2fAg2QI3ANo40FCqwKhJyCk4DW4XDxoE\nKgM2iNMLUIIxyINubLb0K2/9+b4JmtNOY0GzxBqiRiPN6K2thcw5uII2r2JHh9w5\n+Ky9/ev0Tp68uHW+4zeDfjU8saGZasPFV3gcgH0dF4bfZp+PYn8eQl9rOtZ0WbwU\nHY5KseMBKCO/TRWX0V185Bk/IANsJiuKVOkMcbcUa/v+4VTvxbBWd491mYjWDlDy\nIrFkVsg1ws+GjHlC60EwF8w4XKWzDnA0Wy58FXFa0VIZb2JOmVQ8qnN+YCe7OhXT\nKIr+fq8sMpbUWbyeXkrN2VKNjyNGCFqLkblSADVW6DI84nqtXdRu7iP/YnN9wCka\n+UnupOqd/YISdXmyOqD+8FCqSeagrcJl6oYHNM8+z0+CSkCRm2yKL2zUED0GneVT\n5Z1icIzqiM9eBzH9nx6LEhR4AV10zVjd/yZKIusbc4F62NVYMXehV9fVXrfpQVXF\n8TcIJVxdGqoISl5LLY2WU8sx8xiWguvy6bKu9ifaEwGt7FNg6pP2gechI+9ZI3RU\nD1RBv4QQ99DNCHRMTtqMPlpLl47FZO+qjHomhSyHvmtC0Zje1OzB60DXy8ftXmir\nC2A=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-scep-3' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIcBvBu3fsNJ0CAggA\nMBQGCCqGSIb3DQMHBAj46ULT55RwUwSCBMjnP2JN7uCEHcpmgdxb9+Kj4RraE1ZS\nycR+IT7VBEmE3rOjyys3zNMXnN/+hs0JC7dURG9yMCVr/tZzk5tbQiHWAR88VB/V\nn8Nf2WeO1epQFlfclvYhsBgr00BQ/LNvwo3ukUE1oTAYxMotKCJSFJZPLKLDjFcP\n1ZvcGaoFUtTWpIWrHZ6jXqVJtcF0wvklPbAdP89fnXfsA98y6m+ttZ1/fwcKiHFC\nD7rY9fj3nTH51OhFO3qUKZg4t5k+tPF1k+m8tZkMTUwF1WJmFEtyS7dzJdgu7Itt\nUGRoPaagRQpFN4tf1jR0xTYdbOMvcVuVVsX+CmdceRbj5/lU+4IU8zKvRMrR127l\nZ5n4M2gwX4Ey4y0/GCHTa4Zdh4HBhwH5P3GaDF+3jcZicDIUZEYTR6CwXn/Qp9bM\njNgBuFJ5FDtQXLRH9tQD75YAKrdCZeZo0mC5OUZOon8sHwbSkCM90mNYnq5ohyZ8\n5FEmer8LOKl64xRUHKoHrbTbvfvTUI0ldoCzuBWToN/1CA9PiEokcFxZ/yahiStK\n/uait7unV/BvnWH+Ii+QK021t46BexdORs+Iz2cnui1U9ClVSi3A7DjULlmOsVqJ\nA8C3Omsve6LHa5SNmC6Czfv9KDGAKIPwvHmGhj5AQjAuVOvpsp+b2w4/iRbqB7bA\n4BwQjEE9nFpPM6eH5WhpyD6vXAsFaYE1L9ElYwpwZ9SF5ravNf3fDUfG77+q/JQX\ndyDSATsoKM5mJK+3TDja8XPKQlroUUbPuSwWpNITluqL5yBZF9RVG50R8epcaaXl\nIuBw66rB2/Sf39xr9xEUHVq5WmNbHcIJvpqYMOJfsrcCettFGmEeT8J9w0+5Oxdn\nbEdiONHs5XI4V+FvEKyw4qRq+h3Z1AoWvWa+Z7Ny5yyWVRSVXMRGh+GypZAkOrzl\nHamHLVEHWrHwRu55Z5pcsfpkd/U7smCbEG1euHLGdFc+nepgFT1ikGpB4POpJOjq\nxCCvlmYHoD2FakxF4v7W3ntvON5c8wBQdOF+YIRM4Rv5h44bo4ejM654xAHDUrCO\nIyKTbshrwsh4ztKYH35TbGhu6tttfUPg6xS5oijpLhAR6taeumWLY/DpdTinurvY\nGKe9FNVvVwkJ95/jj4EoOSIo6+MxTZ+HsAfZ2vDV/HGe2c8ZWv0V6ph/fOxJz96+\n6GGlnR57KMYO9QIjwJ814FeImVQwYb7U1gKe2ypL6rWrqG+8bwCc2ISDzgeL7jW8\nSx8qNYV6drnrUjRaPBbWMBfH/UfRySvK3rJiV4vwaaemnPb7ILy+xDcHlUs+2e8X\nyMyLpb0iEfqCPdX9ayNn2vHy8eEcEuFaVuu8E5qs+JShMNFqsvajpGHEaf0f/R8L\nW47TzcJBqO13FMlBfPyaWjyEV4VOy5WfxBiS8EvLfkJrOYzBCUwgU9KeGDnDWkfx\nIlOYJzKHaKkmRJBpo3ckai6YTE7sZpqecCFRkbu8wRLk0ZLtFcfdBXkt1ySNpCym\nmApylAgC9xo2fEjckO3lf4xWUcP+FkdADnMbF5F2xeCp0HBO2jw0w5jclTG5AKy6\n92QqXSri3SIAmsxgINtcssq4UbOKo+r6DpSQ1sejRG2xNgd1RhUE2xaMNHubcK+I\n2eo=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-scep-4' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQI5k9+x1H0ZbwCAggA\nMBQGCCqGSIb3DQMHBAgG8qVbdhxFawSCBMgAgjz3mfaXv6MYQUt3fzGrwouT8CKE\npGGqK9S3qeQTCyw4L3O1fP4VNHlHTaNlfkkGzRMF7f5A1gxLa3HAXm6WGUz8La23\nlnpr+ArUFPt5dDkW32DpksefcF+oJgUnb4QROZ7TZjAxRXEfnmdF1n/VlaRi9UN/\nRHz8I8TrGAmYl18IKo2WcU7smsvgydkATleymgFRLqcPQNfZxg4RJIdXavw3OENo\nfqQBtYZ9NdPg7lbbhMY9YYh4riRi+VjP2NG1cfFgRsIaF/Nctlc+uSeBQG/WpT14\nvX665GsKD2VDQqm7jLeQVAj5kM7+bRGEPLA7GvqFnuDw9fukonGLvDfXXxN650jF\nj0l39JdrDL16aPKjiJZ1U+rojwrUvlnGRH2OPYOOxXRBZ3wZq/uCE4O/ZlQOtg19\nq7MaW4MLjJVPOCCLisGvSEmOraWQHKpcEO0KBWvv2IvfAjn0j4q7I8F7Sg/aUSni\nuMt6tSEVC2gO0pcKAQQRqr1ml96g7ijE1NBIiKe/sDHG1KywS8foZpXidBS1nU2/\nf1sYysBQ2pR92oH8e4xrj0HRM2SFwiVqmynst9lLWom4hL8MWLjqi/qlIBIbzjPl\nhg3PkyG5QynMS3ccf9itBUnpDq16Bk+alKRjapa1ncN3apcntAMntjcch1d1hLpY\nyQmrQ+csn78QH4dGMFnA26fklqT7zt/BKlqYeWB5gBv0bKL4zVl450PlZUFceUG6\nv+yjB4HyQDOuhNU5/lD9BxDvnX5pmwnZeSSQmJuFIibSlekQ2Dtq92qkZhoxABXy\nSecgucOFVvjRBDAr/AKnttUYsgbBfj2+wGHLMaQ06YUJuAQg5F45Rx0r8anloFqp\nCDqP0ApVBtmu00x77xaVhS8onXQJKA45EbybfhYXrXHMJAinMcEzFZeb66WgQc9H\nfPmy8ljF/Jp9vp+Dje4rdUq4Us4SsCG1eE1CDWADIlX1idTiMtMWszPi/QMI/DCt\nWy4jSPBPc4jXTS+TZeyiF6XltClzsfWgFOe7AgjSlBYKURFJqhxgVk1PUNKnzIc1\nNvOmBlD2ztgjQtgnxvr4tInu7jm8f3o3Mhn9PfYusgrLtmG+vc934PTbrsd76oXS\naLDj+/cnKQNKlzLeiXF079MxFuOdb+wGxtTBEERxiGszAJd6G9g9+SxtV2j2iqY5\nEeFclYRfW35KFuHQXHGerdfH6aVbR2y4i6Tp1q+YlcEweJUg1VeyY4hR5RN5zzqT\nqtUXGQW0opWW1U/VkD1XHUG1hjbQ9i3Wy+RxEZSLpoiYGvrmQi2EjMm3HGOgvXoG\n7eYcQ/4WFcBGV1YYfIaxAsn4SFTRQCyDM5w6tC9zGBq83taJZb/rSco4QxwSZqHO\nnckAocYYJ08tAM2+lX2JodYqDxnmPj/kQwu0oa1sP2QVbcpF5tzRZ3veEpnUNnA4\njm/rzQE0U6O/inM5jGUgNQTzG3rmDOoW6yFPv2/I0ZDGN37merGKnM/o02a6CMV0\nhmXdTF6tsMMHVGdR2jWNN30A8JoGKLd602R8HlzxiEQZuq3zn9pYQfNmmxzUonO1\nJSY/EEljaBzbE8wX9oIF8TfwGtEuq2SOpjxNHcSUsVoWpaOGeDi6fIWVH313GOE7\nvrU=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-signer-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQILyzREBMoOnUCAggA\nMBQGCCqGSIb3DQMHBAiMUND+9mXdiASCBMiAaehzIltYoj91segTqTrka0boquyc\nxwShYx46snwAwMmCjG6ok1KVI9UnvqnscfAmYcCCHUk9XfMpFqqvc2UPMcyB8snJ\nB8ssjQedkpFfhgrDqdTUXxjbIQbkq0jur5gihlW4+rDXUxaZyooZOzx59YE2vWDt\n5NXP54xXfH0Ed1I99cuRIAUCBmcPsEyK3ifnOIfbnpc4CUoGcrRhx+CwICyqRsMu\nSNJMx3cVwtCeD075iFjox07dJTIcfSJwnE88riwPXtW0RnH2M6ljAe+nkhKqXgf9\ngqdBuIGH3wGP9zl1YiDq7A0h0YnNwWQJ4Y+zCRD5SQLecWK0k7Uoph2+EVPFa/ag\n5kXOqr0qFtqCrYxPtx6J77sGC7OgaW9WHyL6cJ+K92JgffLzr3j8yTauZjSPfUVa\nfkTlhiQWitF0jb1LAMGPQjrK1VYg2jJCyrysT/fL33jxItAvddfmkQ3DP3CaiqqN\nxaa8UEbZofceMnr6zVsCbY92R8cZonJXcWG8YkcX/DVUjrZ0Vwpdn0K5JRl6y9Qw\nEfBFyhv5ahdYv3o3hyIM5Z0EkNhGYUVigTI7F+BFDlUlRIkfHPV+9wSaGm0KeUuY\ny+9YZ9yKMwPHKi8yxQ4OhLV1eFq4wtRbY5iYL+e7MGmkyhZXzUJQi1SB2Tuf5ipY\nNvoPvN/stGMEOVhRXiNpsg64pNOScYeBNjuaHsQJo6BFX9hyOd3gZcgEH3Ur5z0X\nOYYNFT1Ea8fibQ7kmNGd4mE4C0XvECLUw1vAjZ1MBMBDMfvUBx79eqtpYeP9NH7J\noRuumqg5dEYdEwkvFiGMkWn44iGdeYlHf1ymQn3Vcljc4fqPlrN7fk1/O8I7Y/9p\nOqfNQ0smZ7MaukOnWOkYL3e3ZggL1rO19IVX1gU+hzHTZyxoIt83/7InYd+X73n8\nlAUXz9hEu5INbIHQuOujI8E40o569Mxw3qNjEyGrHNo5lARtTak1nTaV+eQREw8h\nvGZ0rO1epsxePn2Ya2C/L5FGbUtjHIcS985wjAYK3uBFFJyE7AvbeFfDQkm+g2D6\n+SVgiuSYkILcxV6Bftc7G/rNzKUo4jJCX0T3e16LSK9tEZa4JxeJ13TM9m70qjN3\nKwQYIR+KZpTYXrkBpNbzi6P44caVNNYwc0/2hwiswRxkrTE6cXtKPlfDubHmmoUV\nuE+se9BN/1NyPj7hTq41S1+wddPlway2595d1SuV2CSh66PrFDe6v/3E413SwOmD\nt38CE1Fw9opuT5OPs3Qh+X/ji5Xos7+YQBzCX1meoIlvtMTXLKNchmGTtFtHpxh+\nBgHcjqhg5gaEcI+MsvUl9G47IDXid+7FiclmfNy/+HMXY7SnlehR9iVodt4VXBaA\n7CHwjLfvJwW9UlNlLlG0RiRJpE+B3DvXzQHNJflcLNzHGkef6LXmBr6m3r9Lvjka\nOySqZSSS6a4HHT8ri/sD2mKNWdHvwyvJ2Au0CB/EdOPcU/hrzpRAKzUcyaQvGxlu\n/TOhYfMhdIc5RGlooJ1xcIO9o0f0zI4HKlxbB9lGoDtwBnMJZk9utl8lhb1Qyg5J\n8fCoy2w+L3Amp9YJIhS+MtyCssunJXQpO9fvycqVTdN+CypMgY5EWryMt0gVFU9A\nXB4=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-signer-2' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIFCwYq5VGtwsCAggA\nMBQGCCqGSIb3DQMHBAjKA/swqMV4LwSCBMhpgMEi0L114FIdIXQcyOK79dLLAWk+\nc3ZzICcmESh0gTtARTrIAbcPR7wT2n7H2PDqrRen54sULp9z2Uh9nkfrML4v1mU8\nwXiTjyudeiaLRiYufK6tKqvPdl6ryD3+OzshwZK0Ko8N3+2AEHqyHsbCxDu0XTGD\niQR+FZpmi/mZsLikyN8tlzSMX4mQsY/GiGHeLmcJxMHGp79hc7yq3QaBBr57r0ZX\nLVJHW7q8cfPdIJDknk0GUWRhOXGe/3s0r4OTn7TpixCfd5OT5uq31oTdLHxiHZx6\nGYucmIFOH2QK+SvHf6VpCN293jCB0NMOsZ/UYswx/froJlxLDaMfm8D5WcJ7y9Sb\n2B1YOCufstr6b1DcicylZeiaRvFRWTC9B74vGyHzHkzMnDcBoazA+Z41IRoc76wX\nzrhbcFJGDOf+6hLBx6yGwVBMSLoBhZzANAxeZb9L4U6birMbpi7R/gOBq8Wj7QXQ\ns691fuv/mNJnScTy7Mj8uzhaasxyGZTQSL4nUXxx3zTNj0S7t7WoSRvTxvng5owj\n9ztnNRKFY0LJw13Ej9jhqncyLLGjCzDnEzyLXqpruV0E2hZ0YyVYCeCZaP34WJvN\ntbOOY/CHF78zQ1WeKIqBo7igUUXd5DhUvSaIFtg8k99Sw7itvI/f1yi7mBE/K00Q\nANCdPZ0NP5SvoTu8aDaxlZ9cb9trmqy2bf3o58EpvxA+K7nhAjGyF5TquXDsiq/h\n4pdQdRGwtT1+liG49Y1bfNeaKb8g1d9vmPj9eTE6mKUwLAv9furnrPMuINyOrodo\nGBW1EoUDRmLVuzF01P7JgVakBm4qyjVkq91X3qLFjpeKkNcbxPbeZIB9+UdybBv5\nWh6g7em9YJSNbWu4difbu3SD/P5toNqPyV2DNLPMp+Z6odBH5W2LUQaK0uLPMAFO\n/RT6QAPYPawtU5ib2vWD6rdAdf3yFe5tfA4L/5FFLriJeR61m4x4eC0GNqsFKvjT\nvhW/gLsTOdP78iCiLsFMU5ZtDYtk5sJpz421f4xe1W++uahmMd39BW/GVhcRLcys\nQWejUBd+7nhE6uqKJDhq2aQ0deKCGWEVuAuXwBAB7XwSUxB0Uj8FcXsnVq1ERknV\nITqccfl8dqgQn8Z4xvkC+i/zSNhaoJzvSk/rNOdvhIPVT8zi9ivOnKSRhTeYbloG\nUTw+V63Ec6DZ7xgWX3UQ4WHiUtNCU9fdWZlueUtF/Mn0J6rb4n+4sFF8VUPY7han\n03a4i4a/AdletQypCGcn3NZ55UGjhygJorj1cFIXhkIVEKIiubO0/EH8G0uZFUSF\nocx8p17V8ejAOQBfVV3E2NnXu9DdFvA1Ku9XOPj/vgt26yg9OWmYbo52Xb9vGHYl\nahyMo+8+TbGjhlcKRnBgsMxtif6XLno69TioqBJCfrMdJDCwp72lILB7pVERtjab\nq44IOw8HGkOFAHqo8Dy0UJA5f8njyI4O6Z3xHzvCDpWWk83gr2cAAxvwDxBeVsMl\noBpFxY78Uwul2mJIxEWsb76Lid1ZGTahL+239vGdF8gUAoiOwFFUuVe88uwNqfPe\nvMtIGcG2v6830AJ3m20cP+LQWPOdjcI7jxIVEmgzUPeLkEjLWbr2KVVsWwdrSG/T\nBMc=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-signer-3' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIzyAYmtBZK8YCAggA\nMBQGCCqGSIb3DQMHBAgkCtupRzYbNASCBMhzoybs2Ofcpfdhukd0Pl/snKg5yT/Z\npCeCpAup8+K3VH0mm+MUT56PbID+RcYFDpONT5l12RLjKuv08UcNOcnltSKc2WHF\n81vDhPLzTbWcE8Hor28wdxPLTJzlQrilhWoMlCYSIb8VdCQdaUNLr9GX1+6vIUAL\nUt7JqNu/j9ignlwjWNWy45muRp2LYduFPmWICrzTyaJu6yRUZSd+S2WUBffs81OM\n5R1IqlrDAljONT4b5NxMHu377MPz9LbTUTAb6kWnx1QywP7MvitJPnq0eH9DdQvK\nOo6VBr5txnDJUIU2ozP30GictHldMCoD+DLfaDFyEBWccca0MV08/QOb5agtfYdh\nuUBDipA53jLHwQ4WJtm8fJ5Y9NJHUYVrTMyOxgy0Wkq0hgNHPxkrX5DbF+nMO4N7\n33MhlHhZFSnhGt3QMsY6U3Z/I8syM2Q2UIe+DcoXnaUZzATSW7w0fjNbgqGKwNYe\nDlyrHV4RdN8P03QX58rVNy/w+xp8HJGshK/OiXEWIJag8WU7RyKSdmR+ytOh7oP6\nzrwUYWQFs7KmFl+8SppntXUZ2+zxS2+pdl/lxi++UBPHtBx4m9yYTWmn+8anl8+W\nyPi/BDSgRmtmODq8dnN3B1OThqbynCfzPkvv0rQ4eHlho51ic7i4nS8ToZ5xKQCW\n3P4iCrWKMS2ccwHcV9OOmFPiXxki/53T7bWQ4GFZbaPXS772DQ8aeRAkhS3+crfD\nbzIAuAb/FBHhM9iYLV0l+hjVl4TIUJUkv0Na0dOYLKDAImYiLvUMmjk7V/Y9h/5/\nkQ7gjtA5co7AnK6c2OqtagBozBkIbmfmz4yVEv2KTvRL6PlwY8/+PvxYsCMKA3FW\nP9whaknA8KovOHpEVj1aWNiRCQtHfu+UhlJRB1/dlZSofyL8Qr2JKMVEnJ1llwaz\nsnf1rfX2+TRPx7lHcWbJ7dFVr8I3k8u/5dN1gBy6fTDP1QP2OGB4jCHHLDCBCPG9\nIczy07n21Dd6P7d/iKs1LNT4oDwrZ9mF0HwBJCgZh/7fBn/A2v1KP8Lo3igbbAxD\nG3OcF4b+oRPu5hMOFJYCEpNOFjBbKNJLNXdJi7TV7TjCBOCqxP1Z2tbHPzGZH+vL\nuV/AawjtvrsGGUb9ZopfPxtOenMWSo7ft4oy5C5NmgfKf3QJC73ops6ykv6H/s/n\nhJmgAVM9+KCWxx7HKPeQnVKXziVk7wStWAzd+5sBvuy7TOjOR8LbsvqBadqSE/75\ntRYlUNeVKDA9edcBEaxfMGgcXbvtJuEgg4RE1xI7Iwa6hdAHMvJijAobyTk7pRNn\nZN0EyrvKlXjwiSfPqD6a849F5ryBACgYvWqbassy72MWpcmryp2K63K7Y50anfKk\nqP8JlnxbKPPDG/WiooDJOxdvEJgit1ehcffivn8hlxSFtpmbHupXgNPZLmIgvJMc\nk/G0J4hLqvQlHmUAlVzAWO/E48QyZoToHPlP9pzOuoaBtFc+ONyYbQM/GtYMtecd\neSMCg0Y0kfjZQR4clTBrAjk1YdGuPB0wkXTyaS72ytl/b5KFFnAqTg01QqYm9vYE\nJKrHlptNTNzKPMO4ANHZgsrj1uSBFncsaTiOVvbojFK0M88+p+gdkVoyl668GRhN\nITQ=\n-----END ENCRYPTED PRIVATE KEY-----",
        'alpha-signer-4' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQI60DiBGp0QjACAggA\nMBQGCCqGSIb3DQMHBAiLsnBmhzgy+ASCBMjYVcHciWavqfqrut3NTgwmuPuY3qzp\npJpeuiBtirAkPj8+jLaCCCmVMzC/E8XPbG6yM9muf1igzi/tOinjjZS7DG7j6CAr\nh9yRKVq6RMyscNzhismtqLoWdJlZ51z7NNek1tRLFGdULSm/xTS6zoaxoe9aUjdg\noiuMm1uVP6hRh4D2ufINOstf8CcziGKtWBSj29E1jISDlxkG6NQmG0Ibtp3wWwcA\nFJ/LCWaLT5Kqq3E4+MxAYshwxorUpzZJVIJXS+0zXkzN+RhplNHo4+iXKbASaPPo\n2+a9y/Fb1xoIdnKjBhnd1HkXd5GHHrwmGmudYWSK1lSYHy9fyXXZtxwgrsLWyU3K\n24R6okhVu2NMQJY953KvZ9ZrRsnC1YfLlaZ6JiZrU7YH4cNDMJXaxRDhwThNc2Un\nyGxS9EdXvTXfrWhGrpnvn/FKMqsEawZiNbz/iqMXPslsndin3/INyE3I7AUKZBqc\nNhzpfE7H92+iVTE/8qnz2ysUdmZ3WYdEQMZGaSO9uXdsMRbsQ5UdLH8iZF6b/m4W\njkjAbN1R67SWZfbDUIpYIFCwnxn7F8pOAaZfsy7HBlBZpCak4akxfe92OtXXWrFy\nsh0I6Q9UGopj24xHTAO9kaTmJt4kLpQNd21cOj0byn1d7mdTzdum2tRDRR/I/F08\nUL4KK+gXdQ27ZrYApN0TmaJY6lBFC8hHZP85qmXHuScKd2AHlL4rb9/c/4IiV4uM\nINad1BbqT1BZsNq3oLir09aocXIt0/RmLlGp0K0Tq/mfTcVuFGTjA5/DLxUrnYAA\nTDeS9qTvUhNEm5VpbiHqtcn/FSX5/iRBO63HYBxDEruF5aUeWjb7qn8J9cm1nmS7\n/cZSO0W0glYXliAiCHv+2XFIEglM4Od9DSmzMLRccTUIzA46BGByGKgiotImF41z\nyL77CytYh7cS60Y/7zs6kXgBGm5SpRj6k3WWMwMM1MWThZDrzJcLjHYTDaJQXNwQ\n7jIiPA82oJq2UzNz0ymbQH4UIcVp2MQR5Sj/WfYgRyv7JdDuOlxvsJpJzZoP7VGx\ndbTAl8b3D4jgZCyJIKfpLLYeZrk1FgAicmk0dHvfOBVdZ9EA+5k9KhMjxaYuOZBp\nkVW045IQ7D+GKre+oH9AYd+zehvKLjMtSY+Br9noPv5OAWClgJT89wXs01+dWP6d\n22pw8AJUNbQs57Gylz3hZmNqIXAFPLtbOKj8K3KrrBtPoQfkeTtV+QfbXGHEzSeC\nJVyuHG147BZPqCWC9biVhIWqBtCTeFjALudQQ6N70M+EKIbssoYFe6h7WPxTFbJU\nyVhYoZQHPeEl8eyGeOzF34Y61CcHz3ugAM1PcIrrLZo8/DRw2F3OAy0j8A0+TaMK\nfSl+g3I7cVz14lHVYaFyQFiJ9FIMDiSfYLJhMAfdYTz0IZ7cPZts5+y2djq6VX70\nojnmqt1qhUNLLMmP+QDr5olidntFCWA2VywcmeZPIIDaW1d36x0fXbHXFJzaTYef\ng9vq17LdWeVYPdEAiNEUxtmhGDPYte4HvvxW5XxAt/HBCxR9G8Ts9VzDnm6zQIRL\nInbQb0/QJJ4bKQPnLQShUfJDNt/lZh0VF5y5j/ISW5/WhHoLY7lphsGoyQWT/LU1\nWMU=\n-----END ENCRYPTED PRIVATE KEY-----",
        'beta-alice-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQI7fLnrRSnrgUCAggA\nMBQGCCqGSIb3DQMHBAjlKOF56ZybyQSCBMiSA9R0nhXfqahXI5vmR7id8RLluQPL\n46SlNTNjybjl/+XkNS3rBnj6GZ35KFMRTwhRKHd1LkiSTzdurN9iTV/Yan2lDStZ\nh0mXN+uMj7KUydqqWzeverPzYgJ8TRxu+OupqHqHKOE4JSB+vVuQiaDdQSDx6Jk8\nXAakjkHHPinXubV96v103Q2o3JXy1OtQLVvnE1/mrHd8EabxxD+r9y1Zw2wn0fKJ\nYR9y54mY7ZqutiNyBTI0+WfYtCFF/I9tcYUV6+FeSZbe3AubQ//dU4mgsbTC0dbO\nHpte2dvgkyxvQjF1zyb8e9r/EopUuNMo7ys8gsM6SUsdJUer4BYu+BdQ1kbh+rHU\n7nmH9/ZOy1KGe1euEuhK/cOzng9So3b8ml2CHPs+90Uyq549BjSU5pDevqkAL9EJ\n/eFXxgLnudh8MD3wRwtmxocHOwu3IJwsP2sUPQLZ0AP0wYUxnUczTbEPrEd+yR4A\ntaJAwOcyav4YTauRJmvuffj+DMR9rNqunL76Q5JSHjufzD17kaB/DTgi2wAaIR6R\ngxOUdIpsxEp80MgXxXjnIszA4sdiOWMcuYrk+mQCVmHrHVRyBiWN/MVG/shuIqDh\no8zUNSbfFIbXmTCugRX+b7MMTJJNrlBsRwvOQ4SHgQU6/gl7wQB7bT1q5Wos2Q38\nSIz5KJD4Wdqg+zKtsp026UjLRJUgli2fenyEjRMmllGcUZgFO2E8wjmRJGB5mfkE\nzNCB7AvkjeTUXfG0ybL1sbHvxqb0e15xI8JZ3Segn4L5qKEZgtnZTa2bWpGRKJIa\nM67hkaN5XGEqOSyBQcRSc2ja3sVOQ6O2e+83eA4p45ItDSCS4HLqDepoGkXAbSTL\nidFvWUqhw8Rh79Jpwu37HDDtleaBM17dFtfXHmvgGga771k9ur5kFGeAIU9YBOy6\nRL1J6LRRx07lA2K3GC0Tap8AhhYYm0SBf6LElbzSBOSv1tbqVtdDlvaJq9CazVCT\nAz8OwcAKSjdHl8GTY+YPHi6Fs9qm+9iYW2h+lENnufqRac8ZyXhqLbVxe9YmWnFB\nNV8lfgTDiRroCLPxhGBw2NQbM9qv3UmgPy2lS6TvW+MqriU16RsR3FhVJhd3HRek\n3rYR+QSmCQQUC60RGvOZhT+gzqrRPlsHXKNW4HpBC0ukBzbu2XdcP9YksABmVFVL\nwK5FplMHQrP7Qp7jU30ML484frP4RjjfSFunyHpOChCi223lhBzG40D37NVeA0QJ\nFOEvpZ6CSVGKRUlvyqjRdMUR+jZL/sUJ+eSx58W2XEgP0/vTRDb34aH1Rm0iW73u\nEBhO1M2oPqjIvV99v4u8PDerWnExs0yllrbNPWB7z/PcQBINtjop/qYCBInNt56I\nLIf6SofwBnPWgkMT/T5yRi3N4aI6kKD2Cn9x9Lzk3AhZ+1gz7vdNXipgtQglkA77\n3BOwTFklM/XOSPTD7RWnipponqeeIcK+Yn40Q7X4fr9dIqhU5rs7YS1BzvhZjvEJ\n3DEw1DZp7GJzqENliP6nYmCmzxIYpndLYrLvFKlSK4XZ/duGyrwuhAagVRljtU+U\n9N6i69wkKlk5fZhrkvS3DSQYXQ/QIVc+gsOoRJWHgxbrYlParwY1YVUksc6mXg9R\n68A=\n-----END ENCRYPTED PRIVATE KEY-----",
        'beta-bob-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIhd4ACmZgzsACAggA\nMBQGCCqGSIb3DQMHBAhnicl9hoY7RASCBMjbi3pUKJ00QzBdNTKHp72dC9a01Y+d\nAfGAJExBMnWOiFB2dW0443fVG08Ux6bktk0P6H+sAUVreUSLaW2MO+axs5nIgdQL\nzVSUauy7qVbxHBwrfA5oZiU9PxfDru9VcJhMNnDeCUoZYtaOgksWbjJuu+R6N7d4\nbvE2ksDer0J7LF55YFCN477MBSHch2fDM7IT62xzbMvA4s80v6Di9vPWQftSp5Wf\nMKUPxO0fGW1qBoMPJ9aVxyEYJH/1GwTcqf7B3ThH/6wh5D+n4U32KwzIeMaiyPI9\n0vKhVu1YrCqd+nwYGzlijYK8L2Ed+CFlSPwHbXdlOyw/n8D4CPT+UYMARTLo99Y2\np/6Wje7WYjetMBdxLE4/G8mceFtNmQ+niYpZj2fq8a5a32LuwKlZcxa3pMgpdeUx\nr3plgotCqvvfd2f1M7wNrviog7M89pj/H9ru+phxmYB3QcETgfnBPa2Hn5m59axe\npE0MNJ9XkvDmjZonyaLTaYTxn/NZuwUYCZqGSeSZOEoUARG5iapSgKMdbGXkl9vC\n8xF0TlWwmJW+K8DS++HTGibn7QJSKIsKkzCoF9MMZdUva6lyE+I04aNADuQyJufV\nH3/8PDZJBg/UpKioCQyXTj62CGBbAmIisHmxecw+qXN6lLrKxXnDOA4hMj0+PfQz\nVrBj6G/+aF/oLtuSf/WkgF7g2XOAosF/EQsZeTXfYPg6bh7nepICnGG82JVmofVt\nLOIVYV8eyjc+cToJLWvkvqVqNklNpKH2mAM4pV9nK7283j89aDOVaBtciohk3ucr\nqvs2AT8MmAlgYtNvhYyliomllOZsFjsZ0n5CaN7rvYkzinvqrsAaT31ZLhyNf/Cg\n+YF2+bNOjHpuuX2tv2j9PrXUcWNVFeSeyzCOt3kSRgcEADuk5e6zaX+ubXdTvDuG\nuFDCtCZY38inLDjhloD4poBBBhScGxhzcjn5OWcxPlrc2ylbilA5tkkXuPmMNh5+\nli0KIF9+azxV3M5eJof+JQjx775XuOdPVYQfBzZsJOcmB9+XiYG5GrBQTkSUeVze\nC9dQUObsdYqq2GR37l8RplhaPiTJTwG5EMuh+HauHbienItcbH7A5DsYjiMwLqIo\nLYVsJgSInFrxv386+fZV6dxmjNxqbUntzFsIyRBG9Zdx/dMWbX6ZxF4GfvIwfa9n\nFLkKdySxU4/27IZ1+Mqb5MwVNti2KEiPdaC2H76JWAYyrEVkapJR+WN0IcOly4GK\nutfzoZEtL716xyJ0ekM67icqR6OC47cm9K8WMjufHG+5sqJGxgfmfuVabNEh6JqF\n45uA0svcvUToc6xb6Y3NRCHCFepAxK8Up7SPk8tQpXbo25o1UtDvgSdjDATfF6qU\nToM2Ypkhsop2d9E67fykoSu+v38Fjoem3wXDoPV/8/GwkqBpwqFgiUIpQsTQNIq+\n7p8lTVpdPOqrk/k2TmQ+nhW5dJIMcycobanG85Y20LI/dVmkng9/cF2XpGqXZodu\n9LprO1l51huU8p5KIQGz+AV/u/a1Y2iK7sjmdQpmkUSon5FHpMRGr1WYc6cIU3tY\n3489d0Bu+bjFsNh8yFZVZaI+mcZLb0hZqUk4KH3bsZP9+kHknafDtUQigS7YBYy6\njng=\n-----END ENCRYPTED PRIVATE KEY-----",
        'beta-datavault-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQINRxsRpRw22MCAggA\nMBQGCCqGSIb3DQMHBAgaMQkARlUjvwSCBMj+niL2NH8wIkW0SimQDlcuPkyEtosJ\nn/qkB+2vIZtx5cFIKJxJ2+Om9gmS5GT3Fxmh4O/bllrJHRUbcOfg6z1K4smSMiUY\nTf8pl9Ww+pY42iYlN0cnREwdmwRjnG5d/Sbcr5G1VDg1wi+mMh3EZfms8UemE+uQ\nDHDzapSdhSM4SQrdPEQzX6V/TR4ziVa1GEYE0pqfkg+3gmptVfmhucK+4SzL7X/H\nfkyn5xSIgQu2X0M/uwRaQiO82DnwfMd+0XBXj3Eywmbhl4iVa/eJtfKXvAbxdc/M\nVg1oiQ5v5TY6x2rY9BbJ79WdqPGVuhOcwR/+r/8BmzFBkQsizn/tTRsEsBuwwwbs\nqT7+q/rXFfjx4t+8QsTN4ZOwcyAFQzopl2b8gG5g/0m9dYhl6YKXVNs9ZSfwBWwq\njyM9iwgQm75QFsnddA/xtxLJOqvH2INgE1s1yDBYEdfVpmpaWRQRJsXqHVqQSx87\nTkEKkj7I1NqBTtWuWjsluzTsNSBNZVLmtOY6SMb+INQP8G9j8LCOkBxZ+z2emSFm\nuefw4J2SK2exhc+9TKE4fig+ErSFRrDjWJRakiVJvbWnnXIRk3LBzyP5NBr0W8yQ\nnXCZkReaKKpkiIHfa34IF4GlkczsNkfpMcyy3bQ5y5h7JDIVplwK5sapU8qG6mu6\nrUDsPHs0y4wCXvPVis1Z9jfBADKiQ2p3ifTzmHASAdkiyZMZ+XC7ATxqZFh3Kh0j\ncNxnjjANuFSk8fBaGv1eCUypbRmCGV7922K6bWr0EAlZDbwCVbYQ/6KhzKEUC/cF\nYx3vEmALs4gSZe0mJ1k4FZ3BcbcVnc72+6HIsU0soCb0BDCswz+XeGsblH0+PYEX\n7/hb/H/ebpevtok2G4QBffy5K8bfK5MdxCTVaNXMIWk8a4hMSDPvp1e3/QFiOKyn\nL8gwer3/8RDT2IWy4+LQKwcofNnH6Q5EbywNFJiVH2jTVqzH77vKvuERk8MifQN3\n+eYH3ue3g261Yj8YcvGkNV+U5Z/gpxy/T+bHgJlwyzCXHXezXYcQJoz5Ift3Fycf\nU10hNCjFFiLEyXTUePqEFz7oWa37cMHIyTEZZ1jEDw3TOvye9BknXy5jk6/vJPMj\npep3iPMgFnIeH3FpLqpfd7w1D8vBGhy+7szaYZmArwkGP7X7tTOlsSilcJ9W6H8W\n0OvJ/k5wM3OemdEfestFUclBbLpiPO9g84+bdXfpWqvhITdww1OT+VwEOiukIM4r\n47NQJ6/qWYZlA9xoNctbBAIqBaMSADbJdODcKodwCQ/HNyG+Q2MN3tfORCDXhf2d\nJ/SyxWi3UZ9CYXNGt9El3mF2QmtBFNLRZD20EVcPT+R3uKyeDLsvqm6amfY9+oGc\nkqn4RBSR6jOoIdnFNQORci+CCs7w9Rx/yaXKplw78bYtXD1c0TlY1+m0ekZ+2Snn\nJsC74STkIlhsueWo2Ttu8RdpsEQV3But9gxHswSJeAa4A/Mn9sE/q7ml75/ne7kC\nF0KIpEfhZ98MxbvJgyZsccgswPMleRqPzJfFQ9T6mMVDQ8nDGhMsYvEj2UeVDNU9\nu8gfT8rAUkVbAj7J3xzQsfbnI8p7VwDgRyGuVTF206G1FXX/w7gAD1zTBgI+7doQ\n96w=\n-----END ENCRYPTED PRIVATE KEY-----",
        'beta-root-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQILEEQGLPdA3sCAggA\nMBQGCCqGSIb3DQMHBAh6JzkGVnbaawSCBMh+DXJxaOYKowJHydO+kXKy4yKLnfK1\njZ1PsW5DJ1XUxLonlnXC9YUV8EhQGgH4iTeD0rK9ostqXbpsraxPEx6MjJgT+FL1\n70d+HIYp89/R1P4LvcF2x0+giXvFGIzH5cIbNm0oTKIPEUA6S1oVWXgWvzhGe8vf\nmZfTHAEDK31DgbKAFJvLKS41Ynr9PiiJ82EnVSH83Y/ttRewitI+wUkq1y6F61/V\n6VAhJL+KNucbYKNCEgeHnVQkHaoUKSX+zFaof+rCeQzGbCQyptCzEXdgoNW+bQtN\nTHl6KYO3xlNDjY2dXEm06pSOyoAZupDI1Mvi0LFq+dW7iQojJKvyZWRImtfhsnkJ\nHShVsaftj/h2lnKltKyh0V+Aka2Vkce+UWtZjkucxgQG5TSXdAhMmjE6hNmPsVWr\n9mF7j+sUtYnJAbuVgmwxKxvurKhLA1el+X0Ux22N4vaCzsehRvpXDBv0Lgyk/1tJ\nuqgjL713zRqckbxrk725oJdcxjq8BhfAO71kNh7Z8iYNHfZsVr/5btkQ0h8ZFkWz\nQYZQzCGNd0EXKgUYk/lqem8FFbaxLQFFjaDpxl4yY4tU5YohpREcf+RizSYlD1VH\nD69FOxAtDrKq9IeDT9Aj2Wv4mA4p/GzljkXFMtB32S5RBreMMDo2HlCoquGeZRbq\n0hLThylwASLEcB8JiBgbLbiBCWu0LGmUkaMYd4QsIP+Qv0O2HuFIIXImTZwplB9W\nIQ+Nkk5yggR/lLejIVTm3KYlMGSYtg7Jd2khppzxrJGkXYS61kTFsaA4BtBEf3O/\npcKVC34JAfmjVT5vgHRrbx/P+SfALFlV9Xd+wxby2nkT7ZOAi6g7SHEuxt4MxrqQ\nWwiZK4FlUYMBeZ8CvyJbeZ3MRpUyg80FsFowX3lIrlFPsDnRTS60F2HConBQ8jBu\nc+54tBOVOIQXry5BcvjMUPMchIo/pU8NVveYX+hFcYR3HhR8fXkjO/BqqnJ3m5xv\nhOtaaiBamWsw+1S2/1ch4jkcJ/oaYhJIUD4Z4uTTRRXMSgZfdlzBS0ioumFn73mp\n5elwIlvkk+ZIgnEZemaCG12SOzONKlEZaPr7aOtU6U0TgkiFiBFTgm527++v/i3H\n1ggKdQJWA+Z/XsDY/gX1GgDMkF/4TyOeMLd5JIJZlqrJv7Vf173W4pR2S3bEpTci\nnMT5zsQxAx0GS0MabDgsM9OIJEXBVN28tq+w0G4ZGUIZvmT1ZaskIzp6O8erc0KB\n4s5r+UMU5LDVhiAH5TsbY0PDym+r/inzJAieYjCtFe3NEWRH7vzur4RC9lOzyyfy\nV0u51QqGqSvMGp0yF/0KO48mtkCl68VMqfEkZTY/xSK8+mFA19Tlh4uZy/WoUL0r\n4Zj2JeU2bI/FPfgknuZK/nK6bUNhorz5d3hng2r2FL6h8FhQd2LC+mcAxelWCDI9\nQM0ebt9w6sbGJUz8saIhxA8hAXgS5Tak4tQi2ly/gMgcWYJo/N0EOcbVHqPUqoxC\nyb0lvdHYEpu7WIP9tWxNE1d6pKV+DSVIBXlGTZMRDmkJTKxld8jIf6hzYue2TuTz\nPwbR7OYTPuT4t9c8zXNTCqo7g9dMnGfLWp0H+FOW+fEhM8PBI4lkvsYHjzAW8FK+\nDUQ=\n-----END ENCRYPTED PRIVATE KEY-----",
        'beta-scep-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQI/RpBRLaHaY4CAggA\nMBQGCCqGSIb3DQMHBAi9oEsMcT4AIASCBMjNZ8+mrpxsGl7N8j8u2cM72a+1JyK2\nPpmvTh7I/k8ohFxXQONFDT/FCCXG3Wz1S9ZdbYPL/IeMAXCoTSM4KmmMuAfIDJoW\nwe+1r7YPICPMMc5U/F8eABavQf0BX9mIn5FMANuH+RhnUUK4DxzaSKEObA9qUVfW\nHD8f/Qr+2If446rVgHQNXvovd0oi7HemBVUSe7k9T+Voqtjl4BnK/hzInJ/p9n2Q\n9TKRNvQvV6BeK0b+ioKJUeCTQsHaB/HCWCu5oHKRjzK1XPAe7m3oZGR5/AbSSOO8\nqUu3udbMA2lGTo1+OaxuR+SYb92WVJJ7TJH9TDfru8tnRk5HFd3ANmyrA/QUnXFC\nhPwRECrp5tt9YLOw+1zOiH8wahg60L79q50sy71T9Wk0grzDeDhtNtFNavrrkHJS\nJHxe17MVho/lHX8rf7lwuUo7c4A0Q7TcpPCFAFk1pTm5a9oTXDWV0wXFJnVSwJQU\niHvrLT34wsdSGaDuWEemTO7jOHlEvUwn24ewRQqoqg8spGQM455U35a1EXc3VGAO\nRjqqYckb7ui1jTnURJs6OCBc8xDqf5i5AF1Udvs3e21Q7JeO8g4UU2pRQDmejs6w\nPH1IvKojKA47lpbyzBdK9CoMxzIMIhjG0vUMfxaCn3pJTXZXi8YgQ56LSHpALxr6\n1wvG0TAD7+D3zvQPMSxnpLVfHR9D+601Mpm0TDzjTHCel2sjsOhogDpp1w673HWU\nB5zgUPIw3QyrTOrS+5xxISKE7FN4GHzPvGHZK4h9YPm6jSjbal9nUSXCdwP9hR8P\nx+jPX5+dTCoekm5gAiRq3fl/S+LfxRiHC/XayvaJHBOGpJ8kT9OKJkETo5DyRTks\nfrbZbPU5eCRwFu36wXZA5CcE9pba2cUCtLN88qw2jwQfas0VJ49levIJ5v+u5Hy+\nQX6htAgGEccw7uAEEWPK4O1JJUPJGUIQxFON73MAYzZkkq6swD/+eCVGoWGTuSht\nz+SFidHtLAHeZfxdaSeRlqGT7+brvXdrFUG4i3L7fikg7LGuuJPdLu/SkBH7UNqC\nYdKlg3u2PMp9oAVkU8bWW7glwnB66CfHfZD+xS8RuMdfAgAOjVryU6c/aWZ/TWbo\nNgwuP8f3x3PqS2TVEzuHtauB4YFlTCEXpcuXMSQD+MFrKLPZnh9VjesoL48lqdVj\nU+ff1mhlO17bkkTPE1faVVdh+Z1F39E/yy/56EMF5mTFDYZ+mozUHUc3xwwPiocb\nPdk6fJ2ClO6Lses55cWKYasO+xDQ1wMgLD6XEpU6w2RXKDxoJ7xxEt4UftH5ociH\n306lQNlBY7by5fjtWBoxAYgbmAXKM/Xe3M1XW8XfVUBLrs2Zthxmd/y1699diqZ+\n5yGRo7SyBYPTPNWu5VRYXdCLps6X255LRlliLMNmC8yK2X08H9q/ulQoOw1dHxAR\nIm/frFtBnlxYiAuNQf5FMo/Embv5KUcV3MiF5zzq3wjzzHWMNteO9fmoKc3hFGRS\nCWporrQIU4RaIN7qHPZrwUwM4pDcm6iBiAPlU3q7ht7DMPXSQWCIgaIzpKGC4DTk\nLLnYoAWLwwaLnoCmIjFDTfHx8XaG+GtCmn1VDKeW29rvNBye3/ucv5SLZyrWmQV8\n5I8=\n-----END ENCRYPTED PRIVATE KEY-----",
        'beta-signer-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIZtULav82yKECAggA\nMBQGCCqGSIb3DQMHBAh0wuYkDeMG9QSCBMhXNiWtuNBWAyDUCNPedRjuZAgY4seo\nfOErk9iz8w+YCbbglWcxLWfqSOJRIPFTtBeAUNwrAqbO6iHK5guSmo65FdUMgMpD\ngm7aYfuqaYuSQ+JqYdh7e+m2vrGhA2gNOriFb1X6tgz/noiA7ckiUQHxaalt2mXV\nNo4taM5bLmHd8UbHotP79UEcXZ/I0I7AjT5F4lzqJzT8Nv9aRrSPCWHE6OCOa2+T\nmpQ9eRElksTYHLlm2CO/DotOY0A7A94EER4lEITjyjpZK3YOpzKgoSuR66pMNGVM\nUjNVIUOMpsAmxyrVmUrGzG23Zf66iLerJZQk4+q03AvHzCwkYfthP1uLNobDqZbE\nHiMEZpeMc/qrChjw8WE2A/34I3SPooVcCRTi/lA+qjAYkUpIWa7m3oQoJV1F471c\nwsIWIdmgkRGrIvmKL8r4b6Mx0gad3t+pY3u3CbDtvjjmMnhmTUEUTE3i+fAGhXEq\nl946proPxYmdTn87MsvqjdyZ8ho/wk+1ldtMI9tyNcc1WRfKDCZ41UXmAzj8+ZiJ\nyPfuPRz7/lEbJrtYFdWddOzVCQnypdhxy+Q/HdEJIlvoJJN3g2cUvRiO7fWwR4PO\nFHBrLX3mHPkxV+lIqn5bHwG9s5tScXJB5UExfs/K0X6lIQ1+ASac2wMWuEHbAmLU\nb1+GRBcHBtr6v89ZzWq5RWx9UQs+85s61oN3wfIrMRRZVAqXqUB80UEqyhHlP5k4\nffwp+iFiev60knWoO6dsbMBlAtzz7dzalmfUblEtHNNWy8ZviClmqoY5GOAEYXBZ\nLdlMuV1bExA8GG4BR9tvUgxQGI2nVuFv+Nr1TWMoHxQAQA0qD/oKmxT3Tn4iQRA8\nkqvZbXnZnSZzJBtYBPIi5aGNDGp0fpmwH0EhK9WVxOd/owRa98X9dKdGq2QLOJFd\nhsvvVDb5VqzpyUN6u5MLBdygp3nOSzoZlxdiW0sZgJ4JK0q4Dc8Y/zjI2RzCuyw7\nGkLn76CfPFTHWjKZ3QiTVdv9guXKXIELUxNM2LTgRoVf63qj/c61i6r+5875cPSj\nLUeA0c1QLRcwviksT+IVhvZMsDMwVcrdNhlDM7WbSXSAxMXSYWW/Q9PGGxTA5Jci\nzLnEKnkfsiyVv70IEI0GUn0+J5Aj4c0NzU/EuOVw1an0SiwIgIsUUQpuocTJIMdw\nMoGBLxMstu3hs1GCdebZr0Kw2cL5FberxrG/kncApdBLF1Zv8NYKXIL7eFSGBjDJ\nimGWxoYsU+X7wmxBnIUuxyLRL4+ce8mr9UqkEhELb3sfvPfTsyoICMYI18M3RILg\nW+jb+nd7Cx81bwt+ZekQNBbbIKKAnTh7W27oP5SKrejE+CkFN9+vW3I73QBCiTuI\nvqXSHdTiGNAnkrjPSDK3RblCpX3v21v/i0KaNKfM5WbcgSOzWeILnLbP7qcZD5Oh\nr0OoAOggDgcuuggWjSQE9ZLZbzK67Ief4QbDtXRXJPCbt2hMSqz9EAugLNzgSo7c\nmb+lohfCleFxGyWrrvK4bHGQkM86RQPCm4Z6f/yLW3eNAV2knvkPxqlN9Pwk03Q9\nUFa7caYUZFhz7IerwI7plYsmg8O9UNTOj5kjotCxBmO70ufZW3lCilpP2UdxNrwH\nSLs=\n-----END ENCRYPTED PRIVATE KEY-----",
        'gamma-bob-1' => "-----BEGIN ENCRYPTED PRIVATE KEY-----\nMIIFDjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIB5mUseb8dK8CAggA\nMBQGCCqGSIb3DQMHBAipZHl+FUsXPQSCBMiBnnefjHYKG+1pvvyAiralygsp+FQl\nSNVYlgI9gpcetRQ5f0wOpP0k7kxCqZhRtVRPFE6+mde3iRdgJ6pH6WhGdI2HXOYv\nsYY7BL5qCKMhZCdNbfPeIJQtZrUBaXjvnt3n/2VmJjSZfzVQx30fBpEA9XCg7fwZ\nlw7FZnl9pnnYHYZuDI5sX3H1K91msSq1T7AiWUHO/CT72717y3nKAnER9juZC+jG\naAESez3d6oc3/r+P4bZokpxIW0gpGF5dcovltwaZhcWMpp3MkU+vrmMxzcr0tcvZ\nUb5PzM1fO+/QgRxs3WjFk8moEMdchU0VIZwp51s2j+OJbH/qm8WQVSm39FZ6iZrt\ncC7NSWn+09yeKFKHlsmUe1umQ/1VW2N/uscARPzKsj8yf4EjGBw2mrMsLuVmC3nt\nHKP0k8BdH+YIe585Kjz9zT4cRNV/XbZZAUKIhhLLsgnMqss0U8x3WcT1oJdKoMZ0\naOXjqnDZSbaWZ5r0jZMhPgxdMLi7ogI5P0/2sURPh3mfbdykGkCH83zuaTbQiZw9\n63Vz8tGKB99LlpmRWv/jsL0QyaVeAdNE/TUyAZ61OLu8CUjz0VDtSHAgNVq+QFxm\nGi8ppr4hyuk09eWEHRimcaY/I4DCt6yCLQu2fwCV8WCkBrRhiUYOqqrfzEertEmY\n9R9uEXSSvtCCB603QrUzniUrOQEJb9Dw/OkZmPEYN9Pz5medtRMR2dr0HoDNtPGQ\n6iZTUXsCfWYuK7hxSiTs8YtCUQ+CRHhOG/O2OSDuwSS/YN9/WO9tsI6AM76hodIB\n3mJqP7+sq/wtDjzyIHl2H5Degudox6gPyTzhwbCmhdyFIR5EgXw0pkMv5YVXwCcA\nG9s+rpAPL+8EjRX/Y/nB+sXS4E7yLnmakVK0kjBynLTfU8Laga8Zuf+EYzDTeC/g\nbV99E5+IsjiLHOfiDvudsihlRNAfcxMZjMSAvei93nNwIp3hynGHArMFDPnihkns\ncfW6GRjvgs9NAEFZK4bISPhiL3vJ7zJVFh1H7SZ9sib+pInf5JSooXbFVoYWPzWL\nmM24YHgdcxAnCxzxnwoXuikn8wFDBbzlC7SN5tolOTPpBOeN//rIhqNr/4cHG+6s\nv4hLxJN1N1omFSZazWEm/xE6Qfye7etUYfDmh00tODVetSWKYvSrzygnd82xyJ3H\nq2X0QVyWnF+dLWvFohciNVkNrzTH4KmpSq6Ck8mvQvecJkbMK/weYY4kIkZYhpzo\ni7nuZhXaq45gRvgAZB0xvlvgtETJNGm8NkJXrJkthOvzjzsP+6BH9KjD77dcSxy9\nFYDZ93/Mf8pugr7IJe+lyPhqcgLy+iq3/L19I5MJUWR5ueYgVEYi8RYlW0c6vBfb\nYq7H4Jb4eTM6KGwy2dp7Q0njcSfRqFpqkOiOeN0iv3BHhtfpNFyckXg203aiGGkK\nue3zoxevVMKVhePwBgPv2xQjljqHSbJcMZQjUa0vp2vKHuOmL6jbkgzSyWaPSwTa\n+GPWZFyH6UApm6Wp+CocXKAFRS/zm2QYAh1XjVbQ9QIR4CXFjkhq9uKUT+OnaPtv\n0jF64l4E2Ckw/aKEynJdaIgiskVk2b5z+FM9yx8Piw/gkK4A/oihFsGBr7tMFI+Z\nLeA=\n-----END ENCRYPTED PRIVATE KEY-----",
    };
}
sub _build_certs {
    return {
        'alpha_alice_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Client Alice 1',
            db => {
                authority_key_identifier => '3D:6C:86:0C:00:F1:8D:E0:5E:0F:84:8B:82:DE:9B:BE:4E:C4:DD:43',
                cert_key => '5',
                data => "-----BEGIN CERTIFICATE-----\nMIIDhDCCAmygAwIBAgIBBTANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGzAZBgNVBAMMEkFMUEhBIFNpZ25pbmcgQ0EgMTAiGA8yMDA2MDEwMTAwMDAwMFoY\nDzIwMDcwMTMxMjM1OTU5WjBdMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZIm\niZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxHTAbBgNVBAMMFEFMUEhB\nIENsaWVudCBBbGljZSAxMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA\ns0r8PZbi7blLQPFAByS6KjeskSOUeXHdr648bzINBpvif1w11djPd6hf+cRptsqE\na9qhsTthormPMWGHMAeW4ACjg40HzPy+PvUL4qXAzkD+1X2EnZAbljHvpwf8s6Tq\n3HxU1jkYgzo8NwEfyYC6izAOlfKNgp/hxkSmJs3JHK8fLuJNvuxq8Y95ABmQKPEx\ndd3w9d7VlWor7Nusv55FozMOiewwJ9fsZQ+tvEfXqxE3oq0Yk4uc2aXFyGVxnPf6\nU4i9+r7K/u0T28W5NXSzuiOsxSk7r/1SRptiaSExmhbEOOv9lpx1ZVtP2zaVpAvg\ni7apXCVPntz4l02mtsV5CwIDAQABo00wSzAJBgNVHRMEAjAAMB0GA1UdDgQWBBRB\ngrjwxnMHM8XDt+GTJnp0wDMSZTAfBgNVHSMEGDAWgBQ9bIYMAPGN4F4PhIuC3pu+\nTsTdQzANBgkqhkiG9w0BAQsFAAOCAQEAPWo6JtHLkvIV47SzO2wGuCq2jppGn8fv\nozMmZ0xDRwQdVIEeckcNMk4pOKVNLLXHj1gCMtoWAJoAqrX5EoyRycRcfszlxPQb\nx2S6CXyDU2rwrYuj5Hl+XndpiNOBDSkouH6A4XgMA+dDcJelhjJjb5SjAvFQ9Acz\nc+3huBjOKHymBxhu7sWewNrGCRsDLb2jpxvUJJpywOi4aWxISeajjE/d9mR1ZQuJ\n+oLJYQzHNfNWnisqvfPQr4V081DSYvLBmM9d+AxneUumctwRoH+vCe1TIgXitTya\n9EqFpjDjV96p30VQ/eW98wbOFFo3Z38RJzqcfEg7Ye2vH9WC29AnpQ==\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'm-txAs1cQpTVgmidIrvqq9U4250',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Signing CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'j6Aps6CGQftSMQM4Gjue2-3RmFw',
                notafter => '1170287999', # 2007-01-31T23:59:59
                notbefore => '1136073600', # 2006-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:b3:4a:fc:3d:96:e2:ed:b9:4b:40:f1:40:07:24:\n    ba:2a:37:ac:91:23:94:79:71:dd:af:ae:3c:6f:32:\n    0d:06:9b:e2:7f:5c:35:d5:d8:cf:77:a8:5f:f9:c4:\n    69:b6:ca:84:6b:da:a1:b1:3b:61:a2:b9:8f:31:61:\n    87:30:07:96:e0:00:a3:83:8d:07:cc:fc:be:3e:f5:\n    0b:e2:a5:c0:ce:40:fe:d5:7d:84:9d:90:1b:96:31:\n    ef:a7:07:fc:b3:a4:ea:dc:7c:54:d6:39:18:83:3a:\n    3c:37:01:1f:c9:80:ba:8b:30:0e:95:f2:8d:82:9f:\n    e1:c6:44:a6:26:cd:c9:1c:af:1f:2e:e2:4d:be:ec:\n    6a:f1:8f:79:00:19:90:28:f1:31:75:dd:f0:f5:de:\n    d5:95:6a:2b:ec:db:ac:bf:9e:45:a3:33:0e:89:ec:\n    30:27:d7:ec:65:0f:ad:bc:47:d7:ab:11:37:a2:ad:\n    18:93:8b:9c:d9:a5:c5:c8:65:71:9c:f7:fa:53:88:\n    bd:fa:be:ca:fe:ed:13:db:c5:b9:35:74:b3:ba:23:\n    ac:c5:29:3b:af:fd:52:46:9b:62:69:21:31:9a:16:\n    c4:38:eb:fd:96:9c:75:65:5b:4f:db:36:95:a4:0b:\n    e0:8b:b6:a9:5c:25:4f:9e:dc:f8:97:4d:a6:b6:c5:\n    79:0b\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Client Alice 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '41:82:B8:F0:C6:73:07:33:C5:C3:B7:E1:93:26:7A:74:C0:33:12:65',
            },
            db_alias => {
                alias => 'alpha-alice-1',
                generation => undef,
                group_id => undef,
            },
        ),

        'alpha_alice_2' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Client Alice 2',
            db => {
                authority_key_identifier => 'A8:BE:4B:1B:52:18:61:F7:DF:06:5A:5E:BD:C5:6B:E8:18:28:3F:ED',
                cert_key => '11',
                data => "-----BEGIN CERTIFICATE-----\nMIIDhDCCAmygAwIBAgIBCzANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGzAZBgNVBAMMEkFMUEhBIFNpZ25pbmcgQ0EgMjAiGA8yMDA3MDEwMTAwMDAwMFoY\nDzIxMDAwMTMxMjM1OTU5WjBdMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZIm\niZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxHTAbBgNVBAMMFEFMUEhB\nIENsaWVudCBBbGljZSAyMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA\n7x2F49CYfCWMlzeQHzuRR702rTLHauBoigfX4Dh3YmFJ5Hzd5rlXPLEGfxXoeg1W\nCkeOdse+0sTML++Fa1wHg3deGvwTjjuX78LKLWlwv+QOsE+reOZ71bkfF8WyK7Z/\n3yinCMTiiIdcv4sMt+kLYDeA3WAT8oZbAuGwKGd+1MSL6Y/+QAndrW54dCEdE9sX\neAZDbaHjQ/3TpfZqk+1m1mk4YoWs6kv/yEFdOTIHb/rleMI4rFM+TyfsxINlWQSU\nd49MpgwTO02elyhum1APKxia9Jqi0H1GEZelUf8pIkHFoXE6TWZJxplQHCX0op5m\nlOZvA1tsVfSkl1+q48+sYwIDAQABo00wSzAJBgNVHRMEAjAAMB0GA1UdDgQWBBRA\nKG+7EtyePgOJ03PacDoWwnG99jAfBgNVHSMEGDAWgBSovksbUhhh998GWl69xWvo\nGCg/7TANBgkqhkiG9w0BAQsFAAOCAQEADfelYahloNn/TNd33h64fqt96WhDf/rr\n1NB3UBBOMerg66V0L/LVa84b6+rzg1ohlZLG0we2XXFCk/jz2JPFkKlhHCztVkQJ\n8VK1mXIpnMgRar0fjJyqOtIt7CuX0uEf2V2IaRxSWgjR0pdzltU/ao4g0uF+SO3j\nGB0jXJ6BeIb5xuiZUja9Vh8D90FJovL1tN62cADUAPO4OxT/5mULipWEVmI5s9oH\nDtZmpd7suJRePfAVT6jq91B29NTU4qm7miguNgaH2zDZo7PivXCAOlrfFP0FObdY\nD3W1SZwcExL8B0BPh+NlBgZGEpuGwYPmMt1D0C6prc3tw2aXV7RzrQ==\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => '9izJ2MMHSIHneoNsEWjdkWyisOA',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Signing CA 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'RE35XR3XIBXiIbAu8P5aGMCmH7o',
                notafter => '4105123199', # 2100-01-31T23:59:59
                notbefore => '1167609600', # 2007-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:ef:1d:85:e3:d0:98:7c:25:8c:97:37:90:1f:3b:\n    91:47:bd:36:ad:32:c7:6a:e0:68:8a:07:d7:e0:38:\n    77:62:61:49:e4:7c:dd:e6:b9:57:3c:b1:06:7f:15:\n    e8:7a:0d:56:0a:47:8e:76:c7:be:d2:c4:cc:2f:ef:\n    85:6b:5c:07:83:77:5e:1a:fc:13:8e:3b:97:ef:c2:\n    ca:2d:69:70:bf:e4:0e:b0:4f:ab:78:e6:7b:d5:b9:\n    1f:17:c5:b2:2b:b6:7f:df:28:a7:08:c4:e2:88:87:\n    5c:bf:8b:0c:b7:e9:0b:60:37:80:dd:60:13:f2:86:\n    5b:02:e1:b0:28:67:7e:d4:c4:8b:e9:8f:fe:40:09:\n    dd:ad:6e:78:74:21:1d:13:db:17:78:06:43:6d:a1:\n    e3:43:fd:d3:a5:f6:6a:93:ed:66:d6:69:38:62:85:\n    ac:ea:4b:ff:c8:41:5d:39:32:07:6f:fa:e5:78:c2:\n    38:ac:53:3e:4f:27:ec:c4:83:65:59:04:94:77:8f:\n    4c:a6:0c:13:3b:4d:9e:97:28:6e:9b:50:0f:2b:18:\n    9a:f4:9a:a2:d0:7d:46:11:97:a5:51:ff:29:22:41:\n    c5:a1:71:3a:4d:66:49:c6:99:50:1c:25:f4:a2:9e:\n    66:94:e6:6f:03:5b:6c:55:f4:a4:97:5f:aa:e3:cf:\n    ac:63\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Client Alice 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '40:28:6F:BB:12:DC:9E:3E:03:89:D3:73:DA:70:3A:16:C2:71:BD:F6',
            },
            db_alias => {
                alias => 'alpha-alice-2',
                generation => undef,
                group_id => undef,
            },
        ),

        'alpha_alice_3' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Client Alice 3',
            db => {
                authority_key_identifier => '62:35:CE:53:91:C7:F4:21:AC:5B:8B:A1:11:60:DE:20:97:97:F0:56',
                cert_key => '19',
                data => "-----BEGIN CERTIFICATE-----\nMIIDhDCCAmygAwIBAgIBEzANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGzAZBgNVBAMMEkFMUEhBIFNpZ25pbmcgQ0EgMzAiGA8yMTAwMDEwMTAwMDAwMFoY\nDzIxMDUwMTMxMjM1OTU5WjBdMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZIm\niZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxHTAbBgNVBAMMFEFMUEhB\nIENsaWVudCBBbGljZSAzMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA\n9UJWUSSfiavd7Km1pnFAyt0o32qCzC9DIAVTNZ8vQEmIKwXqAOVykgg1oUUQw3F1\n1SOmqf0cM4hsiH85n4aA6d5cc+RBongukDqGaxupxHAax6/hlMogV9WBOAIDBdy2\n/LcZEPecMzNhnk9AawZnFD5Zs09/1553deIKHdGqDjiVnPd5DohD5OhU38/uNoVp\nhXNL+XEDujuMQ+YZcC01nIbxRpe6/5+NNhEMK2na5QPFeobnn0v5qgP3DY1k9yAy\nrBuxGrXdWvFdTt3Vo15dODYud0O6mXmtkymFHU9cPZAeyoCVRAEzqXrkr00Ih4s8\neiAM/ATb9OTSZkXqvAZBAQIDAQABo00wSzAJBgNVHRMEAjAAMB0GA1UdDgQWBBSJ\ng+YJd6eJvDvmsR+JdwNNgT0KpTAfBgNVHSMEGDAWgBRiNc5Tkcf0Iaxbi6ERYN4g\nl5fwVjANBgkqhkiG9w0BAQsFAAOCAQEAVr5PtgnhEUBc5TBl9Znc1EtF8u8RJLFv\nSo3OcyGTXQBd2R0Kpn2WuKuCpyYcLaMKkRZYMXSDHYS/NAU6M2jX3VRfSgR7PIqP\nymZ/ixMwbhkUcHGa0BW+3bb4YM+ZXBviylLgqx6oe0dL1mUqzvAAaI5RIaz4G5d8\ntpFxJD+do41FFl0u7KPnJGynNyo9TrpFqm70/jyFs/b/khH3f0SzE2wgRZ5zWuzV\nn/QUzwh/7E6cvLQsL1L29aB2UT+wGQrrZQiaF3bKOhbsW0ZXuDo4do9heyF9BtSl\ndT3fuEA0SIrVv/t89ZhCyjJNKcaosRmqUMzxI3aJnrwKm4tTz6z+uA==\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'g7qHPMXSeYMu5pB3Vx8K4P-j5uA',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Signing CA 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'ZS6zCaRlo1bqb5a33CPnYqptoZE',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '4102444800', # 2100-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:f5:42:56:51:24:9f:89:ab:dd:ec:a9:b5:a6:71:\n    40:ca:dd:28:df:6a:82:cc:2f:43:20:05:53:35:9f:\n    2f:40:49:88:2b:05:ea:00:e5:72:92:08:35:a1:45:\n    10:c3:71:75:d5:23:a6:a9:fd:1c:33:88:6c:88:7f:\n    39:9f:86:80:e9:de:5c:73:e4:41:a2:78:2e:90:3a:\n    86:6b:1b:a9:c4:70:1a:c7:af:e1:94:ca:20:57:d5:\n    81:38:02:03:05:dc:b6:fc:b7:19:10:f7:9c:33:33:\n    61:9e:4f:40:6b:06:67:14:3e:59:b3:4f:7f:d7:9e:\n    77:75:e2:0a:1d:d1:aa:0e:38:95:9c:f7:79:0e:88:\n    43:e4:e8:54:df:cf:ee:36:85:69:85:73:4b:f9:71:\n    03:ba:3b:8c:43:e6:19:70:2d:35:9c:86:f1:46:97:\n    ba:ff:9f:8d:36:11:0c:2b:69:da:e5:03:c5:7a:86:\n    e7:9f:4b:f9:aa:03:f7:0d:8d:64:f7:20:32:ac:1b:\n    b1:1a:b5:dd:5a:f1:5d:4e:dd:d5:a3:5e:5d:38:36:\n    2e:77:43:ba:99:79:ad:93:29:85:1d:4f:5c:3d:90:\n    1e:ca:80:95:44:01:33:a9:7a:e4:af:4d:08:87:8b:\n    3c:7a:20:0c:fc:04:db:f4:e4:d2:66:45:ea:bc:06:\n    41:01\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Client Alice 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '89:83:E6:09:77:A7:89:BC:3B:E6:B1:1F:89:77:03:4D:81:3D:0A:A5',
            },
            db_alias => {
                alias => 'alpha-alice-3',
                generation => undef,
                group_id => undef,
            },
        ),

        'alpha_bob_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Client Bob 1',
            db => {
                authority_key_identifier => '3D:6C:86:0C:00:F1:8D:E0:5E:0F:84:8B:82:DE:9B:BE:4E:C4:DD:43',
                cert_key => '6',
                data => "-----BEGIN CERTIFICATE-----\nMIIDgjCCAmqgAwIBAgIBBjANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGzAZBgNVBAMMEkFMUEhBIFNpZ25pbmcgQ0EgMTAiGA8yMDA2MDEwMTAwMDAwMFoY\nDzIwMDcwMTMxMjM1OTU5WjBbMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZIm\niZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGzAZBgNVBAMMEkFMUEhB\nIENsaWVudCBCb2IgMTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALtV\nBoWWb9VNJwAkx8VRg9ojAdMKERweHNRJ7SG9TUg6ooO6nfOgg1Gq38askGfMQSEK\nhnBO4RKmOWrUfbIdIOlO+rjSkxy/tt/yu+0kb5406tpUtiI93MlzkSbOk13DgZo+\n6N8WnoNIzvZ/MMk/l4mSU9O9S0dAn5zDHXlrutFQ5gjGTF8STY9N/4iQ6IJalyiM\n3y+x5Rk5DA8gMpusnEwivGs/VBFYeDklWEVMbCefXBqleZjOh/aF1e5P/NBG/HwQ\nq5iwJKcBtJVjR+kuOxo0jUU5M2xsnohf9uxyhmO1mIL8zcpUeX8YtfbMdxEgayXF\nCXJmHlUt8WUIHCa6JdsCAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4EFgQUic35\nncXYWF7LCoidyoOE8xDJbo0wHwYDVR0jBBgwFoAUPWyGDADxjeBeD4SLgt6bvk7E\n3UMwDQYJKoZIhvcNAQELBQADggEBADso0QmuzwttQQqRi1E37zVM42yBc/6U7HoF\nmN+Xu3finXla2HjanXcLLcOMd5ClG6G0vnhIfCgGLhu5Npzpvtc/IKgj6ffQhoXa\n7aYoYDnSYsPKdn+ng1vnlh9CrnTgLxh214S7fjtSh3PD16Ji0J65QV0GD9jsgXHR\nsZjeXTOfBQOP5mkCZJWDXv+0oYqxyJutAUW7CQbHBj7li6kS5ScjtD0JCdKtuOf+\nASZIALXMawM3U2TUcOtbXI/lD8uv4zXOSCYTLGBNx/YUuX6HCbHcMb4OYruCoJ2M\nNkKH+xP1ILV14/FbcozWriBJ6G2ifAz9dXKxUVra7sjUtPpd2uA=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => '6j5Vylh95dVTrQsF0dBa9llvBf8',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Signing CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'j6Aps6CGQftSMQM4Gjue2-3RmFw',
                notafter => '1170287999', # 2007-01-31T23:59:59
                notbefore => '1136073600', # 2006-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:bb:55:06:85:96:6f:d5:4d:27:00:24:c7:c5:51:\n    83:da:23:01:d3:0a:11:1c:1e:1c:d4:49:ed:21:bd:\n    4d:48:3a:a2:83:ba:9d:f3:a0:83:51:aa:df:c6:ac:\n    90:67:cc:41:21:0a:86:70:4e:e1:12:a6:39:6a:d4:\n    7d:b2:1d:20:e9:4e:fa:b8:d2:93:1c:bf:b6:df:f2:\n    bb:ed:24:6f:9e:34:ea:da:54:b6:22:3d:dc:c9:73:\n    91:26:ce:93:5d:c3:81:9a:3e:e8:df:16:9e:83:48:\n    ce:f6:7f:30:c9:3f:97:89:92:53:d3:bd:4b:47:40:\n    9f:9c:c3:1d:79:6b:ba:d1:50:e6:08:c6:4c:5f:12:\n    4d:8f:4d:ff:88:90:e8:82:5a:97:28:8c:df:2f:b1:\n    e5:19:39:0c:0f:20:32:9b:ac:9c:4c:22:bc:6b:3f:\n    54:11:58:78:39:25:58:45:4c:6c:27:9f:5c:1a:a5:\n    79:98:ce:87:f6:85:d5:ee:4f:fc:d0:46:fc:7c:10:\n    ab:98:b0:24:a7:01:b4:95:63:47:e9:2e:3b:1a:34:\n    8d:45:39:33:6c:6c:9e:88:5f:f6:ec:72:86:63:b5:\n    98:82:fc:cd:ca:54:79:7f:18:b5:f6:cc:77:11:20:\n    6b:25:c5:09:72:66:1e:55:2d:f1:65:08:1c:26:ba:\n    25:db\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Client Bob 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '89:CD:F9:9D:C5:D8:58:5E:CB:0A:88:9D:CA:83:84:F3:10:C9:6E:8D',
            },
            db_alias => {
                alias => 'alpha-bob-1',
                generation => undef,
                group_id => undef,
            },
        ),

        'alpha_bob_2' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Client Bob 2',
            db => {
                authority_key_identifier => 'A8:BE:4B:1B:52:18:61:F7:DF:06:5A:5E:BD:C5:6B:E8:18:28:3F:ED',
                cert_key => '12',
                data => "-----BEGIN CERTIFICATE-----\nMIIDgjCCAmqgAwIBAgIBDDANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGzAZBgNVBAMMEkFMUEhBIFNpZ25pbmcgQ0EgMjAiGA8yMDA3MDEwMTAwMDAwMFoY\nDzIxMDAwMTMxMjM1OTU5WjBbMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZIm\niZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGzAZBgNVBAMMEkFMUEhB\nIENsaWVudCBCb2IgMjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAPld\nzkXRELECjmZgnaOXRRP9ZVexSHWW/EI83Ud15+v/Rv7ErNdsL8ipEX5cBJGZpKMS\nSyppPC/7YSOXpK21mGv75Msbh2vUqr2i0nclMtvSz3DewY47h+mZ60lqlY/kB2WY\nLc8LCFHKsZHvs3uJoljUsrEbQqk9SzujVjUOf0zS1OeTRa1g0QBiHlLJGVUC9K7M\nBn/vmbvXU0xka03LDlqcD3Lj2izeosmnqI9PoDXPUDAgV0wWy5rUzeRlTcfJUHZf\nxQAc/nLAXW23jaVXTEpZ9lOpm7G94y6beepxsz+0A/umUTB3u95r9STreQeS5+Vs\nI0G/TngbrHuqW3kok/ECAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4EFgQUAHiz\n8QxLWxFFv2CxI4DUazmBoQUwHwYDVR0jBBgwFoAUqL5LG1IYYfffBlpevcVr6Bgo\nP+0wDQYJKoZIhvcNAQELBQADggEBABC2HlKuIYVJxNL8KPrflDOzZ3x6Z1YD6/jf\nvU6BggWNctZekgGOiXeDOlxegkIfb2Y5gN/sMOfrbiAzcHvie7B7zfIYUeKk2wTd\nkSBRurArqAWCQDddkGUCu34kanu1B4UEWUABn6Yf2I+X6Y5eaCFmcMKG+BqtBJnF\ncQWxeXug8Pt8Od1iQpZoyywT6GKlsIdXakzG69HJYL1z0LAuPaoqWED/dFCKUBmR\nJngun7EOy+cfdCZcYF6tGinfgaXe3vd9mjpRyzIc/rvZ05eYr3oM7MXHwBiClkj7\n074QhF3XG08/st5ird0CWkERP4AmPD9QyzaT0g4orUhHxCN4gX8=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'dW0Y-RUzNESiZDiUEU1VQC16US0',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Signing CA 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'RE35XR3XIBXiIbAu8P5aGMCmH7o',
                notafter => '4105123199', # 2100-01-31T23:59:59
                notbefore => '1167609600', # 2007-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:f9:5d:ce:45:d1:10:b1:02:8e:66:60:9d:a3:97:\n    45:13:fd:65:57:b1:48:75:96:fc:42:3c:dd:47:75:\n    e7:eb:ff:46:fe:c4:ac:d7:6c:2f:c8:a9:11:7e:5c:\n    04:91:99:a4:a3:12:4b:2a:69:3c:2f:fb:61:23:97:\n    a4:ad:b5:98:6b:fb:e4:cb:1b:87:6b:d4:aa:bd:a2:\n    d2:77:25:32:db:d2:cf:70:de:c1:8e:3b:87:e9:99:\n    eb:49:6a:95:8f:e4:07:65:98:2d:cf:0b:08:51:ca:\n    b1:91:ef:b3:7b:89:a2:58:d4:b2:b1:1b:42:a9:3d:\n    4b:3b:a3:56:35:0e:7f:4c:d2:d4:e7:93:45:ad:60:\n    d1:00:62:1e:52:c9:19:55:02:f4:ae:cc:06:7f:ef:\n    99:bb:d7:53:4c:64:6b:4d:cb:0e:5a:9c:0f:72:e3:\n    da:2c:de:a2:c9:a7:a8:8f:4f:a0:35:cf:50:30:20:\n    57:4c:16:cb:9a:d4:cd:e4:65:4d:c7:c9:50:76:5f:\n    c5:00:1c:fe:72:c0:5d:6d:b7:8d:a5:57:4c:4a:59:\n    f6:53:a9:9b:b1:bd:e3:2e:9b:79:ea:71:b3:3f:b4:\n    03:fb:a6:51:30:77:bb:de:6b:f5:24:eb:79:07:92:\n    e7:e5:6c:23:41:bf:4e:78:1b:ac:7b:aa:5b:79:28:\n    93:f1\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Client Bob 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '00:78:B3:F1:0C:4B:5B:11:45:BF:60:B1:23:80:D4:6B:39:81:A1:05',
            },
            db_alias => {
                alias => 'alpha-bob-2',
                generation => undef,
                group_id => undef,
            },
        ),

        'alpha_bob_3' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Client Bob 3',
            db => {
                authority_key_identifier => '62:35:CE:53:91:C7:F4:21:AC:5B:8B:A1:11:60:DE:20:97:97:F0:56',
                cert_key => '20',
                data => "-----BEGIN CERTIFICATE-----\nMIIDgjCCAmqgAwIBAgIBFDANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGzAZBgNVBAMMEkFMUEhBIFNpZ25pbmcgQ0EgMzAiGA8yMTAwMDEwMTAwMDAwMFoY\nDzIxMDUwMTMxMjM1OTU5WjBbMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZIm\niZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGzAZBgNVBAMMEkFMUEhB\nIENsaWVudCBCb2IgMzCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAJ2J\nDb69lM3IAGI7iIzD0pitqFciUbGKhNn9qq1xI91EhTrMtdQW7Z6MOsCtNNazhahh\njci6wMyiB1oEC36LmSXs2Zpnjo6sJeKCnfXEFq0t+lPYT9VopmLUiNhb7KulMC3K\nbqNzecagrZq+Wze282rzUx4d85CHCwwO4rZ06O5VP7smI8wd9Mtim+MngNhNJU38\nyk3C2sGQyb6AzlMaJIZ6+H076a2O2UkLrx274SwRTtTXcq/55s2+WalhEBjkcQO7\nyA3SZTpYHxyMf7uLP/Vu4yet8lN7OYNplpZCXF2X+sRdKvHJh8RZ7HT0xIi/7+Fb\nexsCZ0vYXbGOZxdxZqcCAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4EFgQULW7T\nFAC+omleH4idoHKkSNCBum4wHwYDVR0jBBgwFoAUYjXOU5HH9CGsW4uhEWDeIJeX\n8FYwDQYJKoZIhvcNAQELBQADggEBAJIoxs5q0gX1NsZgJ9R6npR8kJfgjQKucDd9\nS+Vi0q56bR/8aU/mWrM5o2bh28CuqraJPbFLHs0eOiYAL1XDMw0P3lexXPtxjp+E\nMZg4QEgT4RiQMXBOGE6NmlqsBxYs0t+dNQflTeGCYNJ0hYwUYh+ZI1qnJZrr0hU/\nUc7qCGmQXnqspW0XOsi5xYCuIsUkNrgjtAbhMSGddqDMWU97gAjEwXRtDeNnDPbr\nB52ItkUfWdk7W0N1VTfG4+g6KgCYHU9jF8JBOuYQN4tsfrNO4AnLAZv0OfT0c3Dr\nKciPNpypk2OhqWGA7WVEJsw1CRAY3KmhZQuLvpOWDbSpvnBDkFw=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'HQHHdUdNoP0NZIS-ZgFLM6NitEA',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Signing CA 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'ZS6zCaRlo1bqb5a33CPnYqptoZE',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '4102444800', # 2100-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:9d:89:0d:be:bd:94:cd:c8:00:62:3b:88:8c:c3:\n    d2:98:ad:a8:57:22:51:b1:8a:84:d9:fd:aa:ad:71:\n    23:dd:44:85:3a:cc:b5:d4:16:ed:9e:8c:3a:c0:ad:\n    34:d6:b3:85:a8:61:8d:c8:ba:c0:cc:a2:07:5a:04:\n    0b:7e:8b:99:25:ec:d9:9a:67:8e:8e:ac:25:e2:82:\n    9d:f5:c4:16:ad:2d:fa:53:d8:4f:d5:68:a6:62:d4:\n    88:d8:5b:ec:ab:a5:30:2d:ca:6e:a3:73:79:c6:a0:\n    ad:9a:be:5b:37:b6:f3:6a:f3:53:1e:1d:f3:90:87:\n    0b:0c:0e:e2:b6:74:e8:ee:55:3f:bb:26:23:cc:1d:\n    f4:cb:62:9b:e3:27:80:d8:4d:25:4d:fc:ca:4d:c2:\n    da:c1:90:c9:be:80:ce:53:1a:24:86:7a:f8:7d:3b:\n    e9:ad:8e:d9:49:0b:af:1d:bb:e1:2c:11:4e:d4:d7:\n    72:af:f9:e6:cd:be:59:a9:61:10:18:e4:71:03:bb:\n    c8:0d:d2:65:3a:58:1f:1c:8c:7f:bb:8b:3f:f5:6e:\n    e3:27:ad:f2:53:7b:39:83:69:96:96:42:5c:5d:97:\n    fa:c4:5d:2a:f1:c9:87:c4:59:ec:74:f4:c4:88:bf:\n    ef:e1:5b:7b:1b:02:67:4b:d8:5d:b1:8e:67:17:71:\n    66:a7\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Client Bob 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '2D:6E:D3:14:00:BE:A2:69:5E:1F:88:9D:A0:72:A4:48:D0:81:BA:6E',
            },
            db_alias => {
                alias => 'alpha-bob-3',
                generation => undef,
                group_id => undef,
            },
        ),

        'alpha_christine_2' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Client Christine 2',
            db => {
                authority_key_identifier => 'A8:BE:4B:1B:52:18:61:F7:DF:06:5A:5E:BD:C5:6B:E8:18:28:3F:ED',
                cert_key => '13',
                data => "-----BEGIN CERTIFICATE-----\nMIIDiDCCAnCgAwIBAgIBDTANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGzAZBgNVBAMMEkFMUEhBIFNpZ25pbmcgQ0EgMjAiGA8yMDA3MDEwMTAwMDAwMFoY\nDzIxMDAwMTMxMjM1OTU5WjBhMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZIm\niZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxITAfBgNVBAMMGEFMUEhB\nIENsaWVudCBDaHJpc3RpbmUgMjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoC\nggEBAKGws35atz8SMvRa8hZeaE3DpcT/DWvjGnYvn5Yh8+oVRyOfZYmrFhvWdQDO\nqvdhZh1RmqYLguaMyyqpL5ezM2/1S7Ot3G4RRSaPmO+G596URj7vDSYN+IgZyG8K\nWhkuM5z1cKfKxWwAJDKiXNlsuK+fPJVpr7FHaSM5QkPAyqlyxxLOMoN6Ho7Sn6sV\n/SsozK71eEfjCBVC/+Dr9gzzSSfIzMM3C0dSdcRVuebkvR10QGGYODAtCp5Iothg\nw1HFsUb9YhW99vWV4B0U0LicOe+AcGN1ogfbyqlEdCIeveJ9HL6ZJvoLi9jhpJgc\n5f91bkeg825v/KEOkffcJ0QD/CECAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4E\nFgQUnPVbHw48H93ifZJpDFC7QwFQjQAwHwYDVR0jBBgwFoAUqL5LG1IYYfffBlpe\nvcVr6BgoP+0wDQYJKoZIhvcNAQELBQADggEBAIckQ7Dc2942x9ZSjVLh/h2jECI0\nplK79UPIPAmrJQwsRx2biVgabn3tP3f+YZn1OtrWwncEB6Lq6joNHgubVraXhjgV\neD3zxQxgZF0Oysiqin7h3gCaXDvKzTk1BPwYBbN9C0PyLGPzjK/hmEYJIcxr9E/J\nnMbP6GTVXW9tscthaEXxLvEhGz2B+S81M8quTSg5mTEWro+AVuwODu5PWAuGSMus\nuv31ImdveS2xRlx/ci1QOp7mb273FKN03PW0Uh7DHRywqxmKCwQWxFqBthzSzLLg\ny2+YoLdKl/SI8LdalhQo+v7BkhrAk6f4ApIxEqeFGCfr4BUP+dCZZOHcfRA=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'ox-qaWaWfSlTWls0gBw40FkuZm0',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Signing CA 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'RE35XR3XIBXiIbAu8P5aGMCmH7o',
                notafter => '4105123199', # 2100-01-31T23:59:59
                notbefore => '1167609600', # 2007-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:a1:b0:b3:7e:5a:b7:3f:12:32:f4:5a:f2:16:5e:\n    68:4d:c3:a5:c4:ff:0d:6b:e3:1a:76:2f:9f:96:21:\n    f3:ea:15:47:23:9f:65:89:ab:16:1b:d6:75:00:ce:\n    aa:f7:61:66:1d:51:9a:a6:0b:82:e6:8c:cb:2a:a9:\n    2f:97:b3:33:6f:f5:4b:b3:ad:dc:6e:11:45:26:8f:\n    98:ef:86:e7:de:94:46:3e:ef:0d:26:0d:f8:88:19:\n    c8:6f:0a:5a:19:2e:33:9c:f5:70:a7:ca:c5:6c:00:\n    24:32:a2:5c:d9:6c:b8:af:9f:3c:95:69:af:b1:47:\n    69:23:39:42:43:c0:ca:a9:72:c7:12:ce:32:83:7a:\n    1e:8e:d2:9f:ab:15:fd:2b:28:cc:ae:f5:78:47:e3:\n    08:15:42:ff:e0:eb:f6:0c:f3:49:27:c8:cc:c3:37:\n    0b:47:52:75:c4:55:b9:e6:e4:bd:1d:74:40:61:98:\n    38:30:2d:0a:9e:48:a2:d8:60:c3:51:c5:b1:46:fd:\n    62:15:bd:f6:f5:95:e0:1d:14:d0:b8:9c:39:ef:80:\n    70:63:75:a2:07:db:ca:a9:44:74:22:1e:bd:e2:7d:\n    1c:be:99:26:fa:0b:8b:d8:e1:a4:98:1c:e5:ff:75:\n    6e:47:a0:f3:6e:6f:fc:a1:0e:91:f7:dc:27:44:03:\n    fc:21\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Client Christine 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '9C:F5:5B:1F:0E:3C:1F:DD:E2:7D:92:69:0C:50:BB:43:01:50:8D:00',
            },
            db_alias => {
                alias => 'alpha-christine-2',
                generation => undef,
                group_id => undef,
            },
        ),

        'alpha_datavault_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA DataVault 1',
            db => {
                authority_key_identifier => 'D7:95:E4:E9:E4:36:9F:28:AB:54:85:DE:E0:1A:32:3D:33:62:09:F8',
                cert_key => '1',
                data => "-----BEGIN CERTIFICATE-----\nMIIDojCCAoqgAwIBAgIBATANBgkqhkiG9w0BAQsFADBaMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGjAYBgNVBAMMEUFMUEhBIERhdGFWYXVsdCAxMCIYDzIwMDYwMTAxMDAwMDAwWhgP\nMjAwNzAxMzEyMzU5NTlaMFoxEzARBgoJkiaJk/IsZAEZFgNPUkcxGDAWBgoJkiaJ\nk/IsZAEZFghPcGVuWFBLSTENMAsGA1UECwwEQUNNRTEaMBgGA1UEAwwRQUxQSEEg\nRGF0YVZhdWx0IDEwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCzTXkz\n4vU2iRzDDx8WbVTsUHpigSlWFqVyzK4QiJH+m5EgS1I4GafCEeC84JhAKGwtQT+G\nZ7Tvcayl5iGpI8namrBq+oNVG9/DCCR08R6d/94jfsbTSwvIHwbKw7YGkDTmpDS0\nXaOf3ltFzvh3S9LPuzBJFxYjHAICsNVwQ30HzKGVI07m15Dbeu6h2iwh+Mab0LNY\nBssdUuo2nsfIeSTGyyRODNx39sT7mLioYX+/VcvafBZgR5Fot55/toN30Lb883Tf\nkQLbUFAG6x/fecw3hG2wz4HUHNepKSNo4nOEwIppnOZvp+fGHZ/goKv3Qz/gokpV\npkGYojhdzk9YzNnlAgMBAAGjbzBtMAkGA1UdEwQCMAAwHQYDVR0OBBYEFNeV5Onk\nNp8oq1SF3uAaMj0zYgn4MB8GA1UdIwQYMBaAFNeV5OnkNp8oq1SF3uAaMj0zYgn4\nMAsGA1UdDwQEAwIFIDATBgNVHSUEDDAKBggrBgEFBQcDBDANBgkqhkiG9w0BAQsF\nAAOCAQEAZgH3mCgK9jeQH9dmE5XQFQd/fMqOdG3Cql/Y0NNeBAw0tlo8dYwHDz+q\n1kNYgG7mI9TtZ79bzvdXt0pLUyx77zMtV2YgpOt1BG3KW2CxfTzQ0ScdHgPGuVYU\nfkKAYiegXhN0IHbdBrKduvFO89ba/fXN1vFdbiprHTblzInp5402WkfBgNH3mkeW\nzVaCBMcGkXcp9fWaLWOq7IHw4UElPcbvthwpSRm/NsfIGNog+9DYFs+n+6FUvkqu\nOoqBkD+mweHtKE1gtJE7cehboW6S5n48OUP9Myvmf5uoXD/pcfK0d/+CfLcybHHz\n5dmpHv7EPP79rcciNeN946fWFu49/A==\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'jABJnMvK7ZWARV5DdmblpcxQmFo',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA DataVault 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'jABJnMvK7ZWARV5DdmblpcxQmFo',
                notafter => '1170287999', # 2007-01-31T23:59:59
                notbefore => '1136073600', # 2006-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:b3:4d:79:33:e2:f5:36:89:1c:c3:0f:1f:16:6d:\n    54:ec:50:7a:62:81:29:56:16:a5:72:cc:ae:10:88:\n    91:fe:9b:91:20:4b:52:38:19:a7:c2:11:e0:bc:e0:\n    98:40:28:6c:2d:41:3f:86:67:b4:ef:71:ac:a5:e6:\n    21:a9:23:c9:da:9a:b0:6a:fa:83:55:1b:df:c3:08:\n    24:74:f1:1e:9d:ff:de:23:7e:c6:d3:4b:0b:c8:1f:\n    06:ca:c3:b6:06:90:34:e6:a4:34:b4:5d:a3:9f:de:\n    5b:45:ce:f8:77:4b:d2:cf:bb:30:49:17:16:23:1c:\n    02:02:b0:d5:70:43:7d:07:cc:a1:95:23:4e:e6:d7:\n    90:db:7a:ee:a1:da:2c:21:f8:c6:9b:d0:b3:58:06:\n    cb:1d:52:ea:36:9e:c7:c8:79:24:c6:cb:24:4e:0c:\n    dc:77:f6:c4:fb:98:b8:a8:61:7f:bf:55:cb:da:7c:\n    16:60:47:91:68:b7:9e:7f:b6:83:77:d0:b6:fc:f3:\n    74:df:91:02:db:50:50:06:eb:1f:df:79:cc:37:84:\n    6d:b0:cf:81:d4:1c:d7:a9:29:23:68:e2:73:84:c0:\n    8a:69:9c:e6:6f:a7:e7:c6:1d:9f:e0:a0:ab:f7:43:\n    3f:e0:a2:4a:55:a6:41:98:a2:38:5d:ce:4f:58:cc:\n    d9:e5\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA DataVault 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'D7:95:E4:E9:E4:36:9F:28:AB:54:85:DE:E0:1A:32:3D:33:62:09:F8',
            },
            db_alias => {
                alias => 'alpha-datavault-1',
                generation => '1',
                group_id => 'alpha-datavault',
            },
        ),

        'alpha_datavault_2' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA DataVault 2',
            db => {
                authority_key_identifier => 'DB:EB:92:11:24:C1:D5:A9:D6:2C:54:58:57:4D:E4:E3:45:48:1E:4D',
                cert_key => '7',
                data => "-----BEGIN CERTIFICATE-----\nMIIDojCCAoqgAwIBAgIBBzANBgkqhkiG9w0BAQsFADBaMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGjAYBgNVBAMMEUFMUEhBIERhdGFWYXVsdCAyMCIYDzIwMDcwMTAxMDAwMDAwWhgP\nMjEwMDAxMzEyMzU5NTlaMFoxEzARBgoJkiaJk/IsZAEZFgNPUkcxGDAWBgoJkiaJ\nk/IsZAEZFghPcGVuWFBLSTENMAsGA1UECwwEQUNNRTEaMBgGA1UEAwwRQUxQSEEg\nRGF0YVZhdWx0IDIwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCmRv2T\n46RslNjiGRJulCltMHqDGwEuvLRlYu9UzNpOrljuvSTOl3BP2H5iVjUz+zksD1m6\nh3jzhOF4Y5pYg2F4vgQAjYYwpVDjuiIqKnrZ/so2y8V1Hfctpbu1hUIK/vA4D0KX\nI827+l3D+B6a9fZ/2dD2ZtB9lYhxf3R81kDVkO3DvufVBvhBtdP5wdgL7IQ0qTdr\ngAHCteK4uzqFWrcnTMGhutieJvSKdyquN7D34sDWv8zPPxHa25NOOJGtfeIoYIL3\nuNa00w1927aL3t2hcHs96KswxS1KcCtaiu6xYBgpd+vdqtkBcLMs02bhD1Hu5080\n6auru/rNxAGg8EVfAgMBAAGjbzBtMAkGA1UdEwQCMAAwHQYDVR0OBBYEFNvrkhEk\nwdWp1ixUWFdN5ONFSB5NMB8GA1UdIwQYMBaAFNvrkhEkwdWp1ixUWFdN5ONFSB5N\nMAsGA1UdDwQEAwIFIDATBgNVHSUEDDAKBggrBgEFBQcDBDANBgkqhkiG9w0BAQsF\nAAOCAQEAO4w2BTO5AInoI7m2AIT3Kj3nUtoj9nUcBD8XwjUyJnrxCPjj6Ub4x1zT\ntDNIp96nWNyBCfq3PZE5evjoXxyNnGxBMAxxIMew7dKqMF0zTWROwqbzZqlVFbeJ\nkw/8FEzZ/YvNsaZe1dze2gFwu9R1aNKdfpw2GJ32AVGuYmUOl39+41pdI4Z5BzGk\n4onVF+MiYKZA5M4ZQGONDPG0MZFfnD65Coc7fLWmzDyuDWCC9eB0zwcl4g77NR1g\nJta/z7T+U8gq7U6ZIunz8FcEtDvBNA7xO/olVv9qRXC105O0NY5x/g9pmB87WW6w\nSskNeP8X64TtTBrCzLJe+jEMZUTfnQ==\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'KmcuOMkK7hJ3GDFi18WQW5h6mOw',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA DataVault 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'KmcuOMkK7hJ3GDFi18WQW5h6mOw',
                notafter => '4105123199', # 2100-01-31T23:59:59
                notbefore => '1167609600', # 2007-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:a6:46:fd:93:e3:a4:6c:94:d8:e2:19:12:6e:94:\n    29:6d:30:7a:83:1b:01:2e:bc:b4:65:62:ef:54:cc:\n    da:4e:ae:58:ee:bd:24:ce:97:70:4f:d8:7e:62:56:\n    35:33:fb:39:2c:0f:59:ba:87:78:f3:84:e1:78:63:\n    9a:58:83:61:78:be:04:00:8d:86:30:a5:50:e3:ba:\n    22:2a:2a:7a:d9:fe:ca:36:cb:c5:75:1d:f7:2d:a5:\n    bb:b5:85:42:0a:fe:f0:38:0f:42:97:23:cd:bb:fa:\n    5d:c3:f8:1e:9a:f5:f6:7f:d9:d0:f6:66:d0:7d:95:\n    88:71:7f:74:7c:d6:40:d5:90:ed:c3:be:e7:d5:06:\n    f8:41:b5:d3:f9:c1:d8:0b:ec:84:34:a9:37:6b:80:\n    01:c2:b5:e2:b8:bb:3a:85:5a:b7:27:4c:c1:a1:ba:\n    d8:9e:26:f4:8a:77:2a:ae:37:b0:f7:e2:c0:d6:bf:\n    cc:cf:3f:11:da:db:93:4e:38:91:ad:7d:e2:28:60:\n    82:f7:b8:d6:b4:d3:0d:7d:db:b6:8b:de:dd:a1:70:\n    7b:3d:e8:ab:30:c5:2d:4a:70:2b:5a:8a:ee:b1:60:\n    18:29:77:eb:dd:aa:d9:01:70:b3:2c:d3:66:e1:0f:\n    51:ee:e7:4f:34:e9:ab:ab:bb:fa:cd:c4:01:a0:f0:\n    45:5f\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA DataVault 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'DB:EB:92:11:24:C1:D5:A9:D6:2C:54:58:57:4D:E4:E3:45:48:1E:4D',
            },
            db_alias => {
                alias => 'alpha-datavault-2',
                generation => '2',
                group_id => 'alpha-datavault',
            },
        ),

        'alpha_datavault_3' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA DataVault 3',
            db => {
                authority_key_identifier => 'A5:8D:66:13:C9:2D:01:CF:95:DE:94:84:AB:C3:83:89:F5:DB:F8:3F',
                cert_key => '15',
                data => "-----BEGIN CERTIFICATE-----\nMIIDojCCAoqgAwIBAgIBDzANBgkqhkiG9w0BAQsFADBaMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGjAYBgNVBAMMEUFMUEhBIERhdGFWYXVsdCAzMCIYDzIxMDAwMTAxMDAwMDAwWhgP\nMjEwNTAxMzEyMzU5NTlaMFoxEzARBgoJkiaJk/IsZAEZFgNPUkcxGDAWBgoJkiaJ\nk/IsZAEZFghPcGVuWFBLSTENMAsGA1UECwwEQUNNRTEaMBgGA1UEAwwRQUxQSEEg\nRGF0YVZhdWx0IDMwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDU5RUL\nQwOvs1HNjExSJikpGVCV6bFsR2//LONZGkQH92Ep6/t+iTZV+kbRay5aEV84Jm0A\nZpVaq72eXMdyNA5uslPXAILZWHU0WBmLZv2GE5KfYQmL/cBT6fvomCnwVsxmCOqe\n/SedJ8ssCCxuLEFbPvKpUN7UGQeocIzWGC2mCKcX6dPRoOSBJUzUVZykl0RPSdZe\nsy6YzFr2t97Nrrz1ku2LvVjONQQ1ZTZX5aeqqCUkCrQgOBurt2Fhb9D0A0wmwVBT\nst99yb+pG1bp8a5hNGOBqNpUCBNqx0ewG++XbTXa0bdZyS752y04FClV+iXeDj5+\nWGrC6eT2zB95tjANAgMBAAGjbzBtMAkGA1UdEwQCMAAwHQYDVR0OBBYEFKWNZhPJ\nLQHPld6UhKvDg4n12/g/MB8GA1UdIwQYMBaAFKWNZhPJLQHPld6UhKvDg4n12/g/\nMAsGA1UdDwQEAwIFIDATBgNVHSUEDDAKBggrBgEFBQcDBDANBgkqhkiG9w0BAQsF\nAAOCAQEAtIXVX/0xsTdmW0JgsUCsxLmZ54tQ4mFbE4PXSq1Pg18lG2du/WfxrdvW\ndU2cOuV7Xd1bCYlo7Pd5qD8c5KruAJWhPgOPjO22oZO9a85dXCta+6NgZz8FKxMq\nzM7WUPzL8KCbtc1+EU1+phulG1uGXzw6NiZwMVbGgFEcbLrwECsO2LlUII/cZ2Fc\nRj0AAQ+KCPhUmlh6f3gjifOCPegLWQSXL1TV+RB4rpxm03sBRo2BoOdBG3GxSsLd\ndBoZjhc6uFKZq2PzMr3IPJUJPWvSbALnfkjLnsmJ6RBTvCJ7GIPb0A1HTdzKlUzF\n1/Q6uW+bAGzJLaX9fPngJ6CO155//A==\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'dI33yoj49iHiHs5X4hJW11fGDo8',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA DataVault 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'dI33yoj49iHiHs5X4hJW11fGDo8',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '4102444800', # 2100-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:d4:e5:15:0b:43:03:af:b3:51:cd:8c:4c:52:26:\n    29:29:19:50:95:e9:b1:6c:47:6f:ff:2c:e3:59:1a:\n    44:07:f7:61:29:eb:fb:7e:89:36:55:fa:46:d1:6b:\n    2e:5a:11:5f:38:26:6d:00:66:95:5a:ab:bd:9e:5c:\n    c7:72:34:0e:6e:b2:53:d7:00:82:d9:58:75:34:58:\n    19:8b:66:fd:86:13:92:9f:61:09:8b:fd:c0:53:e9:\n    fb:e8:98:29:f0:56:cc:66:08:ea:9e:fd:27:9d:27:\n    cb:2c:08:2c:6e:2c:41:5b:3e:f2:a9:50:de:d4:19:\n    07:a8:70:8c:d6:18:2d:a6:08:a7:17:e9:d3:d1:a0:\n    e4:81:25:4c:d4:55:9c:a4:97:44:4f:49:d6:5e:b3:\n    2e:98:cc:5a:f6:b7:de:cd:ae:bc:f5:92:ed:8b:bd:\n    58:ce:35:04:35:65:36:57:e5:a7:aa:a8:25:24:0a:\n    b4:20:38:1b:ab:b7:61:61:6f:d0:f4:03:4c:26:c1:\n    50:53:b2:df:7d:c9:bf:a9:1b:56:e9:f1:ae:61:34:\n    63:81:a8:da:54:08:13:6a:c7:47:b0:1b:ef:97:6d:\n    35:da:d1:b7:59:c9:2e:f9:db:2d:38:14:29:55:fa:\n    25:de:0e:3e:7e:58:6a:c2:e9:e4:f6:cc:1f:79:b6:\n    30:0d\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA DataVault 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'A5:8D:66:13:C9:2D:01:CF:95:DE:94:84:AB:C3:83:89:F5:DB:F8:3F',
            },
            db_alias => {
                alias => 'alpha-datavault-3',
                generation => '3',
                group_id => 'alpha-datavault',
            },
        ),

        'alpha_don_2' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Client Don 2',
            db => {
                authority_key_identifier => 'A8:BE:4B:1B:52:18:61:F7:DF:06:5A:5E:BD:C5:6B:E8:18:28:3F:ED',
                cert_key => '14',
                data => "-----BEGIN CERTIFICATE-----\nMIIDgjCCAmqgAwIBAgIBDjANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGzAZBgNVBAMMEkFMUEhBIFNpZ25pbmcgQ0EgMjAiGA8yMDA3MDEwMTAwMDAwMFoY\nDzIxMDAwMTMxMjM1OTU5WjBbMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZIm\niZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGzAZBgNVBAMMEkFMUEhB\nIENsaWVudCBEb24gMjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBANzz\nyOF1PV4jYrnocNXnTqv9I26Z4/Hye2JGtZTI43QDIo5xiZUIjiZwJR5XUHLjsIRq\nEUvpkLK100G7qIU8zwjSxOJiFOX3ELuwxpt0Ppse/nofFfdJeV9jMywEibbu0Res\nDSSFeAdHpq3XmnvcyY0MQZkC3Bt1C9RS4Rkxav3R7s36evuLLFxKJSaOBZrHUuLl\nt8KvQo/mFZb3cKIpYXcJtBOWRUCAoajh4FapT/hQ2F1ycsUY/1FXBGoU1icDd1ZD\nUBczNFGS6aSzL7yXH9C1pA8vzPy67kmJpTgMlOfbt0VQrqI7cPROiYGIlbYzq4t0\nN7TK2rXiGAVsQEZE75sCAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4EFgQUwjKM\n4VG2Wx6iAC7F3CcVNIgVlYkwHwYDVR0jBBgwFoAUqL5LG1IYYfffBlpevcVr6Bgo\nP+0wDQYJKoZIhvcNAQELBQADggEBAD0zAXRHw+Em8J7Rmivllnme8DK+74tm/lzV\n0mrrEZosg2xWvxF66gvYXAqAEvgBA4GR+SQvbeNJuVZQQmd7Hl63XNe3nXRaSwFe\n1oT5hDPWfauXNbqihTVLB1iL7gEM4BxhabvH56tJSdbZqFdbwo5zOsHT8w6C5YBX\nj2HLaKKGB30Rak9Z2jQKU6aT76P6DGibZ+7deQyAP1tjCyvGPauSKC8eQKCbVcMx\nGyUb9d9dkqCYiMGzVcGFJt54Gu90m0gO/umFEQyI926baznyK/nsuMbyZvZU7wMn\nta2lCG+43DVsx8UxJI5xdRBEWkQ09rwadETsdL8qwnPZ6jtyv+U=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => '5wqoFM_Os44z-thFWV44TK0Bvxc',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Signing CA 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'RE35XR3XIBXiIbAu8P5aGMCmH7o',
                notafter => '4105123199', # 2100-01-31T23:59:59
                notbefore => '1167609600', # 2007-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:dc:f3:c8:e1:75:3d:5e:23:62:b9:e8:70:d5:e7:\n    4e:ab:fd:23:6e:99:e3:f1:f2:7b:62:46:b5:94:c8:\n    e3:74:03:22:8e:71:89:95:08:8e:26:70:25:1e:57:\n    50:72:e3:b0:84:6a:11:4b:e9:90:b2:b5:d3:41:bb:\n    a8:85:3c:cf:08:d2:c4:e2:62:14:e5:f7:10:bb:b0:\n    c6:9b:74:3e:9b:1e:fe:7a:1f:15:f7:49:79:5f:63:\n    33:2c:04:89:b6:ee:d1:17:ac:0d:24:85:78:07:47:\n    a6:ad:d7:9a:7b:dc:c9:8d:0c:41:99:02:dc:1b:75:\n    0b:d4:52:e1:19:31:6a:fd:d1:ee:cd:fa:7a:fb:8b:\n    2c:5c:4a:25:26:8e:05:9a:c7:52:e2:e5:b7:c2:af:\n    42:8f:e6:15:96:f7:70:a2:29:61:77:09:b4:13:96:\n    45:40:80:a1:a8:e1:e0:56:a9:4f:f8:50:d8:5d:72:\n    72:c5:18:ff:51:57:04:6a:14:d6:27:03:77:56:43:\n    50:17:33:34:51:92:e9:a4:b3:2f:bc:97:1f:d0:b5:\n    a4:0f:2f:cc:fc:ba:ee:49:89:a5:38:0c:94:e7:db:\n    b7:45:50:ae:a2:3b:70:f4:4e:89:81:88:95:b6:33:\n    ab:8b:74:37:b4:ca:da:b5:e2:18:05:6c:40:46:44:\n    ef:9b\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Client Don 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'C2:32:8C:E1:51:B6:5B:1E:A2:00:2E:C5:DC:27:15:34:88:15:95:89',
            },
            db_alias => {
                alias => 'alpha-don-2',
                generation => undef,
                group_id => undef,
            },
        ),

        'alpha_scep_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA SCEP 1',
            db => {
                authority_key_identifier => '7B:8F:B0:2C:F9:51:1D:25:CD:51:1F:DF:6D:93:8C:A9:4F:D9:00:74',
                cert_key => '4',
                data => "-----BEGIN CERTIFICATE-----\nMIIDeTCCAmGgAwIBAgIBBDANBgkqhkiG9w0BAQsFADBYMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGDAWBgNVBAMMD0FMUEhBIFJvb3QgQ0EgMTAiGA8yMDA2MDEwMTAwMDAwMFoYDzIw\nMDcwMTMxMjM1OTU5WjBVMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZImiZPy\nLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxFTATBgNVBAMMDEFMUEhBIFND\nRVAgMTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAPFotHSmue+l2THK\nVi2hrDpLgzyMxOSwIY605BHWr7RMuBZq+y1hLhuIUPy3dHLMyjmwbW0+FNUTHb0K\nhI2xEOt+3UjRurEaN9pWzVyIjQT3i7wzpRdFWWKSi0bsHZ3Y62IAwzsSvwmVLs1y\n4As8ByfozPk/8NPoPHpCtYWpIoDZc07GkPidjiMhoIWlZObM+rEfXBWjlJ6EZ/TB\nhv1o0gGBUWkfb8zD4KkRF2e9nCTPj4VbEHvCTmZ5AfWHiWQO9bGs6CgXthz3st0H\n66EfRlwiAcJop7blHrh+YwscOvDYX5r/oAuH9x7unzf+JM3BUCypXBZiyVo04xlf\nF+4Tfi0CAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4EFgQUsJnmToulK95a+iDP\nb85V01jDTpIwHwYDVR0jBBgwFoAUe4+wLPlRHSXNUR/fbZOMqU/ZAHQwDQYJKoZI\nhvcNAQELBQADggEBAG1QWJdQOaa0tcdHVqXovqhQS5OCkFGgqiqoN7ntahUQN2qr\nm0PZk+qWt/gKdSMdoLNZtrJFV+paN9lL7qCYqbOWoL3iefXZ2fO3a71kwh4MYQAR\nnlmKnyKn472drnuHRfHtvLBmKFVK7XIOEG6SQyciTLfVf5DR3cCmSnT3sUIOc5eC\n0lRxKoQEDUJLFesMwcCVowlsX3+0bS4raycLQGVbfw+t0fkhc01HN4JkTnacHX3s\nAL+FUu37uaAn0ERE55Gj0ZIjVSdAtbXyIltwfDTyFGwr5F7S2W1VDwFTIU7CfeDM\nB1riL5irhGZWhA7cNTb+/sI91cFzMAU+021YG6g=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'eAlxHx-AlrH0J7axYJB376im9gg',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Root CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'XN7bRRjtcX4gE9SVl5y7MZ4ZjSk',
                notafter => '1170287999', # 2007-01-31T23:59:59
                notbefore => '1136073600', # 2006-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:f1:68:b4:74:a6:b9:ef:a5:d9:31:ca:56:2d:a1:\n    ac:3a:4b:83:3c:8c:c4:e4:b0:21:8e:b4:e4:11:d6:\n    af:b4:4c:b8:16:6a:fb:2d:61:2e:1b:88:50:fc:b7:\n    74:72:cc:ca:39:b0:6d:6d:3e:14:d5:13:1d:bd:0a:\n    84:8d:b1:10:eb:7e:dd:48:d1:ba:b1:1a:37:da:56:\n    cd:5c:88:8d:04:f7:8b:bc:33:a5:17:45:59:62:92:\n    8b:46:ec:1d:9d:d8:eb:62:00:c3:3b:12:bf:09:95:\n    2e:cd:72:e0:0b:3c:07:27:e8:cc:f9:3f:f0:d3:e8:\n    3c:7a:42:b5:85:a9:22:80:d9:73:4e:c6:90:f8:9d:\n    8e:23:21:a0:85:a5:64:e6:cc:fa:b1:1f:5c:15:a3:\n    94:9e:84:67:f4:c1:86:fd:68:d2:01:81:51:69:1f:\n    6f:cc:c3:e0:a9:11:17:67:bd:9c:24:cf:8f:85:5b:\n    10:7b:c2:4e:66:79:01:f5:87:89:64:0e:f5:b1:ac:\n    e8:28:17:b6:1c:f7:b2:dd:07:eb:a1:1f:46:5c:22:\n    01:c2:68:a7:b6:e5:1e:b8:7e:63:0b:1c:3a:f0:d8:\n    5f:9a:ff:a0:0b:87:f7:1e:ee:9f:37:fe:24:cd:c1:\n    50:2c:a9:5c:16:62:c9:5a:34:e3:19:5f:17:ee:13:\n    7e:2d\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA SCEP 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'B0:99:E6:4E:8B:A5:2B:DE:5A:FA:20:CF:6F:CE:55:D3:58:C3:4E:92',
            },
            db_alias => {
                alias => 'alpha-scep-1',
                generation => '1',
                group_id => 'alpha-scep',
            },
        ),

        'alpha_scep_2' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA SCEP 2',
            db => {
                authority_key_identifier => '11:C2:BA:08:26:F9:EC:BE:0D:9E:EB:66:44:0A:9B:83:7B:D9:0B:C9',
                cert_key => '10',
                data => "-----BEGIN CERTIFICATE-----\nMIIDeTCCAmGgAwIBAgIBCjANBgkqhkiG9w0BAQsFADBYMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGDAWBgNVBAMMD0FMUEhBIFJvb3QgQ0EgMjAiGA8yMDA3MDEwMTAwMDAwMFoYDzIx\nMDAwMTMxMjM1OTU5WjBVMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZImiZPy\nLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxFTATBgNVBAMMDEFMUEhBIFND\nRVAgMjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAM3FfS5AIY9NlW8r\njRTAfsvpd2rKsCFRMWUlm68ERo2gADGW8FZotWIc+BDZkK9WC8fPiEApr0wRqe+k\nbFJ2Sif2bm7YlkfMFSeRzumFS64o2yz+KMKQJoE9dd05hyK0EyL4IY6iKZNbuJJO\nj54x5hkuoaeRjULemcNqoXPu3rD/jbCYyKbSn0r9GcU1SxV4ievdhckZ+mN6lmd+\nthYu6zTxy9mv68Q5HGTrf+l1z25+l4a/qucqezxN4YdnWUngZDpHa3NSNEVFQIqg\nU8Van8zS8QJxZUobIsXo5O96b2sJbejqeObNeZ7EoWfU20UgdJSkLKIearwt9s8j\ngQQeOX8CAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4EFgQU7f6HaqS941/TcxRp\n8D5XIirvQaAwHwYDVR0jBBgwFoAUEcK6CCb57L4NnutmRAqbg3vZC8kwDQYJKoZI\nhvcNAQELBQADggEBAL/yTP8fvNLhWgftUlsoFfB6Oncq2nEYOW6RGp9a9WdI58ns\nW4k6PQXrf5O0I9+B+UKMj0N1Oe+MF+F5xl9oE2826S+Yi9B1fYG/JezkQSqw1tyB\nsZwIbHEYG1fbFiNG8aYrHKAMG7QMdUBuVQg1SXuXRyPV8Ub8JPMedP+Rea563u6B\n/OtL0tPEUMso0wq02IUTG2/P8ami1IXxqX+2lryvTZIwAagbhRyXhMKKKPI/lBsi\nc9dnXb5AdQmeytNNehFf+YmdGgroXLje7QprmtR9rBVHIbvUbUAUzR4gMIo7YPfd\nQgrG3T79GNqLN9+BPRe3ns/YKF82QAKyxjeDUQ8=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'qLwiZjrBYFdWEbyxP_cFSzgXGMg',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Root CA 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'Vw7WTIPQSAHhrbNDC3hwOAqwRCc',
                notafter => '4105123199', # 2100-01-31T23:59:59
                notbefore => '1167609600', # 2007-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:cd:c5:7d:2e:40:21:8f:4d:95:6f:2b:8d:14:c0:\n    7e:cb:e9:77:6a:ca:b0:21:51:31:65:25:9b:af:04:\n    46:8d:a0:00:31:96:f0:56:68:b5:62:1c:f8:10:d9:\n    90:af:56:0b:c7:cf:88:40:29:af:4c:11:a9:ef:a4:\n    6c:52:76:4a:27:f6:6e:6e:d8:96:47:cc:15:27:91:\n    ce:e9:85:4b:ae:28:db:2c:fe:28:c2:90:26:81:3d:\n    75:dd:39:87:22:b4:13:22:f8:21:8e:a2:29:93:5b:\n    b8:92:4e:8f:9e:31:e6:19:2e:a1:a7:91:8d:42:de:\n    99:c3:6a:a1:73:ee:de:b0:ff:8d:b0:98:c8:a6:d2:\n    9f:4a:fd:19:c5:35:4b:15:78:89:eb:dd:85:c9:19:\n    fa:63:7a:96:67:7e:b6:16:2e:eb:34:f1:cb:d9:af:\n    eb:c4:39:1c:64:eb:7f:e9:75:cf:6e:7e:97:86:bf:\n    aa:e7:2a:7b:3c:4d:e1:87:67:59:49:e0:64:3a:47:\n    6b:73:52:34:45:45:40:8a:a0:53:c5:5a:9f:cc:d2:\n    f1:02:71:65:4a:1b:22:c5:e8:e4:ef:7a:6f:6b:09:\n    6d:e8:ea:78:e6:cd:79:9e:c4:a1:67:d4:db:45:20:\n    74:94:a4:2c:a2:1e:6a:bc:2d:f6:cf:23:81:04:1e:\n    39:7f\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA SCEP 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'ED:FE:87:6A:A4:BD:E3:5F:D3:73:14:69:F0:3E:57:22:2A:EF:41:A0',
            },
            db_alias => {
                alias => 'alpha-scep-2',
                generation => '2',
                group_id => 'alpha-scep',
            },
        ),

        'alpha_scep_3' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA SCEP 3',
            db => {
                authority_key_identifier => 'B2:0B:FF:E9:DD:EF:91:58:57:13:A4:C5:84:36:D0:F5:F6:C5:54:2A',
                cert_key => '18',
                data => "-----BEGIN CERTIFICATE-----\nMIIDeTCCAmGgAwIBAgIBEjANBgkqhkiG9w0BAQsFADBYMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGDAWBgNVBAMMD0FMUEhBIFJvb3QgQ0EgMzAiGA8yMTAwMDEwMTAwMDAwMFoYDzIx\nMDUwMTMxMjM1OTU5WjBVMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZImiZPy\nLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxFTATBgNVBAMMDEFMUEhBIFND\nRVAgMzCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAO/3ER03j8XP0wTe\nKVrvIyCO7tie6n5yMzy0/DGOCQdLkvxbFxgTCSPC+pRuoZkLG6AdSvlX59Uv+YL8\ngyR0C8wC9TaqJeATa4bylLkkWUUcmu1upFp7ljFv/Heo6bn36e1RJ3VDmnJ1Cy1Z\nGNHCmNCCO/8/xcZbyAxMlELCpT11eFZJyVspSGdalVNvtLUvSVrdCXKvTTtBQ1lu\nN625A3KBvesF7cI5cAGMEPWsA03usV4FkaoRUjzwMiCA1YtGVPEb8/VcUESwxUf8\nACVIw/35XQjev4bbMzWJJkTn//227HpGeuwZ0cNpVNFFqhZUlKXVsWtPUg46H1ra\nMn6ahgcCAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4EFgQUd+N5O8dOytjSzQWE\nc6p75hhtrYgwHwYDVR0jBBgwFoAUsgv/6d3vkVhXE6TFhDbQ9fbFVCowDQYJKoZI\nhvcNAQELBQADggEBAJcnNSdVwvSw2PBw0DeiISl79x18fFUBSBsFK47YtdNB2i+q\nyQd6UPATjJISq5xfjKs+l7APjRQzIWLdoTMP4JEG11gxPaaVp5SpiaV30vodH6EM\n2h9vCHejngDzBVsfEtkBZTcC5rbEHp9bL46eGpZDMX7odtxWL9H43QemR2kgNUK6\ni5hy28HB8aI54vtmQp3CcK4b6op8YOwbdAigB6c902agIUf6D9vLZwYPGDuYl9Wf\nc2KzBhj8hUvezkMWX299tREpGwz6MPY/ros0VMyOKv4JU+t6uT5pYeemERbx3XaL\n1TKsyqvbXxOWq8+0Hl6YiGzaY4xR1lwqz4ZbOHI=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'QwmF1mYk4R3YVvW4ixJUI8-nrJA',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Root CA 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'IIdXIQJ615WVI7f4JF2OrYS5sPY',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '4102444800', # 2100-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:ef:f7:11:1d:37:8f:c5:cf:d3:04:de:29:5a:ef:\n    23:20:8e:ee:d8:9e:ea:7e:72:33:3c:b4:fc:31:8e:\n    09:07:4b:92:fc:5b:17:18:13:09:23:c2:fa:94:6e:\n    a1:99:0b:1b:a0:1d:4a:f9:57:e7:d5:2f:f9:82:fc:\n    83:24:74:0b:cc:02:f5:36:aa:25:e0:13:6b:86:f2:\n    94:b9:24:59:45:1c:9a:ed:6e:a4:5a:7b:96:31:6f:\n    fc:77:a8:e9:b9:f7:e9:ed:51:27:75:43:9a:72:75:\n    0b:2d:59:18:d1:c2:98:d0:82:3b:ff:3f:c5:c6:5b:\n    c8:0c:4c:94:42:c2:a5:3d:75:78:56:49:c9:5b:29:\n    48:67:5a:95:53:6f:b4:b5:2f:49:5a:dd:09:72:af:\n    4d:3b:41:43:59:6e:37:ad:b9:03:72:81:bd:eb:05:\n    ed:c2:39:70:01:8c:10:f5:ac:03:4d:ee:b1:5e:05:\n    91:aa:11:52:3c:f0:32:20:80:d5:8b:46:54:f1:1b:\n    f3:f5:5c:50:44:b0:c5:47:fc:00:25:48:c3:fd:f9:\n    5d:08:de:bf:86:db:33:35:89:26:44:e7:ff:fd:b6:\n    ec:7a:46:7a:ec:19:d1:c3:69:54:d1:45:aa:16:54:\n    94:a5:d5:b1:6b:4f:52:0e:3a:1f:5a:da:32:7e:9a:\n    86:07\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA SCEP 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '77:E3:79:3B:C7:4E:CA:D8:D2:CD:05:84:73:AA:7B:E6:18:6D:AD:88',
            },
            db_alias => {
                alias => 'alpha-scep-3',
                generation => '3',
                group_id => 'alpha-scep',
            },
        ),

        'alpha_signer_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Signing CA 1',
            db => {
                authority_key_identifier => '7B:8F:B0:2C:F9:51:1D:25:CD:51:1F:DF:6D:93:8C:A9:4F:D9:00:74',
                cert_key => '3',
                data => "-----BEGIN CERTIFICATE-----\nMIIDkjCCAnqgAwIBAgIBAzANBgkqhkiG9w0BAQsFADBYMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGDAWBgNVBAMMD0FMUEhBIFJvb3QgQ0EgMTAiGA8yMDA2MDEwMTAwMDAwMFoYDzIw\nMDcwMTMxMjM1OTU5WjBbMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZImiZPy\nLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGzAZBgNVBAMMEkFMUEhBIFNp\nZ25pbmcgQ0EgMTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAMUUnEXy\neP5Bil8DdoYf+AQYiB8kedb+ZvUqa4sAcGyEARc30mVy5hSuQQiM9YI6ZoiiyOn7\n7M9/jGMNj7KZWeKsLWWJJaJJFmDeckYELhXhFcoQkOBuAiCbcKFBHv8dy8KcUvcz\nC7PPAkImCawzTLUV9oOK7gXDUrrAaBL3BhdZ/ApnbqJGtyAEVAH/PV5UeDFsbaPf\nF0k23l+I4vGWPD3ccYtTcWDcl5+Wc1dnQJ1YLGtsTbnNDsihUx0EC4Pqpm/dqIlZ\n6AdBG8m7HHt0B7yKIY3ulIMFwRyOQazN1Qb09p8Xey4WOEZrbTQ+BLr6bW9+VZ4M\n2WLJHxXe4mMnsAUCAwEAAaNgMF4wDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQU\nPWyGDADxjeBeD4SLgt6bvk7E3UMwHwYDVR0jBBgwFoAUe4+wLPlRHSXNUR/fbZOM\nqU/ZAHQwCwYDVR0PBAQDAgEGMA0GCSqGSIb3DQEBCwUAA4IBAQAn1EMyg2515Rnx\n4ZlGuP2CBlxPMxAiKpfpVu32JysI/z3bsZBBZz/wFamMxGMHHwwaJ9FG/nBsE1Rz\nJ/Ooe0E5RfrjrKmgo6RCmQBF3bkGz4zY0uSsStkfdObjQkyThpMaLDrH24Rgaqxk\nZErwyM4LUyZXf8lBGbV5m4DuPqjtUXCddCLwlt4J2uaCsBX77hYdryFMBLvi4lsg\nCeyx//QhnWtVxz98WCA5JaM2um7VIbNiGteSSzNVx99G+ARlmYomJM4t+VYTY65H\nPJSvTOg2t2Oy+LcIrJ01pZNO3kTzrBhn67YmdrjlQXqW+TDpd/K+ShlBuE9K0ZUk\nIhY2103f\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'j6Aps6CGQftSMQM4Gjue2-3RmFw',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Root CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'XN7bRRjtcX4gE9SVl5y7MZ4ZjSk',
                notafter => '1170287999', # 2007-01-31T23:59:59
                notbefore => '1136073600', # 2006-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:c5:14:9c:45:f2:78:fe:41:8a:5f:03:76:86:1f:\n    f8:04:18:88:1f:24:79:d6:fe:66:f5:2a:6b:8b:00:\n    70:6c:84:01:17:37:d2:65:72:e6:14:ae:41:08:8c:\n    f5:82:3a:66:88:a2:c8:e9:fb:ec:cf:7f:8c:63:0d:\n    8f:b2:99:59:e2:ac:2d:65:89:25:a2:49:16:60:de:\n    72:46:04:2e:15:e1:15:ca:10:90:e0:6e:02:20:9b:\n    70:a1:41:1e:ff:1d:cb:c2:9c:52:f7:33:0b:b3:cf:\n    02:42:26:09:ac:33:4c:b5:15:f6:83:8a:ee:05:c3:\n    52:ba:c0:68:12:f7:06:17:59:fc:0a:67:6e:a2:46:\n    b7:20:04:54:01:ff:3d:5e:54:78:31:6c:6d:a3:df:\n    17:49:36:de:5f:88:e2:f1:96:3c:3d:dc:71:8b:53:\n    71:60:dc:97:9f:96:73:57:67:40:9d:58:2c:6b:6c:\n    4d:b9:cd:0e:c8:a1:53:1d:04:0b:83:ea:a6:6f:dd:\n    a8:89:59:e8:07:41:1b:c9:bb:1c:7b:74:07:bc:8a:\n    21:8d:ee:94:83:05:c1:1c:8e:41:ac:cd:d5:06:f4:\n    f6:9f:17:7b:2e:16:38:46:6b:6d:34:3e:04:ba:fa:\n    6d:6f:7e:55:9e:0c:d9:62:c9:1f:15:de:e2:63:27:\n    b0:05\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Signing CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '3D:6C:86:0C:00:F1:8D:E0:5E:0F:84:8B:82:DE:9B:BE:4E:C4:DD:43',
            },
            db_alias => {
                alias => 'alpha-signer-1',
                generation => '1',
                group_id => 'alpha-signer',
            },
        ),

        'alpha_signer_2' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Signing CA 2',
            db => {
                authority_key_identifier => '11:C2:BA:08:26:F9:EC:BE:0D:9E:EB:66:44:0A:9B:83:7B:D9:0B:C9',
                cert_key => '9',
                data => "-----BEGIN CERTIFICATE-----\nMIIDkjCCAnqgAwIBAgIBCTANBgkqhkiG9w0BAQsFADBYMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGDAWBgNVBAMMD0FMUEhBIFJvb3QgQ0EgMjAiGA8yMDA3MDEwMTAwMDAwMFoYDzIx\nMDAwMTMxMjM1OTU5WjBbMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZImiZPy\nLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGzAZBgNVBAMMEkFMUEhBIFNp\nZ25pbmcgQ0EgMjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAKbcad+x\nL6HptENp199K6GM/ejvQeOh+KGWZR6NaxbHHtv6J46lECTcP3lU6XpOSoP4f6xay\nX7DGBlNhXacpwnI6bYXOKQmP1Ih8GnkGkSrBhHSdU4wo0fS5mq8U5DEkonKR+oIp\n/wjVENjJHfkYV+Jd1TpJ+rWLBHfzmtub6OwmpRNnw6OlQEglrO0OOyULUQipJoKZ\nyTc2A3YbQ0mPRovUz2UWTl0NUGhHOMle+fu8rx03eFY6EFKLOf2nOyP1sKtXtWpG\n76VAoz21Z+XxqVFSjaakfqlffNzxELs4rhF36bVswpncEtj6NJB4Jm0y483cEdp8\nW7GMjr/dkrcc2g0CAwEAAaNgMF4wDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQU\nqL5LG1IYYfffBlpevcVr6BgoP+0wHwYDVR0jBBgwFoAUEcK6CCb57L4NnutmRAqb\ng3vZC8kwCwYDVR0PBAQDAgEGMA0GCSqGSIb3DQEBCwUAA4IBAQAeY76+1o8XYtpY\nNkcyJRvREzt6Iq8VqkO/wsNbdxkvqjjY6oqQSCn9ie+KrTRiqveF8kCED4myXzww\nNHMN86JexRaa/eWTHa15s0FarDtVOOTzF6G6WGHjRwA4f9nquVBozJ3EEOKFa2QD\nVfH9v0f8DxHocC2rpb9jLCbSFZXwcoIEUXAl/zjUjiOR2kXhxeBaJAr5EfdsP0kg\n1Zc8DueUmXGJ9Z4ArFXyBur+dI2i18mY5JEEgTT5izJYphpcQ8x4mwFTd6HOWSwE\n2L4e31ggxt8kr+v8Fv9WVoY8YCKRD/TzeEYP+jqDjKlTk81EUC7HL46BAbGXs0ig\nF67O9h8C\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'RE35XR3XIBXiIbAu8P5aGMCmH7o',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Root CA 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'Vw7WTIPQSAHhrbNDC3hwOAqwRCc',
                notafter => '4105123199', # 2100-01-31T23:59:59
                notbefore => '1167609600', # 2007-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:a6:dc:69:df:b1:2f:a1:e9:b4:43:69:d7:df:4a:\n    e8:63:3f:7a:3b:d0:78:e8:7e:28:65:99:47:a3:5a:\n    c5:b1:c7:b6:fe:89:e3:a9:44:09:37:0f:de:55:3a:\n    5e:93:92:a0:fe:1f:eb:16:b2:5f:b0:c6:06:53:61:\n    5d:a7:29:c2:72:3a:6d:85:ce:29:09:8f:d4:88:7c:\n    1a:79:06:91:2a:c1:84:74:9d:53:8c:28:d1:f4:b9:\n    9a:af:14:e4:31:24:a2:72:91:fa:82:29:ff:08:d5:\n    10:d8:c9:1d:f9:18:57:e2:5d:d5:3a:49:fa:b5:8b:\n    04:77:f3:9a:db:9b:e8:ec:26:a5:13:67:c3:a3:a5:\n    40:48:25:ac:ed:0e:3b:25:0b:51:08:a9:26:82:99:\n    c9:37:36:03:76:1b:43:49:8f:46:8b:d4:cf:65:16:\n    4e:5d:0d:50:68:47:38:c9:5e:f9:fb:bc:af:1d:37:\n    78:56:3a:10:52:8b:39:fd:a7:3b:23:f5:b0:ab:57:\n    b5:6a:46:ef:a5:40:a3:3d:b5:67:e5:f1:a9:51:52:\n    8d:a6:a4:7e:a9:5f:7c:dc:f1:10:bb:38:ae:11:77:\n    e9:b5:6c:c2:99:dc:12:d8:fa:34:90:78:26:6d:32:\n    e3:cd:dc:11:da:7c:5b:b1:8c:8e:bf:dd:92:b7:1c:\n    da:0d\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Signing CA 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'A8:BE:4B:1B:52:18:61:F7:DF:06:5A:5E:BD:C5:6B:E8:18:28:3F:ED',
            },
            db_alias => {
                alias => 'alpha-signer-2',
                generation => '2',
                group_id => 'alpha-signer',
            },
        ),

        'alpha_signer_3' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Signing CA 3',
            db => {
                authority_key_identifier => 'B2:0B:FF:E9:DD:EF:91:58:57:13:A4:C5:84:36:D0:F5:F6:C5:54:2A',
                cert_key => '17',
                data => "-----BEGIN CERTIFICATE-----\nMIIDkjCCAnqgAwIBAgIBETANBgkqhkiG9w0BAQsFADBYMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGDAWBgNVBAMMD0FMUEhBIFJvb3QgQ0EgMzAiGA8yMTAwMDEwMTAwMDAwMFoYDzIx\nMDUwMTMxMjM1OTU5WjBbMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZImiZPy\nLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGzAZBgNVBAMMEkFMUEhBIFNp\nZ25pbmcgQ0EgMzCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALiEKsCV\n6++ZQ5USNOAUhhwlzdOoRpLV9cPm//XdXR7EZsBZigyDaLaqsEdqpYIAM5C/xNFc\nDkGygxXmtcryH+H6swhI89LeHXqzbxvW63v4sjVtEmvAZp9aMZSXAhzFHJOKsnG+\nYBhv2q/vM1HWlRfVRM5/TTM184RXwLKDvHLWssT0KsQIw0T8LGIyn2YkmT0pyppz\n19xsSqM9WXD7OeJVglmXZeptMy4RztAR3us3VjNC2JkVfocKXyxu1vN+JYhRljIx\nVv6W2r+b4eQyz59VWb7PFq6vU8DpZX4inPJ0QwK3vUEG+1XctPbu2vn0UFB1aZ3f\nJooqEQSaNIY48vECAwEAAaNgMF4wDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQU\nYjXOU5HH9CGsW4uhEWDeIJeX8FYwHwYDVR0jBBgwFoAUsgv/6d3vkVhXE6TFhDbQ\n9fbFVCowCwYDVR0PBAQDAgEGMA0GCSqGSIb3DQEBCwUAA4IBAQAldFrviYHbNcNQ\nOmeOB4D67ltzkRHqkCyonm4mQma2xvCZPiF9CnC8A6iClvmMiKr7D/JZ8GJrMKEP\n0CCk6ie8KyMW9P29JQzMGyLoHjx/+IQQ9MvD9BbVcBd0rGRCoPZYMgjBT5/GKapJ\nCeHmAed8JcnIUQZaQmB4ubE2Ok50XbBKtbqf2KieCyDDRphX9Eb2Q4fChJZLOMxE\n5rYo9KT6ni6G4A3pwxZygqAKHoxLrCZf7nDgZJJfY/glnfQnfWqs1sJ7I/PPURZQ\nvxSZvt0AMuHHZcxpT9/QDWwJmSMmCnDWZxNtU/O4HoSL5WqEof8cKFf+/sk7UnE8\n64I9S8Nr\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'ZS6zCaRlo1bqb5a33CPnYqptoZE',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Root CA 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'IIdXIQJ615WVI7f4JF2OrYS5sPY',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '4102444800', # 2100-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:b8:84:2a:c0:95:eb:ef:99:43:95:12:34:e0:14:\n    86:1c:25:cd:d3:a8:46:92:d5:f5:c3:e6:ff:f5:dd:\n    5d:1e:c4:66:c0:59:8a:0c:83:68:b6:aa:b0:47:6a:\n    a5:82:00:33:90:bf:c4:d1:5c:0e:41:b2:83:15:e6:\n    b5:ca:f2:1f:e1:fa:b3:08:48:f3:d2:de:1d:7a:b3:\n    6f:1b:d6:eb:7b:f8:b2:35:6d:12:6b:c0:66:9f:5a:\n    31:94:97:02:1c:c5:1c:93:8a:b2:71:be:60:18:6f:\n    da:af:ef:33:51:d6:95:17:d5:44:ce:7f:4d:33:35:\n    f3:84:57:c0:b2:83:bc:72:d6:b2:c4:f4:2a:c4:08:\n    c3:44:fc:2c:62:32:9f:66:24:99:3d:29:ca:9a:73:\n    d7:dc:6c:4a:a3:3d:59:70:fb:39:e2:55:82:59:97:\n    65:ea:6d:33:2e:11:ce:d0:11:de:eb:37:56:33:42:\n    d8:99:15:7e:87:0a:5f:2c:6e:d6:f3:7e:25:88:51:\n    96:32:31:56:fe:96:da:bf:9b:e1:e4:32:cf:9f:55:\n    59:be:cf:16:ae:af:53:c0:e9:65:7e:22:9c:f2:74:\n    43:02:b7:bd:41:06:fb:55:dc:b4:f6:ee:da:f9:f4:\n    50:50:75:69:9d:df:26:8a:2a:11:04:9a:34:86:38:\n    f2:f1\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Signing CA 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '62:35:CE:53:91:C7:F4:21:AC:5B:8B:A1:11:60:DE:20:97:97:F0:56',
            },
            db_alias => {
                alias => 'alpha-signer-3',
                generation => '3',
                group_id => 'alpha-signer',
            },
        ),

        'alpha_root_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Root CA 1',
            db => {
                authority_key_identifier => '7B:8F:B0:2C:F9:51:1D:25:CD:51:1F:DF:6D:93:8C:A9:4F:D9:00:74',
                cert_key => '2',
                data => "-----BEGIN CERTIFICATE-----\nMIIDjzCCAnegAwIBAgIBAjANBgkqhkiG9w0BAQsFADBYMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGDAWBgNVBAMMD0FMUEhBIFJvb3QgQ0EgMTAiGA8yMDA2MDEwMTAwMDAwMFoYDzIw\nMDcwMTMxMjM1OTU5WjBYMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZImiZPy\nLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGDAWBgNVBAMMD0FMUEhBIFJv\nb3QgQ0EgMTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBANlU3cP/H/Ue\nCcRFcUiMgQ2QwOXDXzgohO8yPCu9FdhhdNKtIP3LLIahhGKFPCiEqutkRZ6f3937\nC1sjdxSoGwP0/VLxigLGopa/+WCOyj4wNkZ99Tp3KhqkWHeQ8IropYdH3jvq2oA+\n7ng6CnAw+Fsv1zbdXO/sucht6S7HsDYglBKKeTnylOe6RNFE265E5VUL5yRzn1dg\n4qB1jTvTdNFLs267ymYwmVtRBxG5tIU3Mud//eyQYuCl9OG8kbaFcFLwJhn0L/AI\n/I1UvsgRmAmu/NRJUxO4YlmNtUDgVFgOfSH5tRylojVK6SPxZ+1NJLb/nFvHTTMu\ncNrtFIh6JfcCAwEAAaNgMF4wDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUe4+w\nLPlRHSXNUR/fbZOMqU/ZAHQwHwYDVR0jBBgwFoAUe4+wLPlRHSXNUR/fbZOMqU/Z\nAHQwCwYDVR0PBAQDAgEGMA0GCSqGSIb3DQEBCwUAA4IBAQDD7yKfTxoQLBkeh+mT\ndXjW291e5FbD7E4Vyotiv/ES2ftuNSuhOtjtDGTtJdEvtblzStEu2KuBqCEd1LZm\nmB+ii4kOsmG7TuCVNXsk1UcIlB3AQVTaZfq2u4OyBYts06o1OraXIO0PDhAX6lrE\ni//frMHtmh25U6wGXFoprg+oDwPxssLhUsTRBiACjb8TLP9k9mRGQI0UjZooGPH9\n6u1opF2OdwGYpCUl4oB2jQCgmhjTkD3F9jo7mK7aCvDNUn9ABc8KqKQnHCtLaBbF\nLddWpgoABPIEGCe4xfb+d3VLcgH/4qTsOBdxQjAqz157E9i307xQQz7xWmIaiVpB\nda9U\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'XN7bRRjtcX4gE9SVl5y7MZ4ZjSk',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Root CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'XN7bRRjtcX4gE9SVl5y7MZ4ZjSk',
                notafter => '1170287999', # 2007-01-31T23:59:59
                notbefore => '1136073600', # 2006-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:d9:54:dd:c3:ff:1f:f5:1e:09:c4:45:71:48:8c:\n    81:0d:90:c0:e5:c3:5f:38:28:84:ef:32:3c:2b:bd:\n    15:d8:61:74:d2:ad:20:fd:cb:2c:86:a1:84:62:85:\n    3c:28:84:aa:eb:64:45:9e:9f:df:dd:fb:0b:5b:23:\n    77:14:a8:1b:03:f4:fd:52:f1:8a:02:c6:a2:96:bf:\n    f9:60:8e:ca:3e:30:36:46:7d:f5:3a:77:2a:1a:a4:\n    58:77:90:f0:8a:e8:a5:87:47:de:3b:ea:da:80:3e:\n    ee:78:3a:0a:70:30:f8:5b:2f:d7:36:dd:5c:ef:ec:\n    b9:c8:6d:e9:2e:c7:b0:36:20:94:12:8a:79:39:f2:\n    94:e7:ba:44:d1:44:db:ae:44:e5:55:0b:e7:24:73:\n    9f:57:60:e2:a0:75:8d:3b:d3:74:d1:4b:b3:6e:bb:\n    ca:66:30:99:5b:51:07:11:b9:b4:85:37:32:e7:7f:\n    fd:ec:90:62:e0:a5:f4:e1:bc:91:b6:85:70:52:f0:\n    26:19:f4:2f:f0:08:fc:8d:54:be:c8:11:98:09:ae:\n    fc:d4:49:53:13:b8:62:59:8d:b5:40:e0:54:58:0e:\n    7d:21:f9:b5:1c:a5:a2:35:4a:e9:23:f1:67:ed:4d:\n    24:b6:ff:9c:5b:c7:4d:33:2e:70:da:ed:14:88:7a:\n    25:f7\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Root CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '7B:8F:B0:2C:F9:51:1D:25:CD:51:1F:DF:6D:93:8C:A9:4F:D9:00:74',
            },
            db_alias => {
                alias => 'root-1',
                generation => '1',
                group_id => 'root',
            },
        ),

        'alpha_root_2' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Root CA 2',
            db => {
                authority_key_identifier => '11:C2:BA:08:26:F9:EC:BE:0D:9E:EB:66:44:0A:9B:83:7B:D9:0B:C9',
                cert_key => '8',
                data => "-----BEGIN CERTIFICATE-----\nMIIDjzCCAnegAwIBAgIBCDANBgkqhkiG9w0BAQsFADBYMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGDAWBgNVBAMMD0FMUEhBIFJvb3QgQ0EgMjAiGA8yMDA3MDEwMTAwMDAwMFoYDzIx\nMDAwMTMxMjM1OTU5WjBYMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZImiZPy\nLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGDAWBgNVBAMMD0FMUEhBIFJv\nb3QgQ0EgMjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBANaxw8j+m1C0\n/MvSmTUaexUz781UjCp9oTlJ/IrwzcrcEhjotKdIGpwNqxe6mDHj0VrAREAQaOlD\nd7kXNUGk4Yxbrirv6yZovcJGrbUD07XRUhQK3ltW9VWRtnuByKcUxvyy9qE9LXp8\na17HVqNhysTDzMGFAQYI6YidQY7sqLyP2dKCUgEbgJLw4ToQHP3WrvOyEMqQtAhj\nf+hOzUpULFMLtJ0TvnVqnefauJzHqcytYTUE75JilVj6DVWjCRLBzKjiPH2Pq+7v\nPiKBC+Emq9WFVzsk1AudR4TgYzgCLDiqVKQVMWCAW7fy8MZrMX8xwW7EFEsTpaXz\n6XbXUAQnOQ8CAwEAAaNgMF4wDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUEcK6\nCCb57L4NnutmRAqbg3vZC8kwHwYDVR0jBBgwFoAUEcK6CCb57L4NnutmRAqbg3vZ\nC8kwCwYDVR0PBAQDAgEGMA0GCSqGSIb3DQEBCwUAA4IBAQA83VB/kPcU0eeFc7Yo\n5/LBslNlEieEal0N0sNwQD814+4JXiylVvpNCLZLlR/sTbNVckoUE4C6Prc795rb\nv7wgtPk/yZyiDIZj4Pa/cq/ekAWsR4rh4rs1xgIbWI38nN5NrT+K8O61sNzsT5iE\nBVxSNaA2zd5YbRzMbRVh6vstyYmXa+MevXhU1DyeM+/L44xt8kW7JWQuy6c69y24\nDBaNYSgjN3cg2dTj6XIoXgRRU8bhLshnAMv3wZgJUIqIt7eHKISMMPhZpIUcQTg0\nlXHfPlyaIM1oXrtD8beYtRrQjktLk2bk05CATlv81yRb91h8sBgOz7qGDmReCrDA\nA/Eh\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'Vw7WTIPQSAHhrbNDC3hwOAqwRCc',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Root CA 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'Vw7WTIPQSAHhrbNDC3hwOAqwRCc',
                notafter => '4105123199', # 2100-01-31T23:59:59
                notbefore => '1167609600', # 2007-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:d6:b1:c3:c8:fe:9b:50:b4:fc:cb:d2:99:35:1a:\n    7b:15:33:ef:cd:54:8c:2a:7d:a1:39:49:fc:8a:f0:\n    cd:ca:dc:12:18:e8:b4:a7:48:1a:9c:0d:ab:17:ba:\n    98:31:e3:d1:5a:c0:44:40:10:68:e9:43:77:b9:17:\n    35:41:a4:e1:8c:5b:ae:2a:ef:eb:26:68:bd:c2:46:\n    ad:b5:03:d3:b5:d1:52:14:0a:de:5b:56:f5:55:91:\n    b6:7b:81:c8:a7:14:c6:fc:b2:f6:a1:3d:2d:7a:7c:\n    6b:5e:c7:56:a3:61:ca:c4:c3:cc:c1:85:01:06:08:\n    e9:88:9d:41:8e:ec:a8:bc:8f:d9:d2:82:52:01:1b:\n    80:92:f0:e1:3a:10:1c:fd:d6:ae:f3:b2:10:ca:90:\n    b4:08:63:7f:e8:4e:cd:4a:54:2c:53:0b:b4:9d:13:\n    be:75:6a:9d:e7:da:b8:9c:c7:a9:cc:ad:61:35:04:\n    ef:92:62:95:58:fa:0d:55:a3:09:12:c1:cc:a8:e2:\n    3c:7d:8f:ab:ee:ef:3e:22:81:0b:e1:26:ab:d5:85:\n    57:3b:24:d4:0b:9d:47:84:e0:63:38:02:2c:38:aa:\n    54:a4:15:31:60:80:5b:b7:f2:f0:c6:6b:31:7f:31:\n    c1:6e:c4:14:4b:13:a5:a5:f3:e9:76:d7:50:04:27:\n    39:0f\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Root CA 2,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '11:C2:BA:08:26:F9:EC:BE:0D:9E:EB:66:44:0A:9B:83:7B:D9:0B:C9',
            },
            db_alias => {
                alias => 'root-2',
                generation => '2',
                group_id => 'root',
            },
        ),

        'alpha_root_3' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'ALPHA Root CA 3',
            db => {
                authority_key_identifier => 'B2:0B:FF:E9:DD:EF:91:58:57:13:A4:C5:84:36:D0:F5:F6:C5:54:2A',
                cert_key => '16',
                data => "-----BEGIN CERTIFICATE-----\nMIIDjzCCAnegAwIBAgIBEDANBgkqhkiG9w0BAQsFADBYMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGDAWBgNVBAMMD0FMUEhBIFJvb3QgQ0EgMzAiGA8yMTAwMDEwMTAwMDAwMFoYDzIx\nMDUwMTMxMjM1OTU5WjBYMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZImiZPy\nLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGDAWBgNVBAMMD0FMUEhBIFJv\nb3QgQ0EgMzCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBANzx0bJUsLAV\n4bLUyLUR6cNJ6WrGBg8IlRFyYcD5P/FK4xr8gmo5c1fzSlt5ltLq5Zp61Pcuozjn\nwgA5vrB5FrzcxLBY8L7eN3g2uuChV9NsHGkWRdNdYbyp63SDwYReUdvtMKRfRhjK\nOjHMMTwsnLjivsAq07Q1XDOkozKvMFsB7sW3x6bvOykZ7CvcaTrCXSFVbCvnWDdd\n0uoARMwgLKcaTWVkw3WB1ES/JNxGeD09oqRVj3lJBEQ1VKuGYNG2SyEtjYtHyYug\ncUmivHpxKGRrDHIvI7eHUyh7qXqVg93YGkpCKE3AV+wiGipapwam1Y6q6kqGNVHO\nTheSmWsXS1MCAwEAAaNgMF4wDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUsgv/\n6d3vkVhXE6TFhDbQ9fbFVCowHwYDVR0jBBgwFoAUsgv/6d3vkVhXE6TFhDbQ9fbF\nVCowCwYDVR0PBAQDAgEGMA0GCSqGSIb3DQEBCwUAA4IBAQAEWvJH3f6/ZVaD/Ruh\nMsw8BWS1I0Yi9ib4QPb2dTMJZo36WsPw9A7cPt12wSV4d80TgGxdkM/rJ3LynF/+\nRYvCb38MF5Cf2mIu48gsj1w8wMZUQvWMfS5CBH5gxaEQ6IAWrItJ8IHhy6fJKjRE\nrCzZPOwAFmA2jXVMCEvTF6XgKcUnRLR4jwPZcmyMUVSL3/aGTeBncEgWdrYKvRSa\nX9EUQy5dDUUapCFdqG1PMnwZkosKg8iZmRDJxTnhuoky1RMdTjLmdKQFRQQj4oVz\nQH+URQdUu81yjFtXAXBsAAOrtq9+qtvFjc21Cuv3DMo3Eg4EDn7En6T5JoXQQUJO\n9FHA\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'IIdXIQJ615WVI7f4JF2OrYS5sPY',
                invalidity_time => undef,
                issuer_dn => 'CN=ALPHA Root CA 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'IIdXIQJ615WVI7f4JF2OrYS5sPY',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '4102444800', # 2100-01-01T00:00:00
                pki_realm => 'alpha',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:dc:f1:d1:b2:54:b0:b0:15:e1:b2:d4:c8:b5:11:\n    e9:c3:49:e9:6a:c6:06:0f:08:95:11:72:61:c0:f9:\n    3f:f1:4a:e3:1a:fc:82:6a:39:73:57:f3:4a:5b:79:\n    96:d2:ea:e5:9a:7a:d4:f7:2e:a3:38:e7:c2:00:39:\n    be:b0:79:16:bc:dc:c4:b0:58:f0:be:de:37:78:36:\n    ba:e0:a1:57:d3:6c:1c:69:16:45:d3:5d:61:bc:a9:\n    eb:74:83:c1:84:5e:51:db:ed:30:a4:5f:46:18:ca:\n    3a:31:cc:31:3c:2c:9c:b8:e2:be:c0:2a:d3:b4:35:\n    5c:33:a4:a3:32:af:30:5b:01:ee:c5:b7:c7:a6:ef:\n    3b:29:19:ec:2b:dc:69:3a:c2:5d:21:55:6c:2b:e7:\n    58:37:5d:d2:ea:00:44:cc:20:2c:a7:1a:4d:65:64:\n    c3:75:81:d4:44:bf:24:dc:46:78:3d:3d:a2:a4:55:\n    8f:79:49:04:44:35:54:ab:86:60:d1:b6:4b:21:2d:\n    8d:8b:47:c9:8b:a0:71:49:a2:bc:7a:71:28:64:6b:\n    0c:72:2f:23:b7:87:53:28:7b:a9:7a:95:83:dd:d8:\n    1a:4a:42:28:4d:c0:57:ec:22:1a:2a:5a:a7:06:a6:\n    d5:8e:aa:ea:4a:86:35:51:ce:4e:17:92:99:6b:17:\n    4b:53\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=ALPHA Root CA 3,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'B2:0B:FF:E9:DD:EF:91:58:57:13:A4:C5:84:36:D0:F5:F6:C5:54:2A',
            },
            db_alias => {
                alias => 'root-3',
                generation => '3',
                group_id => 'root',
            },
        ),

        'beta_alice_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'BETA Client Alice 1',
            db => {
                authority_key_identifier => 'DA:FB:5E:2F:91:C4:4E:9E:F6:46:7B:16:DD:1E:C2:4E:66:9B:2B:48',
                cert_key => '25',
                data => "-----BEGIN CERTIFICATE-----\nMIIDgjCCAmqgAwIBAgIBGTANBgkqhkiG9w0BAQsFADBaMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGjAYBgNVBAMMEUJFVEEgU2lnbmluZyBDQSAxMCIYDzIwMTcwMTAxMDAwMDAwWhgP\nMjEwNTAxMzEyMzU5NTlaMFwxEzARBgoJkiaJk/IsZAEZFgNPUkcxGDAWBgoJkiaJ\nk/IsZAEZFghPcGVuWFBLSTENMAsGA1UECwwEQUNNRTEcMBoGA1UEAwwTQkVUQSBD\nbGllbnQgQWxpY2UgMTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAMG4\n5QS3ZtUkBZ9Iu11emdLS3tJvM7zkulORIgc7klRXWA26t1y0YywVOMkCRzjiWxfG\nN86eritr8WB546lZthCkmfWzGuHTjxBVTZDL5aT1YIkgb003aZXNeElKKr1aduJK\nTkUQD/hKUghpZwN1VPhY652Yu1nS75x4AMnkCWxdZOu8+/io63J7Hb6wXHJitfUw\nmGJy0M0LrpfThYUAyW2Q/NRsMK4tFqldOGmznjGcygFcFGITa5DuyEyybdNHhyiU\n5VrVFdDjjZjXgKNsayPw4n38FvbcOHKrBWdwIY9UuqSmagcMiTf3qMdbjH2X8qPJ\nPrbdq0azranfEPhn6nkCAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4EFgQUOMPc\nCWDnHPkVieG2bIr2JRjNfLQwHwYDVR0jBBgwFoAU2vteL5HETp72RnsW3R7CTmab\nK0gwDQYJKoZIhvcNAQELBQADggEBAJEPGca7d5URSUH3YevT8LgG6fHS7SCGsZnB\nF72nmE5cMSbnYuVdXHTbk/HRXnagKkYQqU808GH+KAkNGMcI8aABt337XQ6lOw9H\n51T00H3d6ZaD4GAhQnL4DNIB5W3QYppKj4KOmlJnpT2w+XWThTcM9QnGH7Rw1hri\nogq+0dUL4+46nCiu38wm4D57YJ+MYyWk+mOg7TKpaOIR3CByHsUeclB1+jgmMBe2\nb3K1qzQ+gNRCJX1d6OmZTlp+uqSafWmNUVWNXSOWW7zGu6X1logvp6S1ftNrg4GY\nrsKUv6TGz28hS1wmGennUMBVgc/RRJzx8YKZV2a7ld95O0vXuBE=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'ZS0Kg-FCgwyg9JBLuyBjJnUehLc',
                invalidity_time => undef,
                issuer_dn => 'CN=BETA Signing CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'yWp_z39FC7zb9PCNX7xjEK7BPWw',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '1483228800', # 2017-01-01T00:00:00
                pki_realm => 'beta',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:c1:b8:e5:04:b7:66:d5:24:05:9f:48:bb:5d:5e:\n    99:d2:d2:de:d2:6f:33:bc:e4:ba:53:91:22:07:3b:\n    92:54:57:58:0d:ba:b7:5c:b4:63:2c:15:38:c9:02:\n    47:38:e2:5b:17:c6:37:ce:9e:ae:2b:6b:f1:60:79:\n    e3:a9:59:b6:10:a4:99:f5:b3:1a:e1:d3:8f:10:55:\n    4d:90:cb:e5:a4:f5:60:89:20:6f:4d:37:69:95:cd:\n    78:49:4a:2a:bd:5a:76:e2:4a:4e:45:10:0f:f8:4a:\n    52:08:69:67:03:75:54:f8:58:eb:9d:98:bb:59:d2:\n    ef:9c:78:00:c9:e4:09:6c:5d:64:eb:bc:fb:f8:a8:\n    eb:72:7b:1d:be:b0:5c:72:62:b5:f5:30:98:62:72:\n    d0:cd:0b:ae:97:d3:85:85:00:c9:6d:90:fc:d4:6c:\n    30:ae:2d:16:a9:5d:38:69:b3:9e:31:9c:ca:01:5c:\n    14:62:13:6b:90:ee:c8:4c:b2:6d:d3:47:87:28:94:\n    e5:5a:d5:15:d0:e3:8d:98:d7:80:a3:6c:6b:23:f0:\n    e2:7d:fc:16:f6:dc:38:72:ab:05:67:70:21:8f:54:\n    ba:a4:a6:6a:07:0c:89:37:f7:a8:c7:5b:8c:7d:97:\n    f2:a3:c9:3e:b6:dd:ab:46:b3:ad:a9:df:10:f8:67:\n    ea:79\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=BETA Client Alice 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '38:C3:DC:09:60:E7:1C:F9:15:89:E1:B6:6C:8A:F6:25:18:CD:7C:B4',
            },
            db_alias => {
                alias => 'beta-alice-1',
                generation => undef,
                group_id => undef,
            },
        ),

        'beta_bob_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'BETA Client Bob 1',
            db => {
                authority_key_identifier => 'DA:FB:5E:2F:91:C4:4E:9E:F6:46:7B:16:DD:1E:C2:4E:66:9B:2B:48',
                cert_key => '26',
                data => "-----BEGIN CERTIFICATE-----\nMIIDgDCCAmigAwIBAgIBGjANBgkqhkiG9w0BAQsFADBaMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGjAYBgNVBAMMEUJFVEEgU2lnbmluZyBDQSAxMCIYDzIwMTcwMTAxMDAwMDAwWhgP\nMjEwNTAxMzEyMzU5NTlaMFoxEzARBgoJkiaJk/IsZAEZFgNPUkcxGDAWBgoJkiaJ\nk/IsZAEZFghPcGVuWFBLSTENMAsGA1UECwwEQUNNRTEaMBgGA1UEAwwRQkVUQSBD\nbGllbnQgQm9iIDEwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDBPe68\niMzsOb8SMIOuMrezv9w6Ypix7vdR+Azk1yajgkPirHPYAuDq8E+oKVNVW2U3za/Y\nGT7ygyYrwbcjlNO9aDKUvx2aBx4ceOpSeG9mycRNaJf25us5veLRTF3P8eBHGF6y\nr9ov0w1bA50udLq+E16vqYzyst9UgsVnrSwi7QRdh+qT5bMeRC6vRJ56Ct/BXewV\nOo5h8xxvaxmBJIT8RNRij8AZmCLzax8h3+P6iPl0koQtHRSkpE/uNahTnXRj7w0z\ni+Ossfhps+Foib5Wzef3tf3n0ofWbAJQpZD7XRGsnBBC89g1p2zd1cgrCyC9f21n\nT4pymxBA0vk4mPeDAgMBAAGjTTBLMAkGA1UdEwQCMAAwHQYDVR0OBBYEFKfNgIJ+\neR3ZLBVxMyLLkwVSpu3OMB8GA1UdIwQYMBaAFNr7Xi+RxE6e9kZ7Ft0ewk5mmytI\nMA0GCSqGSIb3DQEBCwUAA4IBAQCd0nc2G8TwOA73uY1cd3il2Xo+bAVriqRaCmJu\nScwLADCrpWNzmqemI7oEBvVSU7eeyjCL45Cs3GnS8aPKkUXVkjSQdnvs8Hjqswo7\nO6zUf5UEmw0UUM8f4vr97SPFKeBQaHYH8PIgUz6eVhW51CWQ8voe6ZPaYQhrEKe+\nZ1Vxa6+dDQMxEeLIvxcy49+JgDgGj33VrB/VMHxo0/LaEChgGOeMWueoI3GwNKgZ\n0kXRXEGY+3qMVeT1Cl2j5egQQ8QkW8e0FP0pMbeINi+hjVi1flLrt175HuaUpuH0\niIWsf9WmB8XP8XDiQLwn+SrD1H8Dm0D0XL4wtHXkGDEUDotC\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'eWIFuvplej6anRW844iC_2Ib59k',
                invalidity_time => undef,
                issuer_dn => 'CN=BETA Signing CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'yWp_z39FC7zb9PCNX7xjEK7BPWw',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '1483228800', # 2017-01-01T00:00:00
                pki_realm => 'beta',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:c1:3d:ee:bc:88:cc:ec:39:bf:12:30:83:ae:32:\n    b7:b3:bf:dc:3a:62:98:b1:ee:f7:51:f8:0c:e4:d7:\n    26:a3:82:43:e2:ac:73:d8:02:e0:ea:f0:4f:a8:29:\n    53:55:5b:65:37:cd:af:d8:19:3e:f2:83:26:2b:c1:\n    b7:23:94:d3:bd:68:32:94:bf:1d:9a:07:1e:1c:78:\n    ea:52:78:6f:66:c9:c4:4d:68:97:f6:e6:eb:39:bd:\n    e2:d1:4c:5d:cf:f1:e0:47:18:5e:b2:af:da:2f:d3:\n    0d:5b:03:9d:2e:74:ba:be:13:5e:af:a9:8c:f2:b2:\n    df:54:82:c5:67:ad:2c:22:ed:04:5d:87:ea:93:e5:\n    b3:1e:44:2e:af:44:9e:7a:0a:df:c1:5d:ec:15:3a:\n    8e:61:f3:1c:6f:6b:19:81:24:84:fc:44:d4:62:8f:\n    c0:19:98:22:f3:6b:1f:21:df:e3:fa:88:f9:74:92:\n    84:2d:1d:14:a4:a4:4f:ee:35:a8:53:9d:74:63:ef:\n    0d:33:8b:e3:ac:b1:f8:69:b3:e1:68:89:be:56:cd:\n    e7:f7:b5:fd:e7:d2:87:d6:6c:02:50:a5:90:fb:5d:\n    11:ac:9c:10:42:f3:d8:35:a7:6c:dd:d5:c8:2b:0b:\n    20:bd:7f:6d:67:4f:8a:72:9b:10:40:d2:f9:38:98:\n    f7:83\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=BETA Client Bob 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'A7:CD:80:82:7E:79:1D:D9:2C:15:71:33:22:CB:93:05:52:A6:ED:CE',
            },
            db_alias => {
                alias => 'beta-bob-1',
                generation => undef,
                group_id => undef,
            },
        ),

        'beta_datavault_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'BETA DataVault 1',
            db => {
                authority_key_identifier => '37:E0:50:4B:7B:DC:B5:ED:70:F0:6E:DA:3A:F7:86:7B:5E:26:77:E9',
                cert_key => '21',
                data => "-----BEGIN CERTIFICATE-----\nMIIDoDCCAoigAwIBAgIBFTANBgkqhkiG9w0BAQsFADBZMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGTAXBgNVBAMMEEJFVEEgRGF0YVZhdWx0IDEwIhgPMjAxNzAxMDEwMDAwMDBaGA8y\nMTA1MDEzMTIzNTk1OVowWTETMBEGCgmSJomT8ixkARkWA09SRzEYMBYGCgmSJomT\n8ixkARkWCE9wZW5YUEtJMQ0wCwYDVQQLDARBQ01FMRkwFwYDVQQDDBBCRVRBIERh\ndGFWYXVsdCAxMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAwA34/kUh\nvSAhrmLNn7Cp7esL8siv+d9zlIiF5MDGpWsLfx6au9jsaGcFtooShIDwPIqKJQbm\nMNMdkbT9QqWgLcrqEz4uzsdOk/t7bLDSUCxA69YOOFg1LV12zhI6eeZ4YdNgP6oZ\n9x83F80xJ+NJPOlgRt8nN1QMlfgpGZyS4tCkDV90XZv/U7A3I/v/4PIsRSMRHT/1\n9+eF2D4PBAtDGvx4OcVyQNStf8BO7Bbn53wSHDqGWpnJ67qN5WoUxsLONBdQ8o9e\nuTY2zB216MFb6FmEA79DlNEnQ8gVqQl0UpsUA8rvO/tp1emjflotEbKMsUJnb5WN\naPKFsEzbxGr4qQIDAQABo28wbTAJBgNVHRMEAjAAMB0GA1UdDgQWBBQ34FBLe9y1\n7XDwbto694Z7XiZ36TAfBgNVHSMEGDAWgBQ34FBLe9y17XDwbto694Z7XiZ36TAL\nBgNVHQ8EBAMCBSAwEwYDVR0lBAwwCgYIKwYBBQUHAwQwDQYJKoZIhvcNAQELBQAD\nggEBAABjwa1uTPjnkJCnEWfgq3f5g0eSDHOetirhtB2L8oWZu4vKGxB32Q4Yh8KP\nUTtg19JoUotKaZs4YAzjkYKyjpzzhO/HNiRrV3MJ1cnFlO+RIwV+gl4OMFDBhlcA\nqU+VVc+0hjivm7cCAvLnPDiLMnGuyNnWmRqh/RABFupQcq57gEK06fhiwemciwBE\n9sfi2k4tBUFs8LSkFFLGXDFeM+2xleyHifWaEz0vjhP8LA+bvsakCH9WkIfOVbCA\ng2hBoRiDBwh/jVkQTlNBPXBsVkFnOVfRwji1uf84WHdHeeSWnIBV8ZY25zwr1aQP\nOearylCCz4DfUYpPbB66c9Sp29M=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'MolqbCrM99YCJwJlouoa3ESDXKw',
                invalidity_time => undef,
                issuer_dn => 'CN=BETA DataVault 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'MolqbCrM99YCJwJlouoa3ESDXKw',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '1483228800', # 2017-01-01T00:00:00
                pki_realm => 'beta',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:c0:0d:f8:fe:45:21:bd:20:21:ae:62:cd:9f:b0:\n    a9:ed:eb:0b:f2:c8:af:f9:df:73:94:88:85:e4:c0:\n    c6:a5:6b:0b:7f:1e:9a:bb:d8:ec:68:67:05:b6:8a:\n    12:84:80:f0:3c:8a:8a:25:06:e6:30:d3:1d:91:b4:\n    fd:42:a5:a0:2d:ca:ea:13:3e:2e:ce:c7:4e:93:fb:\n    7b:6c:b0:d2:50:2c:40:eb:d6:0e:38:58:35:2d:5d:\n    76:ce:12:3a:79:e6:78:61:d3:60:3f:aa:19:f7:1f:\n    37:17:cd:31:27:e3:49:3c:e9:60:46:df:27:37:54:\n    0c:95:f8:29:19:9c:92:e2:d0:a4:0d:5f:74:5d:9b:\n    ff:53:b0:37:23:fb:ff:e0:f2:2c:45:23:11:1d:3f:\n    f5:f7:e7:85:d8:3e:0f:04:0b:43:1a:fc:78:39:c5:\n    72:40:d4:ad:7f:c0:4e:ec:16:e7:e7:7c:12:1c:3a:\n    86:5a:99:c9:eb:ba:8d:e5:6a:14:c6:c2:ce:34:17:\n    50:f2:8f:5e:b9:36:36:cc:1d:b5:e8:c1:5b:e8:59:\n    84:03:bf:43:94:d1:27:43:c8:15:a9:09:74:52:9b:\n    14:03:ca:ef:3b:fb:69:d5:e9:a3:7e:5a:2d:11:b2:\n    8c:b1:42:67:6f:95:8d:68:f2:85:b0:4c:db:c4:6a:\n    f8:a9\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=BETA DataVault 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '37:E0:50:4B:7B:DC:B5:ED:70:F0:6E:DA:3A:F7:86:7B:5E:26:77:E9',
            },
            db_alias => {
                alias => 'beta-datavault-1',
                generation => '1',
                group_id => 'beta-datavault',
            },
        ),

        'beta_scep_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'BETA SCEP 1',
            db => {
                authority_key_identifier => '53:50:B0:41:55:82:2D:CA:4C:84:09:5C:D4:78:0F:2D:D8:0A:D0:AE',
                cert_key => '24',
                data => "-----BEGIN CERTIFICATE-----\nMIIDdzCCAl+gAwIBAgIBGDANBgkqhkiG9w0BAQsFADBXMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nFzAVBgNVBAMMDkJFVEEgUm9vdCBDQSAxMCIYDzIwMTcwMTAxMDAwMDAwWhgPMjEw\nNTAxMzEyMzU5NTlaMFQxEzARBgoJkiaJk/IsZAEZFgNPUkcxGDAWBgoJkiaJk/Is\nZAEZFghPcGVuWFBLSTENMAsGA1UECwwEQUNNRTEUMBIGA1UEAwwLQkVUQSBTQ0VQ\nIDEwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC5FMQr4BdKQAyjGuYv\nZ3VZvIct2RtzTOuWTifeVBSTcpLp5zn+43iwKZPTDsmAL3o/mdyZDQr7YiQp+LEo\nOXXXi29wz0+/Wp9kbCg9V7pxo5Bwwp1a8BlC+rgGw/1IF6GSZP62m7q/TI8QKFs7\nt0bmZ68qH5pwzHxywWhok9Md2nt1QeKywxtaxJ529rf+RnsjmT9mqyk7B+ZMAHuW\n6WX8v+K3tkuWepODzWrqHSaSCJsDGL/7uHvydHesFO+aGPitGA6yPpMsc9HtumGv\noI5INB6JXgd3Tdz3W/WPNtrc2wE4RG6CbMv9nDlwvwoYzdg7x0/kJf4IBRTI4wOf\n1KdbAgMBAAGjTTBLMAkGA1UdEwQCMAAwHQYDVR0OBBYEFKsEs+Y8aqjbWbg4T5NO\n2hzU3UPXMB8GA1UdIwQYMBaAFFNQsEFVgi3KTIQJXNR4Dy3YCtCuMA0GCSqGSIb3\nDQEBCwUAA4IBAQCuAWFuZSdgi0JIlK9sbBf6QouPHazym/8742MQeN3xJRSOtpxY\n3NlFprsbDEjkHqBPh5BUDGrLpyUkPGBYy9+zCh19yjFGpFU7feU91iYFAT9GoGPr\nt3rXrAkUMlTAlBo+0W2N6l3LPVRkhwdEn90QN3jQ+xgdtquMb7yZpAgigWEOVXrK\nYWfdirq5n+Jpvx5eVYMOHlKWY7ZC2M/di3YJ5dJJ5Kpw7QuwDMdfDIeI+4i+gqPV\nwqQSe79F+c+/ov9jiN5J3JY0dfEliFNbyhDber8qvRrxa0HUCGjpvH3h0VaMx6i6\nI8OhuMbexuOZ8Nv/C67WTLU6ZCjcg4jtwo6M\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => '5kXEkN9WcTQvQhKnAH3r4Eqzek0',
                invalidity_time => undef,
                issuer_dn => 'CN=BETA Root CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'kT9lAWcLbGhCt0Vs-iJGWRHI-wA',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '1483228800', # 2017-01-01T00:00:00
                pki_realm => 'beta',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:b9:14:c4:2b:e0:17:4a:40:0c:a3:1a:e6:2f:67:\n    75:59:bc:87:2d:d9:1b:73:4c:eb:96:4e:27:de:54:\n    14:93:72:92:e9:e7:39:fe:e3:78:b0:29:93:d3:0e:\n    c9:80:2f:7a:3f:99:dc:99:0d:0a:fb:62:24:29:f8:\n    b1:28:39:75:d7:8b:6f:70:cf:4f:bf:5a:9f:64:6c:\n    28:3d:57:ba:71:a3:90:70:c2:9d:5a:f0:19:42:fa:\n    b8:06:c3:fd:48:17:a1:92:64:fe:b6:9b:ba:bf:4c:\n    8f:10:28:5b:3b:b7:46:e6:67:af:2a:1f:9a:70:cc:\n    7c:72:c1:68:68:93:d3:1d:da:7b:75:41:e2:b2:c3:\n    1b:5a:c4:9e:76:f6:b7:fe:46:7b:23:99:3f:66:ab:\n    29:3b:07:e6:4c:00:7b:96:e9:65:fc:bf:e2:b7:b6:\n    4b:96:7a:93:83:cd:6a:ea:1d:26:92:08:9b:03:18:\n    bf:fb:b8:7b:f2:74:77:ac:14:ef:9a:18:f8:ad:18:\n    0e:b2:3e:93:2c:73:d1:ed:ba:61:af:a0:8e:48:34:\n    1e:89:5e:07:77:4d:dc:f7:5b:f5:8f:36:da:dc:db:\n    01:38:44:6e:82:6c:cb:fd:9c:39:70:bf:0a:18:cd:\n    d8:3b:c7:4f:e4:25:fe:08:05:14:c8:e3:03:9f:d4:\n    a7:5b\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=BETA SCEP 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'AB:04:B3:E6:3C:6A:A8:DB:59:B8:38:4F:93:4E:DA:1C:D4:DD:43:D7',
            },
            db_alias => {
                alias => 'beta-scep-1',
                generation => '1',
                group_id => 'beta-scep',
            },
        ),

        'beta_signer_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'BETA Signing CA 1',
            db => {
                authority_key_identifier => '53:50:B0:41:55:82:2D:CA:4C:84:09:5C:D4:78:0F:2D:D8:0A:D0:AE',
                cert_key => '23',
                data => "-----BEGIN CERTIFICATE-----\nMIIDkDCCAnigAwIBAgIBFzANBgkqhkiG9w0BAQsFADBXMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nFzAVBgNVBAMMDkJFVEEgUm9vdCBDQSAxMCIYDzIwMTcwMTAxMDAwMDAwWhgPMjEw\nNTAxMzEyMzU5NTlaMFoxEzARBgoJkiaJk/IsZAEZFgNPUkcxGDAWBgoJkiaJk/Is\nZAEZFghPcGVuWFBLSTENMAsGA1UECwwEQUNNRTEaMBgGA1UEAwwRQkVUQSBTaWdu\naW5nIENBIDEwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCy0xNHV8nj\nEjYJCJrlw2SXGMurCER/2GLbD8wbpTk4nq97tThVtEheoMn658euTAnnqUNfMPx+\nrj3uwWwzsSiUp8ihPYHIw3id9gJT5SYTFMnphRR6yGys2cyCidgVnR1hIHT7yA6U\n2WjGKqJuMZMD6wv+G/RUqfJNDqJIl57zgkHV1MvykRGSi8PDTRRkEkBZapQTD8ct\nnDKkcGIG7io8COJ9hPCCz6a2dCMRAkSZ4t+6GYhXGaassesp43N8IXvq+dop20p+\nQh3qy68YB8NQx3IbeYtWFAhTp1wR9L1V+c8ngYkImxZjxzYWT/ZyazODZQmuJKpT\nt3/pKTmwq0MFAgMBAAGjYDBeMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFNr7\nXi+RxE6e9kZ7Ft0ewk5mmytIMB8GA1UdIwQYMBaAFFNQsEFVgi3KTIQJXNR4Dy3Y\nCtCuMAsGA1UdDwQEAwIBBjANBgkqhkiG9w0BAQsFAAOCAQEAirG7ZciJvgI6WfPC\nLvCtO8+4YNNPzWM5lAQWjx0trhGCFX1MBgaiTzDCYGMeT8GrEM3J/DhJlBLTA/rf\nFlNONBIantnP8pNnGp5xZArhRZpgFs/DSUfiVF23Vzbm3CYBiQEs0D+pp/hvSyCU\nbqnd8zIgoBYha7MERjGuJnSzB6bpt5vqVZTjxvoQWrvg/t1G26+SZI/z41N+lEWZ\nwCKpIS1Jay5StLLlSuH+BoJasYDfHAbzbVKVmBvQZYRw0T7WoafEnkU0IO3HNZEY\n0CLB4Dk9DNqtAbO72FfDiMYzaqTbxoCehQYTrqeKJ8oGnvqzZC2G5Rule/pDDfHs\n4/Byog==\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'yWp_z39FC7zb9PCNX7xjEK7BPWw',
                invalidity_time => undef,
                issuer_dn => 'CN=BETA Root CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'kT9lAWcLbGhCt0Vs-iJGWRHI-wA',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '1483228800', # 2017-01-01T00:00:00
                pki_realm => 'beta',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:b2:d3:13:47:57:c9:e3:12:36:09:08:9a:e5:c3:\n    64:97:18:cb:ab:08:44:7f:d8:62:db:0f:cc:1b:a5:\n    39:38:9e:af:7b:b5:38:55:b4:48:5e:a0:c9:fa:e7:\n    c7:ae:4c:09:e7:a9:43:5f:30:fc:7e:ae:3d:ee:c1:\n    6c:33:b1:28:94:a7:c8:a1:3d:81:c8:c3:78:9d:f6:\n    02:53:e5:26:13:14:c9:e9:85:14:7a:c8:6c:ac:d9:\n    cc:82:89:d8:15:9d:1d:61:20:74:fb:c8:0e:94:d9:\n    68:c6:2a:a2:6e:31:93:03:eb:0b:fe:1b:f4:54:a9:\n    f2:4d:0e:a2:48:97:9e:f3:82:41:d5:d4:cb:f2:91:\n    11:92:8b:c3:c3:4d:14:64:12:40:59:6a:94:13:0f:\n    c7:2d:9c:32:a4:70:62:06:ee:2a:3c:08:e2:7d:84:\n    f0:82:cf:a6:b6:74:23:11:02:44:99:e2:df:ba:19:\n    88:57:19:a6:ac:b1:eb:29:e3:73:7c:21:7b:ea:f9:\n    da:29:db:4a:7e:42:1d:ea:cb:af:18:07:c3:50:c7:\n    72:1b:79:8b:56:14:08:53:a7:5c:11:f4:bd:55:f9:\n    cf:27:81:89:08:9b:16:63:c7:36:16:4f:f6:72:6b:\n    33:83:65:09:ae:24:aa:53:b7:7f:e9:29:39:b0:ab:\n    43:05\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=BETA Signing CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'DA:FB:5E:2F:91:C4:4E:9E:F6:46:7B:16:DD:1E:C2:4E:66:9B:2B:48',
            },
            db_alias => {
                alias => 'beta-signer-1',
                generation => '1',
                group_id => 'beta-signer',
            },
        ),

        'beta_root_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'BETA Root CA 1',
            db => {
                authority_key_identifier => '53:50:B0:41:55:82:2D:CA:4C:84:09:5C:D4:78:0F:2D:D8:0A:D0:AE',
                cert_key => '22',
                data => "-----BEGIN CERTIFICATE-----\nMIIDjTCCAnWgAwIBAgIBFjANBgkqhkiG9w0BAQsFADBXMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nFzAVBgNVBAMMDkJFVEEgUm9vdCBDQSAxMCIYDzIwMTcwMTAxMDAwMDAwWhgPMjEw\nNTAxMzEyMzU5NTlaMFcxEzARBgoJkiaJk/IsZAEZFgNPUkcxGDAWBgoJkiaJk/Is\nZAEZFghPcGVuWFBLSTENMAsGA1UECwwEQUNNRTEXMBUGA1UEAwwOQkVUQSBSb290\nIENBIDEwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDIXyjwCUqIDOkW\nmh2aZnDPYtDIpXIZhbBjT6FZZ3Kiwdom1cBFRcLkeH+mMFMOdsTPYjGLctQSnNJd\nLhzoTBYa1YdvWG20MA5K0JPzZTdlyQJmLNYlyGuXtWTPwriiVsXIUGFC3aHbxeXr\nnIFrU3UPOz5aG3hDdYaxv0A/+up+R98aa8iH7UnpD+ddBFxpwsVAknzrBlJ1ec3X\n3sS1Bd0itcj7EeWr4Pr94m7D+LYdlNx2LK8eTxavAXw0nHMjQBjb/C3U4nfJ2fKW\nUXDTTDpJlCWgv5W5SA6umh3mdLqu0MlXeJCOj4bfrXC9Bz7p4GURDYoae2QZ6HFR\nVkavDGXTAgMBAAGjYDBeMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFFNQsEFV\ngi3KTIQJXNR4Dy3YCtCuMB8GA1UdIwQYMBaAFFNQsEFVgi3KTIQJXNR4Dy3YCtCu\nMAsGA1UdDwQEAwIBBjANBgkqhkiG9w0BAQsFAAOCAQEAxlYZPTQHIz1LeijyaVSE\np1cxoa4XznKmA2ntaIZqnT4VgbgADeKzXwRS2gC2zBkf35H7JT82GSiT9jqtdp1K\n7XHKdQR6NIHXdLrt08SEYxlGSc0WKamusja32+0dKxVnzYJ2Fy2C8SBE5b2Yyo1b\nP/73MHkgd4g6T9G1NcnmJZzdDRZ1PdZ+kaECYlXo7OJZLvWu9MHkOKXczJ2wrqsF\nnjDAefP7VWBiQAgZIGzYF0y2a9TOYKgEibFlsBNTyfC0GW3ZkhZGoQtnVMI5MW/W\nzTEMd9mW0BGkWrRUIubLjpGHPJB73lOd/U2KFV5uSF/zdq8h7tCFv+NigWHQNFAR\n5w==\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'kT9lAWcLbGhCt0Vs-iJGWRHI-wA',
                invalidity_time => undef,
                issuer_dn => 'CN=BETA Root CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => 'kT9lAWcLbGhCt0Vs-iJGWRHI-wA',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '1483228800', # 2017-01-01T00:00:00
                pki_realm => 'beta',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:c8:5f:28:f0:09:4a:88:0c:e9:16:9a:1d:9a:66:\n    70:cf:62:d0:c8:a5:72:19:85:b0:63:4f:a1:59:67:\n    72:a2:c1:da:26:d5:c0:45:45:c2:e4:78:7f:a6:30:\n    53:0e:76:c4:cf:62:31:8b:72:d4:12:9c:d2:5d:2e:\n    1c:e8:4c:16:1a:d5:87:6f:58:6d:b4:30:0e:4a:d0:\n    93:f3:65:37:65:c9:02:66:2c:d6:25:c8:6b:97:b5:\n    64:cf:c2:b8:a2:56:c5:c8:50:61:42:dd:a1:db:c5:\n    e5:eb:9c:81:6b:53:75:0f:3b:3e:5a:1b:78:43:75:\n    86:b1:bf:40:3f:fa:ea:7e:47:df:1a:6b:c8:87:ed:\n    49:e9:0f:e7:5d:04:5c:69:c2:c5:40:92:7c:eb:06:\n    52:75:79:cd:d7:de:c4:b5:05:dd:22:b5:c8:fb:11:\n    e5:ab:e0:fa:fd:e2:6e:c3:f8:b6:1d:94:dc:76:2c:\n    af:1e:4f:16:af:01:7c:34:9c:73:23:40:18:db:fc:\n    2d:d4:e2:77:c9:d9:f2:96:51:70:d3:4c:3a:49:94:\n    25:a0:bf:95:b9:48:0e:ae:9a:1d:e6:74:ba:ae:d0:\n    c9:57:78:90:8e:8f:86:df:ad:70:bd:07:3e:e9:e0:\n    65:11:0d:8a:1a:7b:64:19:e8:71:51:56:46:af:0c:\n    65:d3\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=BETA Root CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => '53:50:B0:41:55:82:2D:CA:4C:84:09:5C:D4:78:0F:2D:D8:0A:D0:AE',
            },
            db_alias => {
                alias => 'root-1',
                generation => '1',
                group_id => 'root',
            },
        ),

        'gamma_bob_1' => OpenXPKI::Test::CertHelper::Database::PEM->new(
            label => 'GAMMA Client Bob 1',
            db => {
                authority_key_identifier => '65:81:A5:F1:F9:78:C3:D5:18:D9:9F:DF:5E:61:92:64:1E:FC:27:10',
                cert_key => '32',
                data => "-----BEGIN CERTIFICATE-----\nMIIDgjCCAmqgAwIBAgIBIDANBgkqhkiG9w0BAQsFADBbMRMwEQYKCZImiZPyLGQB\nGRYDT1JHMRgwFgYKCZImiZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUx\nGzAZBgNVBAMMEkdBTU1BIFNpZ25pbmcgQ0EgMTAiGA8yMDE3MDEwMTAwMDAwMFoY\nDzIxMDUwMTMxMjM1OTU5WjBbMRMwEQYKCZImiZPyLGQBGRYDT1JHMRgwFgYKCZIm\niZPyLGQBGRYIT3BlblhQS0kxDTALBgNVBAsMBEFDTUUxGzAZBgNVBAMMEkdBTU1B\nIENsaWVudCBCb2IgMTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALMy\na9JIvCGZPOxqEjYd9WOhKWu+vGqBObwiEJR/5Kq+88/7VKgqjwC309wqBjndWC7n\niJYU06P/IudKolenQxZc0YOFlqenU6iWhvHqe9oyfLMlEoKc1T+zqImGkgkOAevK\nCvOGyP2H85x/RF0WQWMBkRokjYGENG+m5mj/d9CCN5fy3b/gln867CwDUhBoj3G5\niRic/x5C2KFlipNC1/Ex/Mpd5SM3rPfLGXoLM8qs+SZY3gk8STMoqEPe94TaZVzq\nJphqRjFYuXDx0Fc+4zjidDMiDw2gyy7pi6jOl96CPsQ1wn2/Z9To8OpepU9IAVIV\n2BiZoehX/7jnRrVA/w0CAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4EFgQU4+Ln\nfzWfnsPYAewoFnnv4pkdVoswHwYDVR0jBBgwFoAUZYGl8fl4w9UY2Z/fXmGSZB78\nJxAwDQYJKoZIhvcNAQELBQADggEBAB9zSI1fmuBPIhp+t3figtLYum3SaqM9Nu5K\nB2FTvA88Dj4XIfEwoEzpGTDPfqQTM7W92FcmLN4BHySlRRF/83jFSENLDj/JEFsi\n01S0Lf3Fuf9TF8HpPQ5+phEGyIpWDAhEZ2b2K7pk1j4MmTT4F/ZHwNcbmgZuHUvD\ntp0mYaFcIAyIo86lHdS7e8JkYG2k11dqHcEExMdS21t9MG6QZqDE6WuHDr8cJN0U\nN95M8jHoKIvycd6vaKccT7HyXKOMxn5S/MNCrCdQUs1kkfd6S6zCtKByLBhYUvBh\nuNnECXvzjTVuf9pEWuZ2o+0ip2cu4Cxw3zjdv/rpFd5bI75/myE=\n-----END CERTIFICATE-----",
                hold_instruction_code => undef,
                identifier => 'BZLZtpmFb6OA7yHH3xJ3CojndCE',
                invalidity_time => undef,
                issuer_dn => 'CN=GAMMA Signing CA 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                issuer_identifier => '',
                notafter => '4262889599', # 2105-01-31T23:59:59
                notbefore => '1483228800', # 2017-01-01T00:00:00
                pki_realm => 'gamma',
                public_key => "Public-Key: (2048 bit)\nModulus:\n    00:b3:32:6b:d2:48:bc:21:99:3c:ec:6a:12:36:1d:\n    f5:63:a1:29:6b:be:bc:6a:81:39:bc:22:10:94:7f:\n    e4:aa:be:f3:cf:fb:54:a8:2a:8f:00:b7:d3:dc:2a:\n    06:39:dd:58:2e:e7:88:96:14:d3:a3:ff:22:e7:4a:\n    a2:57:a7:43:16:5c:d1:83:85:96:a7:a7:53:a8:96:\n    86:f1:ea:7b:da:32:7c:b3:25:12:82:9c:d5:3f:b3:\n    a8:89:86:92:09:0e:01:eb:ca:0a:f3:86:c8:fd:87:\n    f3:9c:7f:44:5d:16:41:63:01:91:1a:24:8d:81:84:\n    34:6f:a6:e6:68:ff:77:d0:82:37:97:f2:dd:bf:e0:\n    96:7f:3a:ec:2c:03:52:10:68:8f:71:b9:89:18:9c:\n    ff:1e:42:d8:a1:65:8a:93:42:d7:f1:31:fc:ca:5d:\n    e5:23:37:ac:f7:cb:19:7a:0b:33:ca:ac:f9:26:58:\n    de:09:3c:49:33:28:a8:43:de:f7:84:da:65:5c:ea:\n    26:98:6a:46:31:58:b9:70:f1:d0:57:3e:e3:38:e2:\n    74:33:22:0f:0d:a0:cb:2e:e9:8b:a8:ce:97:de:82:\n    3e:c4:35:c2:7d:bf:67:d4:e8:f0:ea:5e:a5:4f:48:\n    01:52:15:d8:18:99:a1:e8:57:ff:b8:e7:46:b5:40:\n    ff:0d\nExponent: 65537 (0x10001)\n",
                reason_code => undef,
                req_key => undef,
                revocation_time => undef,
                status => 'ISSUED',
                subject => 'CN=GAMMA Client Bob 1,OU=ACME,DC=OpenXPKI,DC=ORG',
                subject_key_identifier => 'E3:E2:E7:7F:35:9F:9E:C3:D8:01:EC:28:16:79:EF:E2:99:1D:56:8B',
            },
            db_alias => {
                alias => 'gamma-bob-1',
                generation => undef,
                group_id => undef,
            },
        ),

    };
}

__PACKAGE__->meta->make_immutable;
